package de.desy.acop.video.analysis;

import java.awt.image.BufferedImage;

import de.desy.acop.video.displayer.ColorMap;
import de.desy.acop.video.timageio.TBufferedImage;
import de.desy.tine.types.IMAGE;

/**
 * <code>ImageAnalyzer</code> implements the image analysis. There can be different implementation of image analysis,
 * depending on the source and type of the image (e.g. beam cross section, interferometer image).
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 */
public abstract class ImageAnalyzer {

    /**
     * Finds the minimum and maximum value in the array looking at the elements between start and start+length.
     * 
     * @param values the values to check
     * @param start the starting index
     * @param length the length of the check
     * @return the triple min/max/index
     */
    protected static double[] findMinMax(double[] values, int start, int length) {
        double min = Double.MAX_VALUE;
        double max = -Double.MAX_VALUE;
        int index = 0;
        for (int i = start; i < start + length; i++) {
            if (values[i] > max) {
                max = values[i];
                index = i;
            }
            if (values[i] < min) min = values[i];
        }
        return new double[]{min,max,index};
    }
    
    /**
     * Perform the analysis and return the analysis image.
     * 
     * @param colorMap
     * @param decoder
     * @param image
     * @param preciseBackground
     * @param imageWidth
     * @param imageHeight
     * @param roiX
     * @param roiY
     * @param roiW
     * @param roiH
     * @param roi2x
     * @param roi2y
     * @param roi2w
     * @param roi2h
     * @param threshold
     * @param calculateThreshold
     * @param performFit
     * @param performSmoothing
     * @return
     */
    public AnalysisImage performAnalysis(ColorMap colorMap, ColorDecoder decoder, IMAGE image,
            double[] preciseBackground, int imageWidth, int imageHeight, int roiX, int roiY, int roiW, int roiH,
            int roi2x, int roi2y, int roi2w, int roi2h, double threshold, boolean calculateThreshold,
            boolean performFit, boolean performSmoothing) {
        IMAGE newImage = image;
        int i, j;
        int[] data;
        double[] values;
        int aH = image.getFrameHeader().aoiHeight;
        int aW = image.getFrameHeader().aoiWidth;
        if (aH != -1) {
            imageWidth = aW;
            imageHeight = aH;
        }
        if (preciseBackground != null) {
            BufferedImage bufferedImage = ImageAnalysisEngine.toBufferedImage(image,ColorMap.NONE);
            if (bufferedImage == null) throw new IllegalStateException("No BufferedImage!");
            data = bufferedImage.getRGB(0,0,imageWidth,imageHeight,null,0,imageWidth);
            values = decoder.transform(data);
            try {
                i = 0;
                //try catch is actually faster than anything else
                for (;;) {
                    values[i] -= preciseBackground[i];
                    i++;
                }
            } catch (ArrayIndexOutOfBoundsException e) {
                //ignore
            }
            encode(data,values);
            int type = 0;
            if (image.getFrameHeader().bytesPerPixel == 1) {
                type = BufferedImage.TYPE_BYTE_GRAY;
            } else if (image.getFrameHeader().bytesPerPixel == 2) {
                type = BufferedImage.TYPE_USHORT_GRAY;
            } else {
                type = BufferedImage.TYPE_INT_RGB;
            }
            bufferedImage = new BufferedImage(bufferedImage.getWidth(),bufferedImage.getHeight(),type);
            bufferedImage.setRGB(0,0,imageWidth,imageHeight,data,0,imageWidth);
            newImage = TBufferedImage.toIMAGE(bufferedImage);
        }
        BufferedImage bufferedImage = ImageAnalysisEngine.toBufferedImage(newImage,ColorMap.GRAYSCALE);
        if (bufferedImage == null) throw new IllegalStateException("No BufferedImage!");
        data = bufferedImage.getRGB(0,0,imageWidth,imageHeight,null,0,imageWidth);
        values = decoder.transform(data);
        //calculate the threshold value
        //we want to "fit" A e^(x-B)^2/2C^2 + D + F x, where D is the threshold
        //to find D, sum up the values in the threshold ROI and outside ROI and calculate the
        //average value. That is the new threshold, if there wasn't any specified by the user
        //      logger.log(Level.FINE, "Analysis: Calculating threshold.");
        int thresholdPointsCount = 0;
        if (calculateThreshold) {
            double sum = 0;
            int r = roiX + roiW, b = roiY + roiH;
            boolean inside = false;
            for (j = 0; j < roi2h; j++) {
                inside = roiY <= roi2y + j && roi2y + j < b;
                for (i = 0; i < roi2w; i++) {
                    //do not include indexes, which are in normal ROI
                    if (roiX <= roi2x + i && roi2x + i < r && inside) continue;
                    sum += values[i + roi2x + imageWidth * (j + roi2y)];
                    thresholdPointsCount++;
                }
            }
            if (thresholdPointsCount != 0) {
                threshold = sum / thresholdPointsCount;
            } else {
                threshold = Double.NaN;
            }
        }
        return doAnalysis(newImage,values,imageWidth,imageHeight,roiX,roiY,roiW,roiH,roi2x,roi2y,roi2w,roi2h,
                thresholdPointsCount,threshold,calculateThreshold,performFit,performSmoothing);
    }

    protected int[] encode(int[] ret, double[] data) {
        int length = data.length;
        int code;
        for (int i = 0; i < length; i++) {
            code = (int)(data[i] + 0.5);
            if (code < 0) code = 0;
            ret[i] = ((255 << 24) & 0xFF000000) + ((code << 16) & 0xFF0000) + ((code << 8) & 0xFF00) + (code & 0xFF);
        }
        return ret;
    }

    protected abstract AnalysisImage doAnalysis(IMAGE newImage, double[] values, int iW, int iH, int rX, int rY, int rW,
            int rH, int r2X, int r2Y, int r2W, int r2H, int thresholdPointsCount, double threshold,
            boolean calculateThreshold, boolean performFit, boolean smoothing);
    
    public abstract AnalysisType getAnalysisType();
}
