/**
 * File: Agency.java
 * Date: Aug 30, 2010
 */
package de.desy.mst.libs.accelerator;

import java.util.HashMap;
import java.util.Map;

/**
 * An agency for tracing methods in particular manner. Currently this
 * means that you can call methods with specified number of calls before
 * you switch to another call which allows to declare a delay time before
 * performing the statement in an if-clause. There also exists a reset
 * method which resets all internal values to their obtained ones. 
 * 
 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
 * @version $Id: Agency.java 40094 2010-11-01 08:52:37Z walla $
 */
public final class Agency
{
	// --- CLASS FIELDS ----------------------------------------------------
	
	private final static Map<String, Tracer> tracers = new HashMap<String, Tracer>();

	// --- CLASS METHODS ---------------------------------------------------
	
	/**
	 * Gets the current {@link Tracer} from the string value. If it 
	 * doesn't exist a new instance is created. If the string is
	 * {@code null} or empty then a {@code null} is returned.
	 * @param value The request string value. It is the key element which 
	 * 		returns the corresponding tracer.
	 * @return Returns {@link Tracer} depending on the current string value.
	 */
	public static Agency.Tracer getTracer(final String value) {
		if (value == null || value.isEmpty())
			return null;
		else if (!Agency.tracers.containsKey(value))
			Agency.tracers.put(value, new Tracer());
		return Agency.tracers.get(value);
	}
	
	/**
	 * Tracer for periodic called method, for example call back methods.
	 *
	 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
	 * @version $Id: Agency.java 40094 2010-11-01 08:52:37Z walla $
	 */
	public static class Tracer {
		
		// --- CLASS FIELDS ----------------------------------------------------
		
		/** Number of calls before the method {@link #isCounterValid()} returns
		 * a {@code false}. This is currently {@code 3}. */
		public final int LIMIT_CALLS = 3;
		/**  Limit of time stamp calls in millisecond. Currently it is set to
		 * 1 minute. */
		public final long LIMIT_TIME_STAMP = 1000 * 60;
		
		// --- INSTANCE FIELDS -------------------------------------------------
		
		private final long limitTimeStamp;
		private int counter;
		private long since;
		
		// --- CONSTRUCTOR -----------------------------------------------------
		
		/**
		 * Constructor. The limit of time stamp difference is set to 60 * 1000 * 10 
		 * milliseconds before this class reacts and resets the time stamp to
		 * the current called one.
		 */
		Tracer() {
			this.limitTimeStamp = LIMIT_TIME_STAMP;
			reset();
		}
		
		/**
		 * Constructor. The limit of time stamp difference is set to 60 * 1000 * 10 
		 * milliseconds before this class reacts and resets the time stamp to
		 * the current called one.
		 */
		Tracer(final long limitTimeStamp) {
			this.limitTimeStamp = Math.max(LIMIT_TIME_STAMP, limitTimeStamp);
			reset();
		}
		
		// --- INSTANCE METHODS ------------------------------------------------
		
		/**
		 * Resets all internal values to their origin state, i.e. the counter
		 * and the current time stamp.
		 */
		public void reset() { 
			this.since = System.currentTimeMillis(); 
			this.counter = 0; 
		}
		
		/**
		 * Request if the counter hasn't reached the limit. If not it increments
		 * the counter.
		 * @return Request if the counter is valid, i.e. if it is lower then
		 * 		{@link #LIMIT_CALLS}.
		 */
		public boolean isCounterValid() {
			final boolean isCounterValid = this.counter < LIMIT_CALLS;
			if (isCounterValid) this.counter += 1;
			return isCounterValid;
		}
		
		/**
		 * Request if the current system time stamp has reached its upper limit,
		 * where the time stamp is taken on runtime via 
		 * {@link java.lang.System#currentTimeMillis()}. 
		 * The upper limit shouldn't defect a time differences which is set in 
		 * the constructor. 
		 * @return Request if the time stamp has reached its upper limit.
		 */
		public boolean isElapsed() {
			final long currentTime = System.currentTimeMillis();
			final boolean isDelayValid = Math.abs(currentTime - this.since) < this.limitTimeStamp; 
			if (!isDelayValid)
				since = currentTime;
			return !isDelayValid;
		}
		
		/**
		 * Request if the current system time stamp has reached its upper limit.
		 * The upper limit shouldn't defect a time differences which is set in 
		 * the constructor. 
		 * @param currentTime current time in milliseconds.
		 * @return Request if the time stamp has reached its upper limit.
		 */
		@Deprecated
		public boolean isElapsed(final long currentTime) {
			final boolean isDelayValid = Math.abs(currentTime - this.since) < this.limitTimeStamp; 
			if (!isDelayValid)
				since = currentTime;
			return !isDelayValid;
		}
	}
}
/* end of file */