/*
 *
 * File: ApplicationInitialization.java
 * Date: Apr 16, 2008; 4:18:40 PM
 */
package de.desy.mst.libs.accelerator;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.List;

import javax.swing.Timer;

import de.desy.mst.libs.accelerator.exception.EnumerationNotFoundException;
import de.desy.mst.libs.accelerator.exception.TineException;
import de.desy.mst.libs.accelerator.model.ICDIDevice;
import de.desy.mst.libs.accelerator.model.ICDIDevice.EApplicationModel;
import de.desy.mst.libs.accelerator.view.CColor;
import de.desy.mst.libs.accelerator.view.RFDialog;
import de.desy.mst.libs.accelerator.view.RFDialog.IVersion;
import de.desy.mst.libs.framework.mstapp.logging.LogLvl;
import de.desy.mst.libs.framework.mstapp.logging.Logger;
import de.desy.mst.libs.framework.mstapp.logging.LoggerManager;
import de.desy.mst.libs.framework.mstapp.logging.appender.ILogAppender;
import de.desy.mst.libs.framework.mstapp.logging.filter.FilterModus;

/**
 * Initializes the application by querying generic command line arguments, for
 * example the logging level for displaying information for the user.
 *
 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
 * @version $Id: ApplicationInitialization.java 53957 2013-02-25 18:46:24Z walla $
 */
public final class ApplicationInitialization {
	/** Class name. */
	public static final String CLAZZ = ApplicationInitialization.class.getSimpleName();
	
	/** Is the default delay time after initializing the application in
	 * seconds. */
	public static final int INIT_DELAY_ERROR_DEFAULT = 15000;

	/** Predefined error string for missing CDI servers. */
	private static final String ERROR_PRE_STRING = "FEHLER bei Initialisierung des Servers ";

	/** Is the delay time after initializing the application. */
	public static int initDelayError = -1;

	// --- CLASS METHODS -------------------------------------------------------
	
	/**
	 * Initialize the whole application and analyzed all command line arguments 
	 * from the delivered pass-through array.
	 * 
	 * @param args_ All command line arguments.
	 */
	public static void initialize(final String... args_) {
		AppArguments.load(args_);
		initialize();
	}

	/**
	 * Initialize the whole application and analyzed all provoked command line 
	 * arguments from the static class {@link AppArguments}.
	 */
	public static void initialize() {
		final ILogAppender consolAppender = LoggerManager.getConsolAppender();
		consolAppender.getFilter().setModus(FilterModus.STANDARD);

		// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		// Request the LOG severity for reporting messages 
		//
		if (!AppArguments.getString(CGeneralConstant.CLA_LOG).isEmpty()) {
			if (AppArguments.getString(CGeneralConstant.CLA_LOG).toUpperCase().equals(CGeneralConstant.DEBUG_DEBUG)) {
				Logger.SWApp.info(ApplicationInitialization.CLAZZ + " - setting the logging level to " + CGeneralConstant.DEBUG_DEBUG);
				consolAppender.getFilter().setLogLvl(LogLvl.debug);
			}
			else if (AppArguments.getString(CGeneralConstant.CLA_LOG).toUpperCase().equals(CGeneralConstant.DEBUG_INFO)) {
				Logger.SWApp.info(ApplicationInitialization.CLAZZ + " - setting the logging level to " + CGeneralConstant.DEBUG_INFO);
				consolAppender.getFilter().setLogLvl(LogLvl.info);
			}
			else if (AppArguments.getString(CGeneralConstant.CLA_LOG).toUpperCase().equals(CGeneralConstant.DEBUG_WARN)) {
				Logger.SWApp.info(ApplicationInitialization.CLAZZ + " - setting the logging level to " + CGeneralConstant.DEBUG_WARN);
				consolAppender.getFilter().setLogLvl(LogLvl.warn);
			}
			else if (AppArguments.getString(CGeneralConstant.CLA_LOG).toUpperCase().equals(CGeneralConstant.DEBUG_ERROR)) {
				Logger.SWApp.info(ApplicationInitialization.CLAZZ + " - setting the logging level to " + CGeneralConstant.DEBUG_ERROR);
				consolAppender.getFilter().setLogLvl(LogLvl.error);
			}
			else if (AppArguments.getString(CGeneralConstant.CLA_LOG).toUpperCase().equals(CGeneralConstant.DEBUG_FATAL)) {
				Logger.SWApp.info(ApplicationInitialization.CLAZZ + " - setting the logging level to " + CGeneralConstant.DEBUG_FATAL);
				consolAppender.getFilter().setLogLvl(LogLvl.fatal);
			}
			else {
				Logger.SWApp.warn(ApplicationInitialization.CLAZZ + " - invalid command line argument for debug (debug/info/warn/error/fatal): " + AppArguments.getString(CGeneralConstant.CLA_LOG) + " -> setting to info");
				consolAppender.getFilter().setLogLvl(LogLvl.info);
			}
		}
		else
			consolAppender.getFilter().setLogLvl(LogLvl.info);
	
		// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		// Request the INITIAL DELAY TIME in seconds
		//
		if (AppArguments.getNumber(CGeneralConstant.CLA_INITDELAY) != null) {
			initDelayError = Math.abs(AppArguments.getNumber(CGeneralConstant.CLA_INITDELAY).intValue());
			if (initDelayError == 0 || initDelayError > 60000) {
				Logger.SWApp.warn(ApplicationInitialization.CLAZZ + " - delay time error before the server starts in milliseconds is 0 or greater then 60000 -> setting to default of " + INIT_DELAY_ERROR_DEFAULT + " milliseconds");
				initDelayError = INIT_DELAY_ERROR_DEFAULT;
			}
			else
				Logger.SWApp.info(ApplicationInitialization.CLAZZ + " - setting delay time before starting the server if a CDI error occurs to " + initDelayError + " milliseconds");
		}
	}

	// -------------------------------------------------------------------------

	/**
	 * This method is <b>deprecated</b>. You should use the method {@link #initializeCDI(EApplicationModel, int, String[], ICDIDevice, IVersion...)}
	 * instead. This method initializes the application model <b>server</b>.
	 *
	 * @param cdiServers_ A string array which contains all possible TINE
	 *  		pre-addresses.
	 * @param cdiDevice_ CDI device singleton.
	 * @param version_ All version instances.
	 * @see #initializeCDI(int, String[], ICDIDevice, IVersion...)
	 * @see #initializeCDI(int, int, String[], ICDIDevice, IVersion...)
	 * @see #initializeCDI(int, int, int, String[], ICDIDevice, IVersion...)
	 */
	@Deprecated
	public final static void initializeCDI(
			final String[] cdiServers_, 
			final ICDIDevice cdiDevice_,
			final IVersion... version_) {
		initializeCDI(EApplicationModel.SERVER, 15, cdiServers_, cdiDevice_, version_);
	}

	/**
	 * This method is <b>deprecated</b>. You should use the method {@link #initializeCDI(EApplicationModel, int, String[], ICDIDevice, IVersion...)}
	 * instead. This method initializes the application model <b>server</b>.
	 *
	 * @param waitAfterError_ The delay time of errors <b>after</b> doing the
	 *  		initialization stuff of all CDI servers. This is declared in 
	 *  		milliseconds if this value is greater then 60 or in seconds if it
	 *  		is lower or equal to 60.
	 * @param cdiServers_ A string array which contains all possible TINE
	 *  		pre-addresses.
	 * @param cdiDevice_ CDI device singleton.
	 * @param version_ All version instances.
	 * @see #initializeCDI(EApplicationModel, int, String[], ICDIDevice, IVersion...)
	 */
	@Deprecated
	public final static void initializeCDI(
			final int waitAfterError_,
			final String[] cdiServers_, 
			final ICDIDevice cdiDevice_,
			final IVersion... version_) {
		initializeCDI(EApplicationModel.SERVER, waitAfterError_, 2, cdiServers_, cdiDevice_, version_);
	}

	/**
	 * <p>Shows the CDI dialog box of initializing all CDI servers and if an
	 * error occurs reports the error messages for the {@code waitAfterError_}
	 * time in seconds. The delay time before the server starts is set to null,
	 * which means that their is no delay time.</p>
	 * 
	 * <p>The default time for retrying if the connection to the CDI server fails
	 * is <b>2</b> times. After this time the error is pronounced in a separate
	 * dialog box which the user must accept.</p>
	 * 
	 * <p>This will method will load all meta device information on each property.</p>
	 *
	 * @param model_ Devides the two possible application model for initializing
	 * 		the application. Internally this method will call the refresh method
	 * 		which reflects via a synchronous TINE call the meta information
	 * 		from the server, which depends on the application model (for example
	 * 		the client would be interested in the unit of a value where the
	 * 		server would be interested in the range of a value). 
	 * @param waitAfterError_ The delay time of errors <b>after</b> doing the
	 *  		initialization stuff of all CDI servers. This is declared in 
	 *  		milliseconds if this value is greater then 60 or in seconds if it
	 *  		is lower or equal to 60.
	 * @param cdiServers_ A string array which contains all possible TINE
	 *  		pre-addresses.
	 * @param cdiDevice_ CDI device singleton.
	 * @param version_ All version instances.
	 * @see #initializeCDI(EApplicationModel, int, int, String[], ICDIDevice, IVersion...)
	 */
	public final static void initializeCDI(
			EApplicationModel model_,
			final int waitAfterError_,
			final String[] cdiServers_, 
			final ICDIDevice cdiDevice_,
			final IVersion... version_) {
		initializeCDI(model_, waitAfterError_, 2, cdiServers_, cdiDevice_, version_);
	}

	/**
	 * <p>This method is <b>deprecated</b>. You should use the method {@link #initializeCDI(EApplicationModel, int, int, String[], ICDIDevice, IVersion...)}
	 * instead. This method initializes the application model <b>server</b>.</p>
	 * 
	 * <p>This will method will load all meta device information on each property.</p>
	 *
	 * @param waitAfterError_ The delay time of errors <b>after</b> doing the
	 *  		initialization stuff of all CDI servers. This is declared in 
	 *  		milliseconds if this value is greater then 60 or in seconds if it
	 *  		is lower or equal to 60.
	 * @param readBackRetries_ Number of retries if reading the CDI meta value 
	 *	 		fails. 
	 * @param cdiServers_ A string array which contains all possible TINE
	 *  		pre-addresses.
	 * @param cdiDevice_ CDI device singleton.
	 * @param version_ All version instances.
	 * @see #initializeCDI(EApplicationModel, int, int, int, String[], ICDIDevice, IVersion...)
	 */
	@Deprecated
	public final static void initializeCDI(
			final int waitAfterError_,
			final int readBackRetries_,
			final String[] cdiServers_, 
			final ICDIDevice cdiDevice_,
			final IVersion... version_) {
		initializeCDI(EApplicationModel.SERVER, 0, waitAfterError_, readBackRetries_, cdiServers_, cdiDevice_, version_);
	}

	/**
	 * <p>Shows the CDI dialog box of initializing all CDI servers and if an
	 * error occurs reports the error messages for the {@code waitAfterError_}
	 * time in seconds. The delay time before the server starts is set to null,
	 * which means that their is no delay time.</p>
	 * 
	 * <p>The default time for retrying if the connection to the CDI server fails
	 * is <b>30</b> times. After this time the error is pronounced in a separate
	 * dialog box which the user must accept.</p>
	 * 
	 * <p>This will method will load all meta device information on each property.</p>
	 *
	 * @param model_ Devides the two possible application model for initializing
	 * 		the application. Internally this method will call the refresh method
	 * 		which reflects via a synchronous TINE call the meta information
	 * 		from the server, which depends on the application model (for example
	 * 		the client would be interested in the unit of a value where the
	 * 		server would be interested in the range of a value). 
	 * @param waitAfterError_ The delay time of errors <b>after</b> doing the
	 *  	initialization stuff of all CDI servers. This is declared in 
	 *  	milliseconds if this value is greater then 60 or in seconds if it
	 *  	is lower or equal to 60.
	 * @param readBackRetries_ Number of retries if reading the CDI meta value 
	 *	 	fails. 
	 * @param cdiServers_ A string array which contains all possible TINE
	 *  	pre-addresses.
	 * @param cdiDevice_ CDI device singleton.
	 * @param version_ All version instances.
	 * @see #initializeCDI(EApplicationModel, int, int, int, String[], ICDIDevice, IVersion...)
	 */
	public final static void initializeCDI(
			final EApplicationModel model_,
			final int waitAfterError_,
			final int readBackRetries_,
			final String[] cdiServers_, 
			final ICDIDevice cdiDevice_,
			final IVersion... version_) {
		initializeCDI(model_, 0, waitAfterError_, readBackRetries_, cdiServers_, cdiDevice_, version_);
	}
	
	/**
	 * <p>This method is <b>deprecated</b>. You should use the method {@link #initializeCDI(EApplicationModel, int, int, int, String[], ICDIDevice, IVersion...)}
	 * instead. This method initializes the application model <b>server</b>.</p>
	 * 
	 * <p>This will method will load all meta device information on each property.</p>
	 *
	 * @param waitBeforeServerStarts_ Is the delay time which is executed
	 * 		<b>before</b> the server starts its initialization. It is declared in
	 * 		milliseconds. Should be greater then 1000 milliseconds. By a wait time
	 * 		of 0 the delay time is ignored.
	 * @param waitAfterError_ The delay time of errors <b>after</b> doing the
	 *  		initialization stuff of all CDI servers. This is declared in 
	 *  		milliseconds if this value is greater then 60 or in seconds if it
	 *  		is lower or equal to 60.
	 * @param readBackRetries_ Number of retries if reading the CDI meta value 
	 *	 		fails. 
	 * @param cdiServers_ A string array which contains all possible TINE
	 *  		pre-addresses.
	 * @param cdiDevice_ CDI device singleton.
	 * @param version_ All version instances.
	 * @see #initializeCDI(EApplicationModel, int, int, int, String[], ICDIDevice, IVersion...)
	 */
	@Deprecated
	public final static void initializeCDI(
			final int waitBeforeServerStarts_ , 
			int waitAfterError_,
			final int readBackRetries_,
			final String[] cdiServers_, 
			final ICDIDevice cdiDevice_,
			final IVersion... version_) {
		initializeCDI(EApplicationModel.SERVER, waitBeforeServerStarts_, waitAfterError_, readBackRetries_, cdiServers_, cdiDevice_, version_);
	}
	
	/**
	 * <p>Shows the CDI dialog box of initializing all CDI servers and if an
	 * error occurs reports the error messages for the {@code waitAfterError_}
	 * time in seconds.</p>
	 * 
	 * <p>This will method will load all meta device information on each property.</p>
	 *
	 * @param model_ Devides the two possible application model for initializing
	 * 		the application. Internally this method will call the refresh method
	 * 		which reflects via a synchronous TINE call the meta information
	 * 		from the server, which depends on the application model (for example
	 * 		the client would be interested in the unit of a value where the
	 * 		server would be interested in the range of a value). 
	 * @param waitBeforeServerStarts_ Is the delay time which is executed
	 * 		<b>before</b> the server starts its initialization. It is declared in
	 * 		milliseconds. Should be greater then 1000 milliseconds. By a wait time
	 * 		of 0 the delay time is ignored.
	 * @param waitAfterError_ The delay time of errors <b>after</b> doing the
	 *  		initialization stuff of all CDI servers. This is declared in 
	 *  		milliseconds if this value is greater then 60 or in seconds if it
	 *  		is lower or equal to 60.
	 * @param readBackRetries_ Number of retries if reading the CDI meta value 
	 *	 		fails. 
	 * @param cdiServers_ A string array which contains all possible TINE
	 *  		pre-addresses.
	 * @param cdiDevice_ CDI device singleton.
	 * @param version_ All version instances.
	 * @see #initializeCDI(EApplicationModel, int, int, int, String[], ICDIDevice, boolean, IVersion...)
	 */
	public final static void initializeCDI(
			final EApplicationModel model_,
			final int waitBeforeServerStarts_ , 
			int waitAfterError_,
			final int readBackRetries_,
			final String[] cdiServers_, 
			final ICDIDevice cdiDevice_,
			final IVersion... version_) {
		initializeCDI(model_, waitBeforeServerStarts_, waitAfterError_, readBackRetries_, cdiServers_, cdiDevice_, true, version_);
	}
	
	/**
	 * Shows the CDI dialog box of initializing all CDI servers and if an
	 * error occurs reports the error messages for the {@code waitAfterError_}
	 * time in seconds.
	 *
	 * @param model_ Devides the two possible application model for initializing
	 * 		the application. Internally this method will call the refresh method
	 * 		which reflects via a synchronous TINE call the meta information
	 * 		from the server, which depends on the application model (for example
	 * 		the client would be interested in the unit of a value where the
	 * 		server would be interested in the range of a value). 
	 * @param waitBeforeServerStarts_ Is the delay time which is executed
	 * 		<b>before</b> the server starts its initialization. It is declared in
	 * 		milliseconds. Should be greater then 1000 milliseconds. By a wait time
	 * 		of 0 the delay time is ignored.
	 * @param waitAfterError_ The delay time of errors <b>after</b> doing the
	 *  		initialization stuff of all CDI servers. This is declared in 
	 *  		milliseconds if this value is greater then 60 or in seconds if it
	 *  		is lower or equal to 60.
	 * @param readBackRetries_ Number of retries if reading the CDI meta value 
	 *	 		fails. 
	 * @param cdiServers_ A string array which contains all possible TINE
	 *  		pre-addresses.
	 * @param cdiDevice_ CDI device singleton.
	 * @param initAllDevices_ Checks if all device elements should be initialized
	 * 		(<tt>true</tt>) or not (<tt>false</tt>). Is this boolean value is
	 * 		{@code false} then only the device names and numbers are initialize. When
	 * 		it is {@code true} it will initialize all device elements by getting the device
	 * 		names, device numbers, units, precision, minimum and max values (range),
	 * 		the addresses (crate and sub-address), and the description.
	 * @param version_ All version instances.
	 */
	public final static void initializeCDI(
			final EApplicationModel model_,
			final int waitBeforeServerStarts_ , 
			int waitAfterError_,
			final int readBackRetries_,
			final String[] cdiServers_, 
			final ICDIDevice cdiDevice_,
			final boolean initAllDevices_,
			final IVersion... version_) {
		if (waitBeforeServerStarts_ != 0 && waitBeforeServerStarts_ < 1000)
			throw new IllegalArgumentException("Number of delay time before server starts is invalid: " + waitBeforeServerStarts_);
		if (!(RFDialog.getDialog( RFDialog.EDialog.CDI_DEVICES, version_) instanceof RFDialog.IDialogProgress))
			throw new IllegalArgumentException("Dialog is not an instance of IDialogProgress");
		
		final List<String> lstErrors = new ArrayList<String>();
		final int waitAfterError;
		if (waitAfterError_ == 0) waitAfterError_ = Math.abs(initDelayError); else waitAfterError_ = Math.abs(waitAfterError_);
		if (waitAfterError_ <= 60) waitAfterError = waitAfterError_; else waitAfterError = waitAfterError_ / 1000;
		
		final RFDialog.IDialogProgress dialogProgress = (RFDialog.IDialogProgress) RFDialog.getDialog( RFDialog.EDialog.CDI_DEVICES, version_);
		dialogProgress.setProgress(0);

		// ---------------------------------------------------------------------

		try {

			if (waitBeforeServerStarts_ != 0)
				movingWaitCDIServers(waitBeforeServerStarts_, dialogProgress);

			if (cdiServers_ != null) {
				int step = 80 / cdiServers_.length;
				boolean failure = false;
				
				for(int i = 0; i < cdiServers_.length && !failure; i++) {
					boolean exit = false;
					int loop = 3;
					do {
						dialogProgress.getHeader().setText("Initialisierung " + cdiServers_[i] + " ...");
						dialogProgress.setProgress(20 + (i + 0) * step);
						
						try {
				    		cdiDevice_.checkEnumElements(model_, cdiServers_[i], true, initAllDevices_);
				    		failure = false;
				    		exit = true;
				    		//throw new TineException(0, cdiServers_[i]);
						}
				    	catch(final EnumerationNotFoundException exception) {
				    		Logger.SWApp.error(ApplicationInitialization.CLAZZ + " - enumeration can not be found.", exception);
				    		if (Utility.errorMessage("Fehler bei Initialisierung",
				    				"Whrend der Initialisierung des Programms trat ein Fehler<br>" +
				    				"auf. Ein oder mehrere TINE Device-Namen stimmen nicht mit<br>" +
				    				"der internen Aufzhlung berein.", exception) != Utility.ERROR_CONTINUE)
				    			System.exit(-1);
				    	}
						catch(final TineException exception) {
							failure = true;
				    		if (--loop <= 0) {
				    			exit = true;
				    			Logger.SWApp.fatal(ApplicationInitialization.CLAZZ + " - tried serveral time to load meta info from server " + cdiServers_[i] + " which fails - exit immediately");
				    		}
				    		else {
				    			exit = false;
				    			if (!lstErrors.contains(ERROR_PRE_STRING + cdiServers_[i]))
				    				lstErrors.add(ERROR_PRE_STRING + cdiServers_[i]);
		
					    		dialogProgress.getHeader().setOpaque(true);
					    		dialogProgress.getHeader().setBackground(CColor.MST_ERROR_BACKGROUND);
					    		dialogProgress.getHeader().setForeground(CColor.MST_ERROR_FOREGROUND);
					    		dialogProgress.getHeader().setText("  FEHLER bei Initialisierung des Servers " + cdiServers_[i] + " (" + exception.getErrorCode() + ")  ");
		
					    		Thread.sleep(5000);
		
					    		dialogProgress.getHeader().setBackground(CColor.MST_DEFAULT_BACKGROUND);
					    		dialogProgress.getHeader().setForeground(CColor.MST_DEFAULT_FOREGROUND);
					    		dialogProgress.getHeader().setOpaque(false);
				    		}
				    	}
					} while(!exit);
		    		
					Thread.sleep(200);
				}

	    		// -------------------------------------------------------------

				dialogProgress.getHeader().setText("Initialisierung wurde abgearbeitet!");
				dialogProgress.setProgress(100);

				Thread.sleep(1000);

				if (lstErrors.size() != 0) {
					dialogProgress.showErrors(model_, waitAfterError, lstErrors);
					int tmp = waitAfterError;
					while(tmp-- >= 0 && dialogProgress.isActive())
						Thread.sleep(1000);
				}
			}

			dialogProgress.close();
    	}
		catch(InterruptedException exception) {
    		Utility.fatalErrorMessage("Fehler whrend Initialisierung",
    				"Whrend der Initialisierung des Programms wurde der Timer jh<br>" +
    				"unterbrochen! Versuchen Sie es noch einmal durch einen erneuten<br>" +
    				"Programmstart", null);
    		System.exit(-1);
		}
		catch(Throwable throwable) {
			Utility.fatalErrorMessage("Fehler whrend der Initialisierung",
					"Whrend der Initialisierung des Programms wurde ein<br>" +
					"unerwarteter Fehler produziert:<br><br>" + throwable.toString(), 
					throwable);
			System.exit(-1);
		}
	}

	/**
	 * Animates a displayed label by moving the citation to the left side. The
	 * animated text string is "Hell isn't other people. Hell is yourself." from
	 * Ludwig Wittgenstein.
	 *
	 * @param waitTimeBeforeServerStarts_ This is the waiting time for one
	 * 		CDI server in milliseconds.
	 * @param dialogProgress_ Is the LibMST information dialog box.
	 * @see #movingWaitCDIServers(int, de.desy.mst.libs.accelerator.view.RFDialog.IDialogProgress, String)
	 */
	private static final void movingWaitCDIServers(
			final int waitTimeBeforeServerStarts_,
			final RFDialog.IDialogProgress dialogProgress_)
	{
		movingWaitCDIServers(waitTimeBeforeServerStarts_,
				dialogProgress_,
				"Hell isn't other people. Hell is yourself. " +
				"(L. Wittgenstein)             ");
	}

	/**
	 * Animates a displayed label by moving message to the left side.
	 *
	 * @param waitTimeBeforeServerStarts_ This is the waiting time for one
	 * 		CDI server in milliseconds.
	 * @param dialogProgress_ Is the LibMST information dialog box.
	 * @param message_ Scrolling text to be displayed.
	 */
	private static final void movingWaitCDIServers(
			final int waitTimeBeforeServerStarts_,
			final RFDialog.IDialogProgress dialogProgress_,
			final String message_)
	{
		final String WOCS = message_;

		final int delay;
		if (waitTimeBeforeServerStarts_ >= 2000)
			delay = waitTimeBeforeServerStarts_;
		else
			delay = 2000;
		
		final int step = delay / 20;
		dialogProgress_.getHeader().setText(WOCS);
		Timer timer = new Timer(100, new ActionListener() {

			private int count = 0;

			public void actionPerformed(final ActionEvent event_)
			{
				dialogProgress_.getHeader().setText(
						WOCS.substring((count % WOCS.length()), WOCS.length())
						+
						WOCS.substring(0, (count % WOCS.length()))
						);
				count++;
			}
		});
		timer.start();
		
		int i = 0;
		do
		{
			dialogProgress_.setProgress(step / 1000 * i);
			Utility.sleep(step);
			i++;
		}
		while(step * i < delay);
		
		timer.stop();
	}
}
/* end of file */