/*
 *
 * File: IEDataMetaValue.java
 * Date: Jan 15, 2008; 10:56:47 AM
 */
package de.desy.mst.libs.accelerator;

/**
 * <p>The enumeration interface of {@code IEDataMetaValue} is a part of the 
 * {@link IEData} interface, which returns a list of <b>short</b>
 * which should first be send to the server before getting the data. The
 * method for these values is {@link IEDataMetaValue#preSendValues()}.</p> 
 * 
 * <p>This enumerations is mostly used for DESY 2 data enumerations,
 * which needs to send and receive constants to the server before it will return 
 * the corresponding values. For example, the TINE property of 
 * <b>{@code sel_sps_block}</b> will send the constant integers of {@code 51}
 * and {@code 26} as a data block number and data word before returning
 * the status of the DESY 2 RF transmitter.<br><br>
 * In other words: 
 * <ul><li>Output: {@code out(0): 51}, {@code out(1): 26} and</li> 
 * <li>{@code in(0): <status>} (a 16-bit integer value)</li></ul></p>
 *  
 * <p>The output values can be read by {@link IEDataMetaValue#preSendValues()}
 * and the input by {@link IEDataMetaValue#preRecvValues()}.</p>
 *
 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
 * @version $Id: IEDataMetaValue.java 12066 2008-05-29 08:35:57Z walla $
 */
public interface IEDataMetaValue
{
	
	/**
	 * <p>This method returns the send meta values as a list of <b>short</b> 
	 * ones. I used this method for TINE property names, which needs to send 
	 * at first values to the server before receiving the expected values. For
	 * getting the received values you should look at the method of 
	 * {@link IEDataMetaValue#preRecvValues()}</p>
	 * <p>The send values are defined per enumeration value in the concrete
	 * enumeration definition and so fore there is no method needed to set the 
	 * value at a latter time with a method (for example 
	 * {@code putPreSendValues(short[]):void}).</p>  
	 *  
	 * @return Returns all pre send meta values.
	 */
	public abstract short[] preSendValues();
	
	/**
	 * <p>This method returns the received meta data as an array of short 
	 * values. The returned meta data is in its length equal or lower then the 
	 * send meta values of the (see {@link IEDataMetaValue#preSendValues()}),
	 * and each element of these values must be equal to the send values from
	 * its position {@code x}.</p>
	 * <p>For example, the send meta data returns the values of <b>51</b>, 
	 * <b>28</b>, and <b>29</b>, where the first value defines the data block 
	 * you wanted to read and the last defines the data words. The received 
	 * values you get by this method only the data words, here: <b>28</b> and 
	 * <b>29</b>.</p>    
	 *  
	 * @return Returns the received meta data as an array of short values.
	 */
	public abstract short[] preRecvValues();
}
/* end of file */