/*
 *
 * File: IEDeviceGroup.java
 * Date: Mar 27, 2007; 2:32:44 PM
 */
package de.desy.mst.libs.accelerator;

import java.util.List;

import de.desy.tine.types.TCompoundDataObject;


/**
 * <p>These enumeration interface represents a group of devices. So, each entry
 * in this interface can be one to many software devices from the real hardware.
 * </p>
 * <p>This is only a dummy interface for a group of <b>device names</b>. It is 
 * a common interface and is inherited by an enumeration in the concrete project.
 * This interface is entailed by the concrete enumeration interfaces of 
 * {@link IEDevice}.</p>
 * 
 * <p>The following properties are returned by the inherited enumeration:</p>
 * 
 * <ul>
 * <li>cardinality or size is returned - it is always greater then zero</li>
 * <li>delay - it waits a time, here realized in millisecond before querying 
 * again</li>
 * <li>dataType - is returned as the TINE data format - this is an value of the
 * TINE class in {@code TFormat}</li>
 * <li>range - is the range of values to be returned and be best explained as a
 * from - to values.</li>
 * <li>range from - is the value from which the range starts.</li>
 * <li>range to - is the value from which the range ends.</li>
 * <li>getDataComponent - is a map of data components this device group belongs 
 * to.</li>
 * <li>element - is the device element. This is important for templates which
 * has a diplocardiac names. The first element can be get by the method of
 * {@link #getDataComponent()} and the second with the method of 
 * {@link #element()}.</li> 
 * </ul>
 *  
 * @author <a href="marcus.walla(at)desy.de">walla</a>
 * @version $Id: IEDeviceGroup.java 22468 2009-03-05 16:42:12Z walla $
 */
public interface IEDeviceGroup extends IEDevice
{
	/**
	 * Returns the element of this device group. This can also return a <tt>null</tt>
	 * value.
	 * 
	 * @return The value of the corresponding element is returned.
	 */
	public abstract IEDeviceElement element();

	/**
	 * Returns the cardinality of an object of this enumeration. This is the
	 * cardinality of the corresponding device groups but <b>NOT</b> of the
	 * multiplied elements per device group.
	 * 
	 * @return The cardinality is returned.
	 */
	public abstract int cardinality();

	/**
	 * Returns a list of all device names from here.
	 * 
	 *  @return Returns the current device names as a list.
	 */
	public abstract List<IEData> getDataComponent();
	
	/**
	 * Returns an array of TINE data objects. which has the super class of
	 * {@code TCompoundDataObject}. 
	 * 
	 * @return Returns an array of TINE data objects. This element should never
	 * 		be {@code null} and full filled with the corresponding element.
	 */
	public abstract TCompoundDataObject[] getDataObjects();

	/**
	 * If the TINE address from the data element binds to a CDI device this 
	 * value allows to get the value calibrated (returns a {@code true}) or not 
	 * (returns a {@code false}). This has an direct influence on the TINE
	 * address and how the access was created. Which means that reading/writing 
	 * values is done via {@code RECV} or {@code RECV.CLBR} respectively 
	 * {@code SEND} and {@code SEND.CLBR}.
  	 * 
	 * @return Returns if the data element is calibrated ({@code true}) or not
	 * 		({@code false}).
	 */
	public abstract boolean isCalibrated();
	
	/**
	 * When this device group is bind to a list of CDI devices then a 
	 * {@code true} is returned otherwise a {@code false}.
	 * 
	 * @return Returns if the device group has a list of CDI devices or not.
	 */
	public abstract boolean isCDI();
}
/* end of file */