/*
 *
 * File: Utility.java
 * Date: Nov 16, 2006; 3:29:57 PM
 */
package de.desy.mst.libs.accelerator;

import java.awt.Component;
import java.awt.Container;
import java.awt.Desktop;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.net.URI;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import javax.swing.GroupLayout;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.LayoutStyle.ComponentPlacement;
import javax.swing.SwingConstants;
import javax.swing.Timer;
import javax.swing.WindowConstants;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableCellRenderer;

import de.desy.acop.transport.AcopTransport;
import de.desy.acop.transport.AcopTransportRequest;
import de.desy.mst.libs.accelerator.exception.TineException;
import de.desy.mst.libs.accelerator.model.EAccess;
import de.desy.mst.libs.accelerator.model.ILinkData;
import de.desy.mst.libs.accelerator.view.CColor;
import de.desy.mst.libs.accelerator.view.ComponentSubject;
import de.desy.mst.libs.accelerator.view.IViewComponent;
import de.desy.mst.libs.framework.mstapp.Main;
import de.desy.mst.libs.framework.mstapp.logging.LogLvl;
import de.desy.mst.libs.framework.mstapp.logging.Logger;
import de.desy.mst.libs.framework.mstapp.logging.LoggerManager;
import de.desy.mst.libs.ser.tclientlib.ITineAccessor;
import de.desy.tine.addrUtils.FECInfo;
import de.desy.tine.addrUtils.TSrvEntry;
import de.desy.tine.client.TLink;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TMode;
import de.desy.tine.queryUtils.TQuery;

/**
 * Different static methods are defined in this {@code Utility} class which can
 * be interesting for miscellaneous variances of a concrete project. Currently
 * in this class are declared the following fields and static methods:
 *
 * <ol>
 * <li>An enumeration of the dialog results (see {@link Utility.DIALOG_RESULT}).</li>
 * <li>Difference static fields about the developer and version information (see for example at {@link Utility#APP_APPLICATION_NAME} etc.).</li>
 * <li>Different static int-fields about the return value of {@link Utility#errorMessage(Component,String,String,Throwable)}, {@link Utility#errorMessage(String, String, Throwable), {@link Utility#fatalErrorMessage()}, {@link Utility#fatalErrorMessage(Component)}, {@link Utility#fatalErrorMessage(String, String, Throwable)}, {@link Utility#fatalErrorMessage(Component, String, String, Throwable)}, {@link Utility#warningMessage(String, String, Throwable)}, and {@link #warningMessage(Component, String, String, Throwable)}.</li>
 * <li>Methods for reporting a failure of the application (for example you should have a look at {@link Utility#errorMessage(String, String, Throwable)} or {@link Utility#fatalErrorMessage(String, String, Throwable)}). Moreover there also exists dialog boxes which are created via the class methods of {@link #createTINEErrorDialog(Container, String, int)} or {@link #createTINEErrorDialog(Container, String, String, int)}</li>
 * <li>Send a mail with a {@code Throwable} instance, see {@link Utility#sendMailTo(Throwable, StringBuilder)}.</li>
 * <li>Throw the exception and print a logging information, see {@link Utility#throwException(Throwable, String, boolean)} and {@link Utility#throwWarning(Throwable, String, boolean)}. There also exist some logging methods called {@link #logTINEAttachedLink(String, AcopTransport, int)}, {@link #logTINEExecuteLink(String, AcopTransport, int)}, {@link #logTINEErrorLink(String, AcopTransport, int)}, and {@link #logTINEErrorLink(String, AcopTransport, String)}.</li>
 * <li>Enable/disable the component and all of its child components, see {@link Utility#setEnableComponent(boolean, Container)}.</li>
 * <li>Set a component and all of its child component to visible/invisible, see {@link Utility#setVisibleComponent(boolean, Container)}.</li>
 * <li>The current thread will sleep a delay time, see {@link Utility#sleep(int)}.</li>
 * <li>Bit operation, see {@link Utility#toBitCode(int)} and {@link Utility#isValidBit(int)}.</li>
 * <li>Debug output, see {@link Utility#debugAbsoluteTime(long, boolean)}, {@link Utility#debugCurrentTime(long, boolean)}, {@link Utility#debugCurrentTime(long, String, boolean)}, and {@link Utility#debugPrintAllComponents()}.</li>
 * <li>Normalizes a string value by transforming special characters into a percent-value in a HTML-styled (or better XML-styled) variant, see {@link Utility#getNormalizedString(String)}.</li>
 * <li>Different methods for the generic data elements, see {@link Utility#getAddress(IEData)}, {@link Utility#getTinename(IEData)}, {@link Utility#containsRange(IEData, IEData, Collection)}, </li>
 * <li>A short utility method of collections which are aggregated in one list, see {@link Utility#getList(Collection...)}.</li>
 * <li>Moving a Swing label with a FPS (frames per second) int-value and a delay time, see {@link Utility#movingLabel(int, String, JLabel)}.</li>
 * <li>Allows to exit the application with the static methods of {@link #exit(String, Throwable, ILinkData, IEDeviceGroup[])} or {@link #shutdown(int, ILinkData, IEDeviceGroup...)}.</li>
 * <li>TINE logging information. For a first access point you can have look at the {@link Utility#logTINELink(LogLvl, String, String, TLink...)} and {@link Utility#logTINELink(LogLvl, String, String, AcopTransport, int)}.</li>
 * <li>Calculates an array of double values with a different step size. This method is especially created for build the major steps of a slider.</li>
 * </ol>
 *
 * @author <a href="marcus.walla(at)desy.de">walla</a>
 * @version $Id: Utility.java 55475 2013-06-19 15:46:46Z walla $
 */
public final class Utility
{
	/** Class name. */
	public static final String CLAZZ = Utility.class.getSimpleName();
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	/**
	 * The result of the displayed modal dialog box. This is currently returned
	 * by the static methods of {@link Utility#fatalErrorMessage()},
	 * {@link Utility#fatalErrorMessage(String, String, Throwable)},
	 * {@link Utility#errorMessage(String, String, Throwable)}, or the link
	 * information dialog box called {@link #showLinkInfo(TLink...)}.
	 *
	 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
	 * @version $Id: Utility.java 55475 2013-06-19 15:46:46Z walla $
	 */
	public enum DIALOG_RESULT { OK, CONTINUE, EXIT, SEND_EMAIL, NO,
		CANCEL};

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	/** This is the folder for storing the exceptions on the local drive. */
	public static final String JAVA_APPLICATION_PATH = "Java";

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	/** Is the reflected  'unknown' description of the reported email message. */
	public static final String UNKNOWN = "unknown";

	/** Is the reflected  'user name' description of the reported email message. */
	public static final String USER_NAME = "user.name";

	/** Is the reflected  'user home directory' description of the reported email message. */
	public static final String USER_HOME = "user.home";

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	/** Is the reflected  'application name' description of the reported email message. */
	public static final String APP_APPLICATION_NAME = "application.name";

	/** Is the reflected  'application version' description of the reported email message. */
	public static final String APP_VERSION = "application.version";
	
	/** Is the reflected application versions release date. */
	public static final String APP_RELEASE_DATE = "application.releaseDate";

	/** Is the reflected  application developer in its full name. */
	public static final String APP_DEVELOPER = "application.developer";

	/** Is the reflected  application developer in its short name, i.e. "walla". */
	public static final String APP_AUTHOR = "application.author";

	/** Is the phone number of the developer, i.e. "2628". */
	public static final String APP_PHONE = "application.phone";

	/** Is the reflected  building of the developer, i.e. "30" or "Arztpraxis". */
	public static final String APP_BUILDING = "application.building";

	/** Is the reflected  building of the developer, i.e. "517". */
	public static final String APP_ROOM = "application.room";

	/** Is the reflected  group description of the developer, i.e. "MCS1" */
	public static final String APP_GROUP = "application.group";

	/** Is the reflected  application e-mail address, i.e. "marcus.walla(at)desy.de". */
	public static final String APP_EMAIL = "application.email";

	/** Is the reflected  application mailing list. */
	public static final String APP_MAILINGLIST = "application.mailinglist";

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	
	/** This is the result of option pane on pressing send email button. */
	public static final int OPTIONPANE_SEND_EMAIL = 1;

	/** This is the result of option pane on pressing OK button. */
	public static final int OPTIONPANE_SEND_OK = 0;

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	/** Is the highest number of bits in a short value. */
	public static final int HIGHEST_BIT_VALUE_SHORT = 16;

	/** Is the highest number of bits in an integer value. */
	public static final int HIGHEST_BIT_VALUE_INT = 32;

	/** Is the highest number of bits in a long value. */
	public static final int HIGHEST_BIT_VALUE_LONG = 64;

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	/** Continue button index of the function
	 * {@link #warningMessage(String, String, Throwable)} or
	 * {@link #warningMessage(Component, String, String, Throwable)}.*/
	public static final int WARNING_CONTINUE = 0;

	/** Exit button index of the function
	 * {@link #warningMessage(String, String, Throwable)} or
	 * {@link #warningMessage(Component, String, String, Throwable)}.*/
	public static final int WARNING_EXIT = 1;

	/** Send E-Mail message and exit button index of the function
	 * {@link #warningMessage(String, String, Throwable)} or
	 * {@link #warningMessage(Component, String, String, Throwable)}.*/
	public static final int WARNING_SEND_EMAIL = 2;

	/** Continue button index of the function
	 * {@link #errorMessage(String, String, Throwable)} or
	 * {@link #errorMessage(Component, String, String, Throwable)}.*/
	public static final int ERROR_CONTINUE = 0;

	/** Exit button index of the function
	 * {@link #errorMessage(String, String, Throwable)} or
	 * {@link #errorMessage(Component, String, String, Throwable)}.*/
	public static final int ERROR_EXIT = 1;

	/** Send E-Mail message and exit button index of the function
	 * {@link #errorMessage(String, String, Throwable)} or
	 * {@link #errorMessage(Component, String, String, Throwable)}.*/
	public static final int ERROR_SEND_EMAIL = 2;

	/** Exit button index of the function
	 * {@link #fatalErrorMessage()}, {@link #fatalErrorMessage(Component)},
	 * {@link #fatalErrorMessage(String, String, Throwable)},
	 * or {@link #fatalErrorMessage(Component, String, String, Throwable)}.*/
	public static final int FATAL_EXIT = 0;

	/** Send E-Mail and exit button index of the function
	 * {@link #fatalErrorMessage()}, {@link #fatalErrorMessage(Component)},
	 * {@link #fatalErrorMessage(String, String, Throwable)},
	 * or {@link #fatalErrorMessage(Component, String, String, Throwable)}.*/
	public static final int FATAL_SEND_EMAIL = 1;
	
	// -------------------------------------------------------------------------
	
	/**
	 * Returns a list of the stack trace elements in a human readable way. 
	 * The number of elements which are included in the string value depends
	 * on the current stack trace element list and can reach up to 10 
	 * elements.
	 * 
	 * @param stackTraces_ Stack trace elements.
	 * @return Returns a human readable stack trace elements.
	 */
	public final static String getCallingMethod(final StackTraceElement[] stackTraces_)
	{
		final StringBuilder builder = new StringBuilder();
		int lines = 1;
		
		if (stackTraces_ == null)
			builder.append(CLAZZ + " - stack trace instances is null\n");
		else if (stackTraces_.length <= 1) 
			// remember that the first entry is of this class!
			builder.append(CLAZZ + " - stack trace instances is empty\n");
		else
		{
			do
			{
				builder.append('\t')
					.append(stackTraces_[lines].getClassName()).append('.')
					.append(stackTraces_[lines].getMethodName()).append(" on line ")
					.append(stackTraces_[lines].getLineNumber()).append('\n');
				lines++;
			}
			while (lines < 10 && lines < stackTraces_.length);

			if (lines < stackTraces_.length)
				builder.append("...\n");
		}
		
		return builder.toString();
	}
	
	/**
	 * Returns a list of the stack trace elements in a human readable way. 
	 * The number of elements which are included in the string value depends
	 * on the current stack trace element list and can reach up to 10 
	 * elements.
	 * 
	 * @param thread_ The calling thread.
	 * @return Returns a human readable stack trace elements.
	 */
	public final static String getCallingMethod(final Thread thread_)
	{
		if (thread_ != null)
			return getCallingMethod(thread_.getStackTrace());
		else
			return (CLAZZ + " - thread instance is null\n");
	}
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	
	/**
	 * Returns the length of the character array until a null-terminated character. 
	 * It searches for the null-character by look at each character from the
	 * beginning and reflects if the integer value is zero. If it is not it will
	 * proceed otherwise it will return the current position.
	 * 
	 * @param arrayOfChars_ Character array.
	 * @return Returns the length of the character array by searching the
	 * 		first null-terminate character.
	 */
	public final static int getLengthOfChar(final char[] arrayOfChars_)
	{
		if (arrayOfChars_ == null || arrayOfChars_.length == 0)
			return 0;
		for (int i = 0; i < arrayOfChars_.length; i++)
			if (arrayOfChars_[i] == 0)
				return i;
		return arrayOfChars_.length;
	}

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	
	/**
     * <p>Returns the bit values of <b>short</b>. The length of the
     * delivered value can range from 1st to 16th bit and is declared in the constant
     * field called {@link #HIGHTEST_BIT_VALUE_SHORT}. It interprets the
     * {@code short}-value as signed.</p>
     * <p><b>ATTENTION:</b> You should use the new method of {@link #toBit(short)}
     * instead of this one.</p>
     *
     * @param value_ Short value to calculated.
     * @return A string is returned, each nibble are separated with a space 
     * 		character.
     * @deprecated Use method {@link #toBit(short)} instead.
     */
	@Deprecated
    public static final String toBitCode(final short value_)
    {
    	return toBit(value_);
    }
    
	/**
     * Returns the bit values of a <b>short</b>. The length of the
     * delivered value can range from 1st to 16th and is declared in the constant
     * field called {@link #HIGHTEST_BIT_VALUE_SHORT}. It interprets the
     * {@code short}-value as signed.
     *
     * @param value_ Short value to calculated.
     * @return A string is returned, each nibble are separated with a space 
     * 		character.
     */
    public static final String toBit(final short value_)
    {
    	int currentValue = (value_ & ((1 << HIGHEST_BIT_VALUE_SHORT) - 1));
    	StringBuilder buffer = new StringBuilder();
    	for(int i = HIGHEST_BIT_VALUE_SHORT - 1; i >= 0; i--)
    	{
    		if((currentValue  & (1 << i)) == (1 << i))
    			buffer.append("1");
    		else
    			buffer.append("0");
    		if ((i % 4) == 0 && i != 0)
    			buffer.append(" ");
    	}
    	return buffer.toString();
    }
    
	/**
     * Returns the bit values of an <b>integer</b>. The length is always
     * 32 bits. It interprets the {@code int}-value as signed or the 
     * {@code short}-value as unsigned. Remember that the value should always 
     * be casted to an integer value!
     *
     * @param value_ Integer value to calculated.
     * @return A string is returned, each nibble are separated with a space 
     * 		character.
     */
    public static final String toBit(final int value_)
    {
    	int currentValue = (value_ & -1);
    	StringBuilder buffer = new StringBuilder();
    	for(int i = 31; i >= 0; i--)
    	{
    		if((currentValue  & (1 << i)) == (1 << i))
    			buffer.append("1");
    		else
    			buffer.append("0");
    		if ((i % 4) == 0 && i != 0)
    			buffer.append(" ");
    	}
    	return buffer.toString();
    }
    
	/**
     * Returns the bit values as a string of <b>long</b>. The length is always
     * 64 bits. It interprets the {@code long}-value as signed and the 
     * {@code short}- or {@code int}-value as unsigned. Remember that the value 
     * should always casted to a long value!
     *
     * @param value_ The long value to calculated.
     * @return A string is returned, each nibble are separated with a space 
     * 		character.
     */
    public static final String toBit(final long value_)
    {
    	long currentValue = (value_ & -1l);
    	StringBuilder buffer = new StringBuilder();
    	for(int i = 63; i >= 0; i--)
    	{
    		if((currentValue  & (1l << i)) == (1l << i))
    			buffer.append("1");
    		else
    			buffer.append("0");
    		if ((i % 4) == 0 && i != 0)
    			buffer.append(" ");
    	}
    	return buffer.toString();
    }

    /**
	 * Checks if only one bit is set in a {@code long} value {@code bit_}. It
	 * does this by querying from the first bit position until one bit is set to 1. 
	 * If no bit is set to 1 a {@code false} is returned. If it is set it controls in a
	 * second query if more bits are set. If yes it returns with a {@code false}
	 * otherwise {@code true}.
	 *
	 * @param bit_ Is the bit value as a {@code long}. Remember that this value 
	 * 		is performed with negative values if the highest bit is set.
	 * @return Returns a {@code true} if in the long value only one bit is 
	 * 		set. Otherwise a {@code false} is returned. 
	 */
	public static final boolean isValidBit(final long bit_)
	{
		int i = 0;

		do
		{
 			if ((1<<i) == (bit_ & 1<<i))
			{
				i++;
				while(i < 64)
				{
					if ((1<<i) == (bit_ & (long) 1<<i))
						return false;
					i++;
				}
				return true;
			}
			i++;
		}
		while(i < 64);

		return false;
	}

	/**
	 * Checks if only one bit is set in an {@code int} value. It
	 * does this by querying until the first bit is set to 1. If nothing can be found
	 * this method response with {@code false}. If it is set it controls in a
	 * second query if more bits are set. If yes it returns with a {@code false}
	 * otherwise {@code true}.
	 *
	 * @param bit_ Is the bit value as a {@code int}. Remember that this value 
	 * 		is performed with negative values if the highest bit is set.
	 * @return Returns a {@code true} if in the long value only one bit is 
	 * 		set. Otherwise a {@code false} is returned. 
	 */
	public static final boolean isValidBit(final int bit_) {
		return isValidBit((long) bit_);
	}

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	/**
	 * Compares an epsilon domain around the compare value and the current one:
	 * <pre>
	 * &#x2223;current&#x2223; &#x2266; &#x2223;compare&#x2223; + &#x03B5;
	 * </pre>
	 * 
	 * @param current Current value.
	 * @param compare Compare value.
	 * @param epsilon Epsilon domain. Will be changed to a positive value.
	 * @return Returns a boolean value if the compared value is inside the 
	 * 		domain ({@code true}) or not ({@code false}).
	 */
	public final static boolean epsilon(double current, final double compare, final double epsilon) {
		final double tmpEpsilon = Math.abs(epsilon);
		return ((compare - tmpEpsilon) <= current && (compare + tmpEpsilon) >= current);
	}
	
	/**
	 * Compares an epsilon domain around the compare value and the current one:
	 * 
	 * <pre>
	 * &#x2223;current&#x2223; &#x2266; &#x2223;compare&#x2223; + &#x03B5;
	 * </pre>
	 * 
	 * @param current Current value.
	 * @param compare Compare value.
	 * @param epsilon Epsilon domain. Must be greater then zero.
	 * @see #epsilon(double, double, double)
	 */
	public final static boolean epsilon(float current, float compare, float epsilon) {
		return epsilon((double) current, (double) compare, (double) epsilon);
	}

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	/**
	 * Creates an array of doubles which starts at min and ends at max. Between 
	 * this end points 10 steps are created. The returned array always has a
	 * length of 11 elements (1 at the beginning and 10 incrementing step elements).
	 * 
	 * @param min First element.
	 * @param max Last element.
	 * @return Returns the calculated ticks.
	 */
	public static final double[] createTicksStructuring(final double min, final double max) {
		return createTicksStructuring(min, max, 10);
	}

	/**
	 * Creates an array of double values beginning at min and reaches to max. 
	 * Between the end points it creates ticks. Each step increments the before
	 * defined value with {@code (maximum - minimum)/steps}. The returned array 
	 * always has a length of {@code steps + 1} elements.
	 * 
	 * @param min First element.
	 * @param max Last element.
	 * @param steps Steps in the range.
	 * @return Returns a double array with a length of {@code step + 1} elements.
	 */
	public static final double[] createTicksStructuring(final double min, final double max, final int steps) {
		final int tmpSteps = steps < 1 ? 1 : steps;
		final double[] ticks = new double[tmpSteps + 1];
		final double increment = (max - min) / tmpSteps;
		
		ticks[0] = min;
		for(int i = 1; i < ticks.length; i++)
			ticks[i] = ticks[i - 1] + increment;
		ticks[ticks.length - 1] = max; // only for safety

		return ticks;
	}
	
	/**
	 * <p>Calculates the major ticks. These ticks reach from the delivered minimum 
	 * up to the maximum tick value, where it is separated from the first value with the
	 * {@code stepBetweenValues) interval between two major steps. The tolerance
	 * is in this case stated to 0.9.</p>
	 * 
	 * @param min Minimum value.
	 * @param max Maximum value.
	 * @param stepBetweenValues Quantum between two major ticks.
	 * @return Returns a double array of ticks.
	 */
	public static final double[] createTicksStep(final double min, final double max, final double stepBetweenValues) {
		final double tmpStepBetweenValues = (stepBetweenValues <= 0) ? 1.0 : stepBetweenValues; 
		final double minimum = Math.min(min, max);
		final double maximum = Math.max(min, max);
		
		int counter = 0;
		do {
			counter += 1;
		}
		while((minimum + tmpStepBetweenValues * counter) < maximum);

		final double[] result;
		if ((minimum + stepBetweenValues * (counter - 1)) + stepBetweenValues * 0.9 > maximum ) {
			result = new double[counter];
			for(int i = 0; i < counter; i++)
				result[i] = minimum + stepBetweenValues * i;
			result[counter - 1] = maximum;
		}
		else {
			result = new double[counter + 1];
			for(int i = 0; i < counter + 1; i++)
				result[i] = minimum + stepBetweenValues * i;
			result[counter] = maximum;
		}
		
		return result;
	}

	/**
	 * Returns an array of double values which starts at the minimum value and
	 * ends at the maximum value. Between this end points this method creates
	 * 9 sub ticks, so that the returned array always has a length of 11 
	 * elements.
	 * 
	 * @param min The first element.
	 * @param max The last element.
	 * @return Returns a double array with a length of 11 elements.
	 * @deprecated Use method {@link #createTicksStructuring(double, double} instead.
	 */
	@Deprecated
	public static final double[] getMajorTicks(final double min, final double max) {
		return createTicksStructuring(min, max, 10);
	}
	
	/**
	 * Returns an array of double values which starts at the minimum value and
	 * ends at the maximum value. Between this end points this method creates
	 * sub ticks, which has a differences between two sub ticks of 
	 * {@code (maximum - minimum)/step}, so that the returned array always has a
	 * length of {@code steps + 1} elements.
	 * 
	 * @param minimum_ The first element.
	 * @param maximum_ The last element.
	 * @param steps_ Steps in the range.
	 * @return Returns a double array with a length of {@code step + 1} 
	 * 		elements.
	 * @deprecated Use method {@link #createTicksStructuring(double, double, double) instead.
	 */
	@Deprecated
	public static final double[] getMajorTicks(final double min, final double max, final int steps) {
		return createTicksStructuring(min, max, steps);
	}

	/**
	 * <p>Calculates the major ticks. These ticks reach from the delivered minimum 
	 * up to the maximum tick value, where it is separated from the first value with the
	 * {@code stepBetweenValues) interval between two major steps. The tolerance
	 * is in this case stated to 0.9.</p>
	 * 
	 * @param min Minimum value.
	 * @param max Maximum value.
	 * @param stepBetweenValues Quantum between two major ticks.
	 * @return Returns a list of the major ticks.
	 * @deprecated Use method {@link #createTicksStep(double, double, double)} instead.
	 */
	@Deprecated
	public static final double[] getMajorTicks(final double min, final double max, final double stepBetweenValues) {
		return createTicksStep(min, max, stepBetweenValues);
	}
	
	/**
	 * <p>
	 * Calculates the major ticks. These ticks reach from the delivered min 
	 * up to max, where the range starts from the first value with
	 * {@code stepBetweenValues) interval between two major steps. The double 
	 * value of {@code tolerance} indicates a difference between the next to last of the
	 * last major ticks and describes if the next major tick will be included
	 * in the returned collection or not. The higher you choose the tolerance 
	 * the lesser will is the next major tick will be included.
	 * </p>
	 * <p>
	 * Min and max values can be exchange if necessary.
	 * </p>
	 * 
	 * @param min Minimum value.
	 * @param max Maximum value.
	 * @param stepBetweenValues Quantum between two major ticks.
	 * @param tolerance Indicates till which differences the next to last major
	 * 		tick will be included in the returned list. Can reach from 0.5 up to
	 * 		1.0.
	 * @return Returns a list of the major ticks.
	 * @deprecated tolerance is not used here. Instead use the method {@link #getMajorTicks(double, double, double)}. 
	 */
	@Deprecated
	public static final double[] getMajorTicks(final double min, final double max, final double stepBetweenValues, final double tolerance) {
		return getMajorTicks(min, max, stepBetweenValues);
	}

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	
	/** Message: Missing data. */
	public final static String MESSAGE_MISSING_DATA = "* MISSING DATA *";
	/** Message: Missing TINE address. */
	public final static String MESSAGE_MISSING_TINE_ADDRESS = "* MISSING TINE ADDRESS *";
	/** Message: Missing ACOP transport. */
	public final static String MESSAGE_MISSING_ACOP_TRANSPORT = "* MISSING ACOP TRANSPORT *";
	/** Message: Missing TINE link. */
	public final static String MESSAGE_MISSING_TINE_LINK = "* MISSING TINE LINK *";
	/** Message: Missing device group. */
	public final static String MESSAGE_MISSING_DEVICE_GROUP = "* MISSING DEVICE GROUP *";
	/** Message: Host name */
	public final static String MESSAGE_HOSTNAME = "HOSTNAME=";
	
	/** <p>Map of all server entries from TINE. The key value is the TINE context
	 * plus the device server name (each separated with a slash: 
	 * /&lt;context&gt;/&lt;device server&gt;) and can be retrieved 
	 * from the enumeration interface {@link IEEnum#getName(EName)} where the 
	 * parameter is {@link EName#PRE_TINEADDRESS}.</p> 
	 */
	public static final Map<String, TSrvEntry> SERVERS = 
		new HashMap<String, TSrvEntry>();
	
	/** <p>Map of all FEC information entries from TINE. The key value is the
	 * TINE context plus the device server name (each separated with a slash:
	 * /&lt;context&gt;/&lt;device server&gt;) and can be retrieved from the
	 * enumeration interface (@link {@link IEEnum#getName(EName)} where the
	 * parameter is {@link EName#PRE_TINEADDRESS}.</p>
	 * 
	 */
	public static final Map<String, FECInfo> FEC_INFO =
		new HashMap<String, FECInfo>();
	
	/**
	 * Returns the components crate, sub-address and address of the data device
	 * as a string value if it exist. This is probably done for all CDI devices.
	 * Otherwise, if the data type is not a CDI one it will return the pre TINE
	 * address and the property name. The output is in HTML style.
	 *
	 * @param data_ Is the enumeration interface of the corresponding device
	 * 		data component.
	 * @return A string value of the corresponding SEDAC or TINE address path.
	 * @see #getAddress(IEData, boolean, boolean)
	 */
	public static final String getAddress(final IEData data_)
	{
		return getAddress(data_, true);
	}
	
	/**
	 * Returns the components crate, sub-address and address of the data device
	 * as a string value if it exist. This is probably done for all CDI devices.
	 * Otherwise, if the data type is not a CDI one it will return the pre TINE
	 * address and the property name. The output is in HTML style.
	 *
	 * @param data_ Is the enumeration interface of the corresponding device
	 * 		data component.
	 * @param isHostname_ Displayes the host name if this parameters is set to	
	 * 		{@code true}.
	 * @return A string value of the corresponding SEDAC or TINE address path.
	 * @see #getAddress(IEData, boolean, boolean)
	 */
	public static final String getAddress(final IEData data_, final boolean isHostname_)
	{
		return getAddress(data_, true, true);
	}
	
	private static final Map<String, Integer> tinePreAddresses = new HashMap<String, Integer>();
	
	/**
	 * Returns the components crate, sub-address and address of the data device
	 * as a string value if it exist. This is probably done for all CDI devices.
	 * Otherwise, if the data type is not a CDI one it will return the pre TINE
	 * address and the property name. The output can be elected with HTML style 
	 * or not.
	 *
	 * @param data_ Is the enumeration interface of the corresponding device
	 * 		data component.
	 * @param isHostname_ Displayes the host name if this parameters is set to	
	 * 		{@code true}.
	 * @param isHTMLStyle_ Switches between the HTML flavored implementation or
	 * 		with a normal carriage return.
	 * @return A string value of the corresponding SEDAC or TINE address path.
	 */
	public final static String getAddress(final IEData data_, final boolean isHostname_, final boolean isHTMLStyle_)
	{
		if (data_ == null)
			return MESSAGE_MISSING_DATA;

		final String tinepath = data_.getName(EName.PRE_TINEADDRESS);
		if (tinepath == null || tinepath.isEmpty())
			return MESSAGE_MISSING_TINE_ADDRESS;

		final StringBuilder sb = new StringBuilder();

		if (!data_.isCDI())
		{
			final String device = data_.getName(EName.DEVICENAME);
			final String property = data_.getName(EName.PROPERTY);

			sb.append("TINE: ").append(tinepath);
			
			if (!Utility.tinePreAddresses.containsKey(tinepath))
				Utility.tinePreAddresses.put(tinepath, fetchDelimiter(tinepath));
			
			if (Utility.tinePreAddresses.get(tinepath) == 2 && device != null && !device.isEmpty())
				sb.append('/').append(device);
			
			if (property != null && !property.isEmpty())
				sb.append(" [").append(property).append(']');
			
			if (data_ instanceof IEDataSendNumber && ((IEDataSendNumber) data_).sendNumber() != null) {
				final short[] values = ((IEDataSendNumber) data_).sendNumber();
				if (values.length > 4)
					sb.append(", send=[").append(values[0]).append(",").append(values[1]).append(",").append(values[2]).append(",").append(values[3]).append(",...]");
				else if (((IEDataSendNumber) data_).sendNumber().length >= 1)
					sb.append(", send=").append(Arrays.toString(((IEDataSendNumber) data_).sendNumber()));
			}
			else if (data_ instanceof IEDataMetaValue && ((IEDataMetaValue) data_).preSendValues() != null) {
				final short[] values = ((IEDataMetaValue) data_).preSendValues();
				if (values.length > 4)
					sb.append(", meta=[").append(values[0]).append(",").append(values[1]).append(",").append(values[2]).append(",").append(values[3]).append(",...]");
				else if (((IEDataMetaValue) data_).preSendValues().length >= 1)
					sb.append(", meta=").append(Arrays.toString(((IEDataMetaValue) data_).preSendValues()));
			}
			if (isHostname_)
				sb.append(isHTMLStyle_?"<br>":"\n").append(getHostname(tinepath));
			
			return sb.toString();
		}
		
		final IEDataAddress address = (IEDataAddress) data_;
		
		sb.append("HARDWARE: Crate=");
		sb.append(address.crate());
		sb.append(", BSA=");
		sb.append((address.address() - (address.address() % 32)));
		sb.append(", BSA+=");
		sb.append((address.address() % 32));
		
		if (isHostname_)
			sb.append(isHTMLStyle_?"<br>":"\n").append(getHostname(tinepath));
		
		return sb.toString();
	}
	
	public final static int fetchDelimiter(final String tinepath) {
		int numberOfDelimiter = 0;
		for(final char c: tinepath.toCharArray())
			if (c == '/')
				numberOfDelimiter++;
		return numberOfDelimiter;
	}
	
	/**
	 * <p>Queries the TINE context, server, device name and the corresponding
	 * property which is returned as a full TINE name, separated with terminating
	 * back slashes.</p>
	 *
	 * @param data_ The data component.
	 * @return Returns the full TINE context to the property.
	 */
	public static final String getTinename(final IEData data_)
	{
		if (data_ == null)
			return MESSAGE_MISSING_DATA;
		
		if (data_.getName(EName.PRE_TINEADDRESS) == null || data_.getName(EName.PRE_TINEADDRESS).isEmpty())
			return MESSAGE_MISSING_TINE_ADDRESS;
		
		return data_.getName(EName.PRE_TINEADDRESS) + '/' + data_.toString();
	}

	// - - GET TINE ADDRESS - - - - - - - - - - - - - - - - - - - - - - - - - -

	/**
	 * Returns on the TINE context, device server, and the property the full
	 * TINE address in a printable way.
	 * 
	 * @param context_ TINE context.
	 * @param deviceServer_ TINE device server.
	 * @param property_ TINE property.
	 * @return The full TINE address.
	 * @see #getTINEAddress(String, String, String, String, boolean, boolean)
	 */
	public final static String getTINEAddress(
			final String context_,
			final String deviceServer_,
			final String property_)
	{
		return getTINEAddress(context_, deviceServer_, null, property_);
	}
	
	/**
	 * Returns on the TINE context, device server, device name, and the 
	 * property the full TINE address in a printable way.
	 * 
	 * @param context_ TINE context.
	 * @param deviceServer_ TINE device server.
	 * @param deviceName_ TINE device name.
	 * @param property_ TINE property.
	 * @return The full TINE address.
	 * @see #getTINEAddress(String, String, String, String, boolean, boolean)
	 */
	public final static String getTINEAddress(
			final String context_, 
			final String deviceServer_, 
			final String deviceName_, 
			final String property_)
	{
		return getTINEAddress(context_, deviceServer_, deviceName_, property_, true);
	}
	
	/**
	 * Returns on the TINE context, device server, device name, and the 
	 * property the full TINE address in a printable way. The output is in HTML
	 * style.
	 * 
	 * @param context_ TINE context.
	 * @param deviceServer_ TINE device server.
	 * @param deviceName_ TINE device name.
	 * @param property_ TINE property.
	 * @param isHostname_ Displayes the host name if set to {@code true}.
	 * @return The full TINE address.
	 * @see #getTINEAddress(String, String, String, String, boolean, boolean)
	 */
	public final static String getTINEAddress(
			final String context_, 
			final String deviceServer_, 
			final String deviceName_, 
			final String property_,
			final boolean isHostname_)
	{
		return getTINEAddress(context_, deviceServer_, deviceName_, property_, isHostname_, true);
	}

	/**
	 * Returns on the TINE context, device server, device name, and the 
	 * property the full TINE address in a printable way. The output is in HTML
	 * style.
	 * 
	 * @param context_ TINE context.
	 * @param deviceServer_ TINE device server.
	 * @param deviceName_ TINE device name.
	 * @param property_ TINE property.
	 * @param isHostname_ Displayes the host name if set to {@code true}.
	 * @param isHTMLStyle_ Enables the HTML style output with a value of {@code true}.
	 * @return The full TINE address.
	 * @see #getTINEAddress(String, String, String, String, boolean, boolean)
	 */
	public final static String getTINEAddress(
			final String context_, 
			final String deviceServer_, 
			final String deviceName_, 
			final String property_,
			final boolean isHostname_,
			final boolean isHTMLStyle_)
	{
		return getTINEAddress(context_, deviceServer_, deviceName_, property_, isHostname_, isHTMLStyle_, true);
	}

	/**
	 * Returns on the TINE context, device server, device name, and the 
	 * property the full TINE address in a printable way.
	 * 
	 * @param context_ TINE context.
	 * @param deviceServer_ TINE device server.
	 * @param deviceName_ TINE device name.
	 * @param property_ TINE property.
	 * @param isHostname_ Displayes the host name if set to {@code true}.
	 * @param isHTMLStyle_ Enables the HTML style output with a value of {@code true}.
	 * @param isMultiline Enables a multiline output.
	 * @return The full TINE address.
	 */
	public final static String getTINEAddress( 
			final String context_, 
			final String deviceServer_, 
			final String deviceName_, 
			final String property_,
			final boolean isHostname_,
			final boolean isHTMLStyle_,
			final boolean isMultiline_)
	{
		final StringBuilder builder = new StringBuilder();
		
		if (context_ != null && !context_.isEmpty())
			builder.append('/').append(context_);
		if (deviceServer_ != null && !deviceServer_.isEmpty())
			builder.append('/').append(deviceServer_);
		if (deviceName_ != null && !deviceName_.isEmpty())
			builder.append('/').append(deviceName_);
		else
			builder.append("/#0");
		if (property_ != null && !property_.isEmpty())
			builder.append(" [").append(property_).append(']');
		if (isHostname_)
			builder.append(isMultiline_?(isHTMLStyle_?"<br>":"\n"):"").append(getHostname(context_, deviceServer_));
		
		return builder.toString();
	}
	
	/**
	 * Returns on the ACOP transport instance the TINE address including the
	 * access rate and protocol.
	 * 
	 * @param transport_ ACOP transport.
	 * @return Returns the TINE address.
	 */
	public final static String getTINEAddress(
			final AcopTransport transport_)
	{
		return getTINEAddress(transport_, true);
	}
	
	/**
	 * Returns on the ACOP transport instance the TINE address including the
	 * access rate and protocol.
	 * 
	 * @param transport_ ACOP transport.
	 * @param isHostname_ Displayes the host name if set to {@code true}.
	 * @return Returns the TINE address.
	 */
	public final static String getTINEAddress(
			final AcopTransport transport_,
			final boolean isHostname_)
	{
		return getTINEAddress(transport_, isHostname_, true);
	}
	
	/**
	 * Returns on the ACOP transport instance the TINE address including the
	 * access rate and protocol.
	 * 
	 * @param transport_ ACOP transport.
	 * @param isHostname_ Displayes the host name if set to {@code true}.
	 * @param isHTMLStyle_ Enables the HTML style output with a value of {@code true}.
	 * @return Returns the TINE address.
	 */
	public final static String getTINEAddress(
			final AcopTransport transport_,
			final boolean isHostname_,
			final boolean isHTMLStyle_)
	{
		return getTINEAddress(transport_, isHostname_, isHTMLStyle_, true);
	}
	
	/**
	 * Returns on the ACOP transport instance the TINE address including the
	 * access rate and protocol.
	 * 
	 * @param transport_ ACOP transport.
	 * @param isHostname_ Displayes the host name if set to {@code true}.
	 * @param isHTMLStyle_ Enables the HTML style output with a value of {@code true}.
	 * @param isMultiline Enables a multiline output.
	 * @return Returns the TINE address.
	 */
	public final static String getTINEAddress(
			final AcopTransport transport_,
			final boolean isHostname_,
			final boolean isHTMLStyle_,
			final boolean isMultiline_)
	{
		if (transport_ == null)
			return MESSAGE_MISSING_ACOP_TRANSPORT;
		
		return getTINEAddress(transport_.getDeviceContext(), transport_.getDeviceGroup(), transport_.getDeviceName(), transport_.getDeviceProperty(), isHostname_, isHTMLStyle_, isMultiline_) 
			+ " @ " + transport_.getAccessRate() + " on " + transport_.getAccessProtocol();
	}

	private final static String HOSTNAME = " (HOSTNAME=%s)";
	private final static String HOSTNAME_AND_LOCATION = " (HOSTNAME=%s - %s)";
	private final static String NO_TCP_IP_CONNECTION = " (NO TCP/IP CONNECTION: %s)";
	private final static String EXCEPTION = " (EXCEPTION: %s)";
	private final static String CANNOT_CONVERT = " (CANNOT CONVERT: %s)";
	
	/**
	 * Returns the host name on the give TINE context and device server in the
	 * prefix TINE address string.
	 * 
	 * @param tinePrefix_ TINE prefix address string.
	 * @return Host name.
	 */
	public final static String getHostname(final String tinePrefix_)
	{
		final String[] parts = tinePrefix_.split("/");
		
		if (parts.length >= 3 
				&& parts[0].isEmpty() 
				&& !parts[1].isEmpty() 
				&& !parts[2].isEmpty())
			return getHostname(parts[1], parts[2]);
		else
			return String.format(CANNOT_CONVERT, tinePrefix_);
	}
	
	/**
	 * Returns the host name on the give TINE context and device server.
	 * 
	 * @param context_ TINE context.
	 * @param deviceServer_ TINE device server.
	 * @return Host name.
	 */
	public final static String getHostname(final String context_, final String deviceServer_)
	{
		if (context_ == null || deviceServer_ == null)
			return String.format(EXCEPTION, "no context/device server");

		final String tinePrefix = '/' + context_ + '/' + deviceServer_;
		
		try
		{
			if (!SERVERS.containsKey(tinePrefix))
			{
				final TSrvEntry server = 
					TQuery.getModuleAddress(deviceServer_, context_);
				
				SERVERS.put(tinePrefix, server);
			}
					
			if (SERVERS.get(tinePrefix) != null && !FEC_INFO.containsKey(tinePrefix))
			{
				final FECInfo fecInfo =
					TQuery.getServerInformation(SERVERS.get(tinePrefix).getFecName());

				FEC_INFO.put(tinePrefix, fecInfo);
			}

			if (SERVERS.containsKey(tinePrefix) && FEC_INFO.containsKey(tinePrefix))
				return String.format(HOSTNAME_AND_LOCATION, 
						SERVERS.get(tinePrefix).getFecAddr().fecHost.getHostName(),
						FEC_INFO.get(tinePrefix).getLocation());
			else if (SERVERS.containsKey(tinePrefix))
				return String.format(HOSTNAME, 
						SERVERS.get(tinePrefix).getFecAddr().fecHost.getHostName());
			else
				return String.format(NO_TCP_IP_CONNECTION, tinePrefix);
		}
		catch(Throwable exception)
		{
			return String.format(EXCEPTION, tinePrefix);
		}
	}
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	
	/**
	 * Queries if the boundaries of the range (which means the values of
	 * <tt>from_</tt> and <tt>to_</tt>) are members of the list and both are
	 * inside the list in the right order. If this results in a true the value
	 * of <tt>true</tt> is return otherwise a <tt>false</tt>.
	 *
	 * @param from_ Is the minimum border of the range.
	 * @param to_ Is the maximum border of the range.
	 * @param list_ Is a collection, which should include the range.
	 * @return A <tt>true</tt> is returned, if the elements of the range are
	 * 		inside the list and both are in a natural order otherwise a
	 * 		<tt>false</tt> is returned.
	 */
	public static final boolean containsRange(final IEData from_,
			final IEData to_, final Collection<IEData> list_)
	{
		if (from_ == null || to_ == null || list_ == null)
			throw new NullPointerException("One or more elements are null.");

		if (!(list_.contains(from_) && list_.contains(to_)))
			return false;
		if (from_.equals(to_))
			return true;

		boolean isInside = false;
		IEData data = from_;

		for(IEData current: list_)
		{
			if (current.equals(data) && isInside)
				return true;
			else if (current.equals(data) && !isInside)
			{
				isInside = true;
				data = to_;
			}
		}
		return false;
	}

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	/**
	 * Collects all collections in an array of data elements and installs them
	 * all in one list which. If the delivered collections are empty an empty
	 * list is returned.
	 *
	 * @param collections_ An array of all collections.
	 * @return A list is returned which carries all delivered collections.
	 */
	public static final List<IEData> getList(
			final Collection<IEData>... collections_)
	{
		List<IEData> list = new ArrayList<IEData>(collections_.length);
		for(Collection<IEData> current: collections_)
			list.addAll(current);
		return list;
	}

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	/**
	 * Sleeps {@code wait_} milliseconds before continuing work. If the delay
	 * time is lower then 100 it will wait 100 milliseconds. If the delay
	 * time is greater then 60000 it will wait 60000 milliseconds.
	 *
	 * @param wait_ Is the time in milliseconds. If it is lesser then 100 it
	 * 		will return immediately.
	 */
	public static final void sleep(final int wait_)
	{

		if (wait_ <= 0)
			return;

		try
		{
			if (wait_ < 100)
			{
				Logger.SWApp.warn("Delay time is lower then 100 milliseconds " +
						"-> waiting 100 millseconds.");
				Thread.sleep(100);
			}
			else if (wait_ > 60000)
			{
				Logger.SWApp.error("Delay time is greater then 60000 milliseconds " +
					"-> waiting 60000 millseconds.");
				Thread.sleep(60000);
			}
			else
				Thread.sleep(wait_);
		}
		catch(InterruptedException exception)
		{
		}
	}

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	/**
	 * Exits this application in a proper way by terminating all opened TINE 
	 * links and printing the corresponding message in the log file. 
	 * 
	 * @param message_ A message. Can also be {@code null}.
	 * @param exception_ The exception. Can also be {@code null}.
	 * @param linkData_ The link data which can also be {@code null}. If it is
	 * 		{@code null} then this exit method will <b>NOT</b> close all TINE
	 * 		links from this application before it exits.
	 * @param devices_ The device group. This can also be {@code null}. Then the
	 * 		TINE links are <b>NOT</b> closed before existing this application.
	 */
	public final static void exit(
			final String message_, 
			final Throwable exception_,
			final ILinkData linkData_,
			final IEDeviceGroup[] devices_)
	{
		if (message_ != null || exception_ != null)
			Logger.SWApp.error(message_, exception_);
		else if (message_ == null || exception_ != null)
			Logger.SWApp.error(CLAZZ + " - catch a throwable", exception_);
		
		Logger.SWApp.info(CLAZZ + " - exiting this application in 10 seconds...");

		new Thread(new Runnable() {
			public void run()
			{
				Utility.sleep(10000);
				
				if (linkData_ != null)
					linkData_.removeAllLinks(devices_);

				Main.exit(-1);
			}
		}, "ShutdownApplicationThread").start();
	}

	/**
	 * Exits this application gracefully, for example when an unhandled 
	 * exception is thrown.
	 * 
	 * @param exit_ The exit code number.
	 */
	public final static void shutdown(final int exit_)
	{
		Main.exit(exit_);
	}

	/**
	 * Exits this application gracefully, for example when an unhandled 
	 * exception is thrown.
	 * 
	 * @param exit_ The exit code number.
	 * @param linkData_ Link data structure.
	 * @param devices_ Device group collection.
	 */
	public final static void shutdown(
			final int exit_,
			final ILinkData linkData_, 
			final IEDeviceGroup... devices_)
	{
		if (linkData_ != null)
			linkData_.removeAllLinks(devices_);
		
		shutdown(exit_);
	}
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	/**
	 * Sets all component to be visible or not depending in the boolean value.
	 *
	 * @param isEnabled_ Is the enable (<tt>true</tt>) or disable (<tt>false</tt>)
	 * 		argument.
	 * @param container_ Is the container which should be enabled or disabled
	 * 		an all of its child elements.
	 */
	public static final void setVisibleComponent(final boolean isEnabled_,
			final Container container_)
	{
		if (container_ == null)
			return;
		container_.setVisible(isEnabled_);
		if (container_.getComponents() == null)
			return;

		for(Component component: container_.getComponents())
		{
			if (component instanceof Container)
			{
				if (((Container) component).getComponents() != null)
					setVisibleComponent(isEnabled_,
							((Container) component).getComponents());
				component.setVisible(isEnabled_);
			}
		}
	}

	/**
	 * Sets all component to be visible or not depending in the boolean value.
	 *
	 * @param isEnabled_ Is the enable (<tt>true</tt>) or disable (<tt>false</tt>)
	 * 		argument.
	 * @param components_ Is the list of all components. This should never be
	 * 		null.
	 */
	private static final void setVisibleComponent(final boolean isEnabled_,
			final Component[] components_)
	{
		if (components_ == null)
			throw new NullPointerException("The list of all components is null.");

		for(Component component: components_)
		{
			if (component instanceof Container)
			{
				if (((Container) component).getComponents() != null)
					setVisibleComponent(isEnabled_,
							((Container) component).getComponents());
				component.setVisible(isEnabled_);
			}
		}
	}

	/**
	 * Enables or disables all components as its second argument.
	 *
	 * @param isEnabled_ Is the enable (<tt>true</tt>) or disable (<tt>false</tt>)
	 * 		argument.
	 * @param container_ Is the container which should be enabled or disabled
	 * 		an all of its child elements.
	 */
	public static final void setEnableComponent(final boolean isEnabled_,
			final Container container_)
	{
		if (container_ == null)
			return;

		container_.setEnabled(isEnabled_);

		if (container_.getComponents() == null)
			return;

		for(Component component: container_.getComponents())
		{
			if (component instanceof Container)
			{
				if (((Container) component).getComponents() != null)
					setEnableComponent(isEnabled_,
							((Container) component).getComponents());
				component.setEnabled(isEnabled_);
			}
		}
	}

	/**
	 * Enables or disables all components as its second argument.
	 *
	 * @param isEnabled_ Is the enable (<tt>true</tt>) or disable (<tt>false</tt>)
	 * 		argument.
	 * @param components_ Is the list of all components. This should never be
	 * 		null.
	 */
	private static final void setEnableComponent(final boolean isEnabled_,
			final Component[] components_)
	{
		if (components_ == null)
			throw new NullPointerException("The list of all components is null.");

		for(Component component: components_)
		{
			if (component instanceof Container)
			{
				if (((Container) component).getComponents() != null)
					setEnableComponent(isEnabled_,
							((Container) component).getComponents());
				component.setEnabled(isEnabled_);
			}
		}
	}

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	/**
	 * Throws an exception and prints it with the logger error expression.
	 *
	 * @param throwable_ Is the <tt>Throwable</tt> class. This should never be
	 * 		null, otherwise a fatal error is printed on screen.
	 * @param message_ Is the message. If it is not null it will be printed
	 * 		before the throwable is printed.
	 * @param isStackTrace_ If this value is true it will also print the hole
	 * 		stack trace otherwise only the first line will be printed (i.e.
	 * 		the exception and the line including the line number).
	 */
	public static final void throwException(final Throwable throwable_,
			final String message_, final boolean isStackTrace_)
	{
		if (throwable_ == null)
		{
			Logger.SWApp.error("A fatal error occurred on the static method" +
					" Utility.throwException!");
			return;
		}

		if (isStackTrace_)
		{
			if (message_ != null)
				Logger.SWApp.error(message_, throwable_);
			else
				Logger.SWApp.error(throwable_);
		}
		else
		{
			if (message_ != null)
				Logger.SWApp.error(message_ + ": " + throwable_.toString() +
						((throwable_.getStackTrace().length != 0)
							?" on " + throwable_.getStackTrace()[0].toString() +
									'(' + throwable_.getStackTrace()[0].getLineNumber() + ')'
							:""));
			else
				Logger.SWApp.error(throwable_.toString() +
						((throwable_.getStackTrace().length != 0)
							?" on " + throwable_.getStackTrace()[0].toString() +
									'(' + throwable_.getStackTrace()[0].getLineNumber() + ')'
							:""));
		}
	}

	/**
	 * Throws an exception and prints it with the logger warn expression.
	 *
	 * @param throwable_ Is the <tt>Throwable</tt> class. This should never be
	 * 		null, otherwise a fatal error is printed on screen.
	 * @param message_ Is the message. If it is not null it will be printed
	 * 		before the throwable is printed.
	 * @param isStackTrace_ If this value is true it will also print the hole
	 * 		stack trace otherwise only the first line will be printed (i.e.
	 * 		the exception and the line including the line number).
	 */
	public static final void throwWarning(final Throwable throwable_,
			final String message_, final boolean isStackTrace_)
	{
		if (throwable_ == null)
		{
			Logger.SWApp.error("A fatal error occurred on the static method" +
					" Utility.throwException!");
			return;
		}

		if (isStackTrace_)
		{
			if (message_ != null)
				Logger.SWApp.warn(message_, throwable_);
			else
				Logger.SWApp.warn(throwable_);
		}
		else
		{
			if (message_ != null)
				Logger.SWApp.warn(message_ + ": " + throwable_.toString() +
						((throwable_.getStackTrace().length != 0)
							?" on " + throwable_.getStackTrace()[0].toString() +
									'(' + throwable_.getStackTrace()[0].getLineNumber() + ')'
							:""));
			else
				Logger.SWApp.warn(throwable_.toString() +
						((throwable_.getStackTrace().length != 0)
							?" on " + throwable_.getStackTrace()[0].toString() +
									'(' + throwable_.getStackTrace()[0].getLineNumber() + ')'
							:""));
		}
	}

	/**
	 * This method will send a message to the corresponding developer of the
	 * application with different other information which the developer needs
	 * replicate the error message. May be also a short description of what the
	 * user has done can also entered in the open e-mail application.
	 *
	 * @param throwable_ Is the error exception.
	 * @param builder_ Is a string builder instance which represents all the
	 * 		displayed information.
	 */
	public static final void sendMailTo(final Throwable throwable_,
			final StringBuilder builder_)
	{
		// TODO: Here you should call different method(s) for sending an
		// email to the corresponding developer of the application (and possibly
		// also second user, if the regarded developer is on holiday) and other
		// information about the environment (may be all the properties which
		// are set by java, a list of all properties set by OS, etc. ...).
		if (Desktop.isDesktopSupported())
		{
			String homedir =
				System.getProperty(USER_HOME)!=null?System.getProperty(USER_HOME):UNKNOWN;
			String appname =
				System.getProperty(APP_APPLICATION_NAME)!=null?System.getProperty(APP_APPLICATION_NAME):UNKNOWN;
			String appdeveloper =
				System.getProperty(APP_DEVELOPER)!=null?System.getProperty(APP_DEVELOPER):UNKNOWN;
			String appauthor =
				System.getProperty(APP_AUTHOR)!=null?System.getProperty(APP_AUTHOR):UNKNOWN;
			String appphone =
				System.getProperty(APP_PHONE)!=null?System.getProperty(APP_PHONE):UNKNOWN;
			String appgroup =
				System.getProperty(APP_GROUP)!=null?System.getProperty(APP_GROUP):UNKNOWN;
			String appbuilding =
				System.getProperty(APP_BUILDING)!=null?System.getProperty(APP_BUILDING):UNKNOWN;
			String approom =
				System.getProperty(APP_ROOM)!=null?System.getProperty(APP_ROOM):UNKNOWN;
			String appemail =
				System.getProperty(APP_EMAIL)!=null?System.getProperty(APP_EMAIL):UNKNOWN;
			String appmailinglist =
				System.getProperty(APP_MAILINGLIST)!=null?System.getProperty(APP_MAILINGLIST):UNKNOWN;
			String appversion =
				System.getProperty(APP_VERSION)!=null?System.getProperty(APP_VERSION):UNKNOWN;
			String username =
				System.getProperty(USER_NAME)!=null?System.getProperty(USER_NAME):UNKNOWN;

			String hostname = null;
			if (System.getenv("COMPUTERNAME") != null)
				hostname = System.getenv("COMPUTERNAME");
			else if (System.getenv("HOSTNAME") != null)
				hostname = System.getenv("HOSTNAME");
			else
				hostname = UNKNOWN;

			String cr = System.getProperty("line.separator");
			File file = null;
			try
			{
				file = createErrorFile(homedir, appname, appversion);
				PrintStream stream = new PrintStream(file);
				stream.printf("Hostname: %s" + cr, hostname);
				stream.printf("Username (on the local computer system): %s" + cr, username);
				stream.printf("Username (developer of the application): %s" + cr, appauthor);
				stream.printf("Username (developer of the application): %s" + cr, appdeveloper);
				stream.printf("Phone: %s" + cr,  appphone);
				stream.printf("Group: %s" + cr, appgroup);
				stream.printf("Room: %s/%s" + cr,  appbuilding, approom);
				stream.printf("Mailingliste: %s" + cr, appmailinglist);
				stream.printf("Application, name: %s" + cr, appname);
				stream.printf("Application, version: %s" + cr + cr,	appversion);

				stream.print("Fatal error:" + cr);
				if (throwable_ != null)
					throwable_.printStackTrace(stream);
				else
					stream.printf("*** unknown ***");
				stream.print(cr + cr);

				stream.print("OS environment variables:" + cr);
				for(String key: System.getenv().keySet())
					stream.printf("%s=%s" + cr, key, System.getenv(key));

				stream.flush();
				stream.close();
			}
			catch(IOException exception)
			{
				Logger.SWApp.error("An IOException is thrown.", exception);
			}
			catch(SecurityException exception)
			{
				Logger.SWApp.error("A security exception is thrown.", exception);
			}
			catch(Exception exception)
			{
				Logger.SWApp.error("An exception is thrown.", exception);
			}

			// query if the e-mail address exits and is valid
			if (System.getProperty(APP_EMAIL) != null &&
					System.getProperty(APP_EMAIL).matches(CGeneralConstant.VALID_EMAIL))
			{
				try
				{
					String uriString =
						"mailto:" + appemail + "&subject=Exception" +
							((appname!=null)
									?("%20in%20" + appname)
									:("")) +
							((appversion!=null)
									?(",%20version%20" + appversion)
									:("")) +
							((hostname!=null)
									?(",%20hostname%20" + hostname)
									:("")) +
							"&body=" + getNormalizedString(file.getAbsolutePath());
					uriString = uriString.replace(" ", "%20");
					URI uri = new URI(uriString);
					Desktop.getDesktop().mail(uri);
				}
				catch(Exception exception)
				{
					exception.printStackTrace();
					return;
				}
			}
			else
			{
				JOptionPane.showOptionDialog(
						null,
						
						"<html><center>" +
						"Die Applikation enthlt einen Fehler in der Darstellung<br>" +
						"der E-Mail Adresse: " + System.getProperty(APP_EMAIL) + "<br><br>" +
						"Bitte wenden Sie sich an den Entwickler dieses Programms:<br><br>" +
						appdeveloper + " - " + appgroup + ", Geb. " + appbuilding + "/"	+ approom + ", Tel. " + appphone +
						"</center></html>",
						
						"Falsche Darstellung der E-Mail Adresse",
						
						JOptionPane.OK_OPTION, 
						JOptionPane.ERROR_MESSAGE,
						CGeneralConstant.ICON_MESSAGE_FATAL_32x32, // icon
						
						new Object[] {"OK"}, "OK");
			}
		}
		else
			Logger.SWApp.error("This environment doesn't support the singelton" +
					" of java.awt.Desktop.");
	}

	/**
	 * Creates a new error file in the home directory of the application. All
	 * delivered parameters must be set.
	 *
	 * @param homedir_ Is the home directory.
	 * @param appname_ Is the application name.
	 * @param appversion_ Is the application version.
	 * @return The file is returned if it can be created.
	 * @throws IOException This exception is thrown if the file could not be
	 * 		created.
	 * @throws SecurityException This exceptions is thrown when the application
	 * 		is not allowed to created or write to the file.
	 */
	private static final File createErrorFile(final String homedir_,
			final String appname_, final String appversion_)
		throws IOException, SecurityException
	{
		if (homedir_ == null || appname_ == null || appversion_ == null)
			throw new IllegalArgumentException("All or some values are null.");

		String separator = System.getProperty("file.separator");
		String application =
			homedir_ + separator + JAVA_APPLICATION_PATH + separator
			+ ((appname_!=null)?(appname_):("unknown"))
			+ ((appversion_!=null)?('-' + appversion_):(""));
		File fileapplication = new File(application);
		fileapplication.mkdirs();

		String error = application + separator + "error_" +
			System.currentTimeMillis() + ".txt";
		File fileerror = new File(error);
		if (fileerror.exists())
			throw new IOException("The file " + error + " already exits.");
		if (!fileerror.createNewFile())
			throw new IOException("Unable to create the file " + error);

		return fileerror;
	}


	/**
	 * Creates a message dialog box in the middle of the screen. It reports that
	 * a fatal error has occurred and gives the user the opportunity to select
	 * different options. As the first option you can select the 'OK' button
	 * which returns a value of 0 and 'Sende E-Mail Nachricht' as 1.
	 *
	 * @return Returns the selected option as an integer value. This option can
	 * 		be 0 or 1.
	 * @see #fatalErrorMessage(Component)
	 */
	public static final int fatalErrorMessage()
	{
		return fatalErrorMessage(null);
	}

	/**
	 * Creates a modal message dialog box in the middle of the application or
	 * when the application is null in the middle of the screen. It
	 * reports an <b>fatal error</b> message to the user.The user can send an 
	 * e-mail to the developer with the stack trace in the mail content. 
	 * The application is finished after performing the e-mail, where the e-mail
	 * is send by this method. Their exists two constant fields which represent 
	 * the selected option: {@link #FATAL_EXIT}, or {@link #FATAL_SEND_EMAIL}.
	 *
	 * @param parent_ Is the parent component in which this dialog box should be
	 * 		centered. If this component is {@code null} it is shown in the
	 * 		center of the desktop screen.
	 * @return Returns the selected option as an integer value. This option is one
	 * 		of the constant fields in this class: {@link #FATAL_EXIT},
	 * 		or {@link #FATAL_SEND_EMAIL}, depending on the current input value 
	 * 		of {@code exception_}.
	 */
	public static final int fatalErrorMessage(final Component parent_)
	{
		String appname =
			System.getProperty(APP_APPLICATION_NAME)!=null?System.getProperty(APP_APPLICATION_NAME):UNKNOWN;
		String appdeveloper =
			System.getProperty(APP_DEVELOPER)!=null?System.getProperty(APP_DEVELOPER):UNKNOWN;
		String appphone =
			System.getProperty(APP_PHONE)!=null?System.getProperty(APP_PHONE):UNKNOWN;
		String appgroup =
			System.getProperty(APP_GROUP)!=null?System.getProperty(APP_GROUP):UNKNOWN;
		String appbuilding =
			System.getProperty(APP_BUILDING)!=null?System.getProperty(APP_BUILDING):UNKNOWN;
		String approom =
			System.getProperty(APP_ROOM)!=null?System.getProperty(APP_ROOM):UNKNOWN;
		return JOptionPane.showOptionDialog(
				parent_,
				
				"<html><center>" +
				"Die Applikation enthlt einen schwerwiegenden Fehler.<br>" +
				"Applikation: " + appname + "<br><br>" +
				"Bitte wenden Sie sich an den Entwickler des Programms:<br><br>" +
				appdeveloper + " - " + appgroup + ", Geb. " + appbuilding + "/" + approom + ", Tel. " + appphone + "<br><br>" +
				"Es kann ebenfalls eine Nachricht an den Entwickler ber<br>" +
				"E-Mail verschickt werden." +
				"</center></html>",
				
				"Schwerwiegender Fehler",
				
				JOptionPane.YES_NO_OPTION, 
				
				JOptionPane.ERROR_MESSAGE,
				
				CGeneralConstant.ICON_MESSAGE_FATAL_32x32, // icon
				
				new Object[] {"Beenden", "Sende E-Mail Nachricht"}, "OK");
	}

	/**
	 * Creates a modal message dialog box in the middle of the application or
	 * when the application is null in the middle of the screen. It
	 * reports an <b>fatal error</b> message to the user. Depending on the delivered
	 * exception you have different options to select from. Without the exception
	 * the user can finish the application immediately.
	 * With the exception the user can also send an e-mail to the developer with
	 * the stack trace in the mail content. The application is then finished
	 * after performing the e-mail, where the e-mail
	 * is send by this method. Their exists two constant fields which
	 * represent the selected option: {@link #FATAL_EXIT}, or {@link #FATAL_SEND_EMAIL}.
	 *
	 * @param title_ Is the title of the option pane. This can also be null.
	 * @param message_ Is the message of the option pane. This can also be null.
	 * @param exception_ Is the exception by itself. This can also be null.
	 * @return Returns the selected option as an integer value. This option is one
	 * 		of the constant fields in this class: {@link #FATAL_EXIT},
	 * 		or {@link #FATAL_SEND_EMAIL}, depending on the current input value 
	 * 		of {@code exception_}.
	 * @see #fatalErrorMessage(Component, String, String, Throwable)
	 */
	public static final int fatalErrorMessage(final String title_,
			final String message_, final Throwable exception_)
	{
		return fatalErrorMessage(null, title_, message_, exception_);
	}

	/**
	 * Creates a modal message dialog box in the middle of the application or
	 * when the application is null in the middle of the screen. It
	 * reports an <b>fatal error</b> message to the user. Depending on the delivered
	 * exception you have different options to select from. Without the exception
	 * the user can finish the application immediately.
	 * With the exception the user can also send an e-mail to the developer with
	 * the stack trace in the mail content. The application is then finished
	 * after performing the e-mail, where the e-mail
	 * is send by this method. Their exists two constant fields which
	 * represent the selected option: {@link #FATAL_EXIT}, or {@link #FATAL_SEND_EMAIL}.
	 *
	 * @param parent_ Is the parent component in which this dialog box should be
	 * 		centered. If this component is {@code null} it is shown in the
	 * 		center of the desktop screen.
	 * @param title_ Is the title of the option pane. This can also be null.
	 * @param message_ Is the message of the option pane. This can also be null.
	 * @param exception_ Is the exception by itself. This can also be null.
	 * @return Returns the selected option as an integer value. This option is one
	 * 		of the constant fields in this class: {@link #FATAL_EXIT},
	 * 		or {@link #FATAL_SEND_EMAIL}, depending on the current input value 
	 * 		of {@code exception_}.
	 */
	public static final int fatalErrorMessage(final Component parent_,
			final String title_, final String message_,
			final Throwable exception_)
	{
		int option = 0;
		String[] results;
		String okResult;
		String appname =
			System.getProperty(APP_APPLICATION_NAME)!=null?System.getProperty(APP_APPLICATION_NAME):UNKNOWN;
		String appdeveloper =
			System.getProperty(APP_DEVELOPER)!=null?System.getProperty(APP_DEVELOPER):UNKNOWN;
		String appphone =
			System.getProperty(APP_PHONE)!=null?System.getProperty(APP_PHONE):UNKNOWN;
		String appgroup =
			System.getProperty(APP_GROUP)!=null?System.getProperty(APP_GROUP):UNKNOWN;
		String appbuilding =
			System.getProperty(APP_BUILDING)!=null?System.getProperty(APP_BUILDING):UNKNOWN;
		String approom =
			System.getProperty(APP_ROOM)!=null?System.getProperty(APP_ROOM):UNKNOWN;

		String sendError;
		if (exception_ != null)
		{
			option = JOptionPane.YES_NO_OPTION;
			results = new String[] {"Beenden", "Sende E-Mail Nachricht"};
			okResult = "Beenden";
			sendError = "<br><br>" +
				"Es kann ebenfalls eine Nachricht an den Entwickler ber<br>" +
				"E-Mail verschickt werden.";
		}
		else
		{
			option = JOptionPane.YES_OPTION;
			results = new String[] {"Beenden"};
			okResult = "Beenden";
			sendError = "";
		}
		int result = JOptionPane.showOptionDialog(
				parent_,
				
				"<html><center>" + 
				((message_==null)?("Applikation: " + appname):(message_)) + "<br><br>" +
				"Bitte wenden Sie sich an den Entwickler des Programms:<br><br>" +
				appdeveloper + " - " + appgroup + ", Geb. " + appbuilding + "/" + approom + ", Tel. " + appphone + 
				sendError
				+ "</center></html>",
				
				((title_ == null)?"Schwerwiegender Fehler":title_),
				
				option, // option type: YES_NO_CHANCEL, ...
				
				JOptionPane.ERROR_MESSAGE,  // message type: ERROR_MESSAGE, ...
				
				CGeneralConstant.ICON_MESSAGE_ERROR_32x32, // icon
				
				results, okResult);
		if (result == FATAL_SEND_EMAIL)
			sendMailTo(exception_, null);
		if (exception_ != null)
			Logger.SWApp.error(exception_);
		return result;
	}

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	/**
	 * Creates a modal message dialog box in the middle of the application or
	 * when the application is null in the middle of the screen. It
	 * reports an <b>error</b> message to the user. Depending on the delivered
	 * exception you have different options to select from. Without the exception
	 * the user can finish the application immediately or continue in executing.
	 * With the exception the user can also send an e-mail to the developer with
	 * the stack trace in the mail content. The application is then finished
	 * after performing the e-mail, where the e-mail
	 * is send by this method. Their exists three constant fields which
	 * represent the selected option: {@link #ERROR_CONTINUE}, {@link #ERROR_EXIT},
	 * or {@link #ERROR_SEND_EMAIL}.
	 *
	 * @param title_ Is the title of the option pane. This can also be null.
	 * @param message_ Is the message of the option pane. This can also be null.
	 * @param exception_ Is the exception by itself. This can also be null.
	 * @return Returns the selected option as an integer value. This option is one
	 * 		of the constant fields in this class: {@link #ERROR_CONTINUE},
	 * 		{@link #ERROR_EXIT}, or {@link #ERROR_SEND_EMAIL}, depending on the
	 * 		current input value of {@code exception_}.
	 * @see #errorMessage(String, String, Throwable)
	 */
	public static final int errorMessage(final String title_,
			final String message_, final Throwable exception_)
	{
		return errorMessage(null, title_, message_, exception_);
	}

	/**
	 * Creates a modal message dialog box in the middle of the application or
	 * when the application is null in the middle of the screen. It
	 * reports an <b>error</b> message to the user. Depending on the delivered
	 * exception you have different options to select from. Without the exception
	 * the user can finish the application immediately or continue in executing.
	 * With the exception the user can also send an e-mail to the developer with
	 * the stack trace in the mail content. The application is then finished
	 * after performing the e-mail, where the e-mail
	 * is send by this method. Their exists three constant fields which
	 * represent the selected option: {@link #ERROR_CONTINUE}, {@link #ERROR_EXIT},
	 * or {@link #ERROR_SEND_EMAIL}.
	 *
	 * @param parent_ Is the parent component in which this dialog box should be
	 * 		centered. If this component is {@code null} it is shown in the
	 * 		center of the desktop screen.
	 * @param title_ Is the title of the option pane. This can also be null.
	 * @param message_ Is the message of the option pane. This can also be null.
	 * @param exception_ Is the exception by itself. This can also be null.
	 * @return Returns the selected option as an integer value. This option is one
	 * 		of the constant fields in this class: {@link #ERROR_CONTINUE},
	 * 		{@link #ERROR_EXIT}, or {@link #ERROR_SEND_EMAIL}, depending on the
	 * 		current input value of {@code exception_}.
	 */
	public static final int errorMessage(
			final Component parent_, 
			final String title_,
			final String message_, 
			final Throwable exception_)
	{
		int option = 0;
		String[] results;
		String okResult;
		String appname =
			System.getProperty(APP_APPLICATION_NAME)!=null?System.getProperty(APP_APPLICATION_NAME):UNKNOWN;
		String appdeveloper =
			System.getProperty(APP_DEVELOPER)!=null?System.getProperty(APP_DEVELOPER):UNKNOWN;
		String appphone =
			System.getProperty(APP_PHONE)!=null?System.getProperty(APP_PHONE):UNKNOWN;
		String appgroup =
			System.getProperty(APP_GROUP)!=null?System.getProperty(APP_GROUP):UNKNOWN;
		String appbuilding =
			System.getProperty(APP_BUILDING)!=null?System.getProperty(APP_BUILDING):UNKNOWN;
		String approom =
			System.getProperty(APP_ROOM)!=null?System.getProperty(APP_ROOM):UNKNOWN;

		if (exception_ != null)
		{
			option = JOptionPane.YES_NO_CANCEL_OPTION;
			results = new String[] {"Forfahren", "Beenden", "E-Mail und Beenden"};
			okResult = results[0];
		}
		else
		{
			option = JOptionPane.YES_NO_OPTION;
			results = new String[] {"Fortfahren", "Beenden"};
			okResult = results[0];
		}
		int result = JOptionPane.showOptionDialog(
				parent_,
				
				"<html><center>" + 
				((message_==null)?("Applikation: " + appname):(message_)) + "<br><br>" +
				"Bitte wenden Sie sich an den Entwickler des Programms:<br><br>" +
				appdeveloper + " - " + appgroup + ", Geb. " + appbuilding + "/" + approom + ", Tel. " + appphone +
				((exception_ != null)? "<br><br>" +
				"Es kann ebenfalls eine Nachricht an den Entwickler ber<br>" +
				"E-Mail verschickt werden.":"") +
				"</center></html>",
				
				((title_ == null)?"Fehler":title_),
				
				option, // message type: ERROR_MESSAGE, ...
				
				JOptionPane.ERROR_MESSAGE, // option type: YES_NO_CHANCEL, ...
				
				CGeneralConstant.ICON_MESSAGE_ERROR_32x32, // icon
				
				results, 
				okResult
				);
		if (exception_ != null && result == ERROR_SEND_EMAIL)
			sendMailTo(exception_, null);
		return result;
	}
	
	/**
	 * Creates a modal message dialog box in the middle of the application or
	 * when the application is null in the middle of the screen. It
	 * reports an <b>error</b> message to the user. The user can finish the 
	 * application immediately or continue in executing.
     * Their exists two possible constant fields which represent the selected 
     * option: {@link #ERROR_CONTINUE} or {@link #ERROR_EXIT}.
	 * 
	 * @param parent_ Is the parent component in which this dialog box should be
	 * 		centered. If this component is {@code null} it is shown in the
	 * 		center of the desktop screen.
	 * @param title_ Is the title of the option pane. This can also be null.
	 * @param message_ Is the message of the option pane. This can also be null.
	 * @param options_ A list of options as a string array. At least one option
	 * 		must be defined.
	 * @return Returns the selected option as an integer value. This option is one
	 * 		of the constant fields in this class: {@link #ERROR_CONTINUE} or
	 * 		{@link #ERROR_EXIT}.
	 */
	public static final int errorMessageOptions(
			final Component parent_, 
			final String title_,
			final String message_, 
			final String[] options_)
	{
		return errorMessageOptions(parent_, title_, message_, options_, true);
	}

	
	/**
	 * Creates a modal message dialog box in the middle of the application or
	 * when the application is null in the middle of the screen. It
	 * reports an <b>error</b> message to the user. The user can finish the 
	 * application immediately or continue in executing.
     * Their exists two possible constant fields which represent the selected 
     * option: {@link #ERROR_CONTINUE} or {@link #ERROR_EXIT}.
	 * 
	 * @param parent_ Is the parent component in which this dialog box should be
	 * 		centered. If this component is {@code null} it is shown in the
	 * 		center of the desktop screen.
	 * @param title_ Is the title of the option pane. This can also be null.
	 * @param message_ Is the message of the option pane. This can also be null.
	 * @param options_ A list of options as a string array. At least one option
	 * 		must be defined.
	 * @param isDeveloperAvailable_ Boolean value which defines if the developer
	 * 		information is available or not.
	 * @return Returns the selected option as an integer value. This option is one
	 * 		of the constant fields in this class: {@link #ERROR_CONTINUE} or
	 * 		{@link #ERROR_EXIT}.
	 */
	public static final int errorMessageOptions(
			final Component parent_, 
			final String title_,
			final String message_, 
			final String[] options_,
			final boolean isDeveloperAvailable_)
	{
		if (options_ == null)
			throw new NullPointerException("Options is null.");
		if (options_.length == 0)
			throw new IllegalArgumentException("At least one option must be" +
					" definied.");
		
		final String okResult = options_[0];
		String appname =
			System.getProperty(APP_APPLICATION_NAME)!=null?System.getProperty(APP_APPLICATION_NAME):UNKNOWN;
		String appdeveloper =
			System.getProperty(APP_DEVELOPER)!=null?System.getProperty(APP_DEVELOPER):UNKNOWN;
		String appphone =
			System.getProperty(APP_PHONE)!=null?System.getProperty(APP_PHONE):UNKNOWN;
		String appgroup =
			System.getProperty(APP_GROUP)!=null?System.getProperty(APP_GROUP):UNKNOWN;
		String appbuilding =
			System.getProperty(APP_BUILDING)!=null?System.getProperty(APP_BUILDING):UNKNOWN;
		String approom =
			System.getProperty(APP_ROOM)!=null?System.getProperty(APP_ROOM):UNKNOWN;
			
		String developer =
			(isDeveloperAvailable_) ? ("<br><br>" +
					 "Sie knnen sich an den Entwickler des Programms wenden:<br><br>" +
					 appdeveloper + " - " + appgroup + ", Geb. " + appbuilding + "/"
					 + approom + ", Tel. " + appphone) : ("");
		String message =
			(message_==null)?("Applikation: " + appname):(message_);
			
			
		int result = JOptionPane.showOptionDialog(
				parent_,
				
				"<html><center>" + 
				message + 
				developer + 
				"</center></html>",
				
				((title_ == null)?"Fehler":title_),
				
				0, // option type: YES_NO_CHANCEL, ...
				
				JOptionPane.ERROR_MESSAGE, 
				
				CGeneralConstant.ICON_MESSAGE_ERROR_32x32, // icon
				
				options_, 
				okResult
				);
		return result;
	}
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	
	/**
	 * It
	 * reports a <b>warning</b> message that has occurred and gives the user 
	 * the opportunity to select different options for the ongoing execution.</p>
	 *  
	 * <p>The first choice is when you appoints no exception as an argument
	 * (means its value is {@code null}). That opens only two options for
	 * the user: to continue the execution or to  finish the application.</p>
	 * 
	 * <p>The second choice is when you appoint an exception as an argument. That
	 * opens three possible options for the user: to continue the execution, to 
	 * finish the application, or to finish the application and send an e-mail to 
	 * the developer with the stack trace inside the mail content.</p>
	 * 
	 * <p>The e-mail is automatically created in this static method and you don't
	 * need to build and send it to the developer on your own!</p>
	 * 
	 * <p>Their exists three constant fields which deliver the selected options
	 * from the method: {@link #WARNING_CONTINUE}, {@link #WARNING_EXIT}, and
	 * {@link #WARNING_SEND_EMAIL}.</p> 
	 *
	 * @param title_ Is the title of the option pane. This can also be null.
	 * @param message_ Is the message of the option pane. This can also be null.
	 * @param exception_ Is the exception by itself. This can also be null.
	 * @return Returns the selected option as an integer value. This option is one
	 * 		of the constant fields in this class: {@link #WARNING_CONTINUE},
	 * 		{@link #WARNING_EXIT}, or {@link #WARNING_SEND_EMAIL}, depending on the
	 * 		current input value of {@code exception_}.
	 * @see #warningMessage(String, String, Throwable)
	 */
	public static final int warningMessage(final String title_,
			final String message_, final Throwable exception_)
	{
		return warningMessage(null, title_, message_, exception_);
	}

	/**
	 * <p>Creates a modal message dialog box in the middle of the application or
	 * where the application is null in the center of the screen. It
	 * reports a <b>warning</b> message that has occurred and gives the user 
	 * the opportunity to select different options for the ongoing execution.</p>
	 * 
	 * <p>The first choice is when you appoints no exception as an argument
	 * (means its value is {@code null}). That opens only two options for
	 * the user: to continue the execution or to  finish the application.</p>
	 * 
	 * <p>The second choice is when you appoint an exception as an argument. That
	 * opens three possible options for the user: to continue the execution, to 
	 * finish the application, or to finish the application and send an e-mail to 
	 * the developer with the stack trace inside the mail content.</p>
	 * 
	 * <p>The e-mail is automatically created in this static method and you don't
	 * need to build and send it to the developer on your own!</p>
	 * 
	 * <p>Their exists three constant fields which deliver the selected options
	 * from the method: {@link #WARNING_CONTINUE}, {@link #WARNING_EXIT}, and
	 * {@link #WARNING_SEND_EMAIL}.</p> 
	 *
	 * @param parent_ Is the parent component in which this dialog box should be
	 * 		centered. If this component is {@code null} it is shown in the
	 * 		center of the desktop screen.
	 * @param title_ Is the title of the option pane. This can also be null.
	 * @param message_ Is the message of the option pane. This can also be null.
	 * @param exception_ Is the exception by itself. This can also be null.
	 * @return Returns the selected option as an integer value. This option is one
	 * 		of the constant fields in this class: {@link #WARNING_CONTINUE},
	 * 		{@link #WARNING_EXIT}, or {@link #WARNING_SEND_EMAIL}, depending on the
	 * 		current input value of {@code exception_}.
	 */
	public static final int warningMessage(final Component parent_, final String title_,
			final String message_, final Throwable exception_)
	{
		int option = 0;
		String[] results;
		String okResult;
		String appname =
			System.getProperty(APP_APPLICATION_NAME)!=null?System.getProperty(APP_APPLICATION_NAME):UNKNOWN;
		String appdeveloper =
			System.getProperty(APP_DEVELOPER)!=null?System.getProperty(APP_DEVELOPER):UNKNOWN;
		String appphone =
			System.getProperty(APP_PHONE)!=null?System.getProperty(APP_PHONE):UNKNOWN;
		String appgroup =
			System.getProperty(APP_GROUP)!=null?System.getProperty(APP_GROUP):UNKNOWN;
		String appbuilding =
			System.getProperty(APP_BUILDING)!=null?System.getProperty(APP_BUILDING):UNKNOWN;
		String approom =
			System.getProperty(APP_ROOM)!=null?System.getProperty(APP_ROOM):UNKNOWN;

		if (exception_ != null)
		{
			option = JOptionPane.DEFAULT_OPTION;
			results = new String[] {"Forfahren", "Beenden", "E-Mail und Beenden"};
			okResult = results[0];
		}
		else
		{
			option = JOptionPane.YES_NO_OPTION;
			results = new String[] {"Fortfahren", "Beenden"};
			okResult = results[0];
		}
		int result = JOptionPane.showOptionDialog(
				parent_,
				
				"<html><center>" + 
				((message_==null)?("Applikation: " + appname):(message_)) + 
				"<br><br>" +
				"Bitte wenden Sie sich an den Entwickler des Programms:<br><br>" +
				appdeveloper + " - " + appgroup + ", Geb. " + appbuilding + "/"
				+ approom + ", Tel. " + appphone +
				((exception_ != null)? "<br><br>" +
				"Es kann ebenfalls eine Nachricht an den Entwickler ber<br>" +
				"E-Mail verschickt werden.":"") +
				"</center></html>",
				
				((title_ == null)?"Warnung":title_),
				
				option, // message type: ERROR_MESSAGE, ...
				
				JOptionPane.WARNING_MESSAGE, // option type: YES_NO_CHANCEL, ...
				
				CGeneralConstant.ICON_MESSAGE_WARNING_32x32, // icon
				
				results, 
				okResult
				);
		if (exception_ != null && result == WARNING_SEND_EMAIL)
			sendMailTo(exception_, null);
		return result;
	}

	/**
	 * <p>Creates a modal message dialog box in the middle of the screen. It
	 * reports that a <b>warning</b> has occurred and gives the user the opportunity to
	 * select different options.</p>
	 * 
	 * @param parent_ Is the parent component in which this dialog box should be
	 * 		centered. If this component is {@code null} it is shown in the
	 * 		center of the desktop screen.
	 * @param title_ Is the title of the option pane. This can also be null.
	 * @param message_ Is the message of the option pane. This can also be null.
	 * @param options_ A list of options as a string array. At least one option
	 * 		must be defined.
	 * @return Returns the selected option as an integer value. This option is one
	 * 		of the constant fields in this class: {@link #WARNING_CONTINUE}, or 
	 * 		{@link #WARNING_SEND_EMAIL}.
	 */
	public static final int warningMessageOptions(
			final Component parent_, 
			final String title_,
			final String message_, 
			final String[] options_)
	{
		if (options_ == null)
			throw new NullPointerException("Options is null.");
		if (options_.length == 0)
			throw new IllegalArgumentException("At least one option must be" +
					" definied.");
		
		final String okResult = options_[0];
		String appname =
			System.getProperty(APP_APPLICATION_NAME)!=null?System.getProperty(APP_APPLICATION_NAME):UNKNOWN;
		String appdeveloper =
			System.getProperty(APP_DEVELOPER)!=null?System.getProperty(APP_DEVELOPER):UNKNOWN;
		String appphone =
			System.getProperty(APP_PHONE)!=null?System.getProperty(APP_PHONE):UNKNOWN;
		String appgroup =
			System.getProperty(APP_GROUP)!=null?System.getProperty(APP_GROUP):UNKNOWN;
		String appbuilding =
			System.getProperty(APP_BUILDING)!=null?System.getProperty(APP_BUILDING):UNKNOWN;
		String approom =
			System.getProperty(APP_ROOM)!=null?System.getProperty(APP_ROOM):UNKNOWN;

		int result = JOptionPane.showOptionDialog(
				parent_,
				
				"<html><center>" + 
				((message_==null)?("Applikation: " + appname):(message_)) + 
				"<br><br>" +
				"Bitte wenden Sie sich an den Entwickler des Programms:<br><br>" +
				appdeveloper + " - " + appgroup + ", Geb. " + appbuilding + "/"
				+ approom + ", Tel. " + appphone +
				"</center></html>",
				
				((title_ == null)?"Warnung":title_),
				
				0, // option type: YES_NO_CHANCEL, ...
				
				JOptionPane.WARNING_MESSAGE, 
				
				CGeneralConstant.ICON_MESSAGE_WARNING_32x32, // icon
				
				options_, 
				okResult
				);
		return result;
	}
	
	// - - DISPLAY TINE DIALOG BOXES - - - - - - - - - - - - - - - - - - - - - -

	/**
	 * Creates a TINE error dialog box popping up on screen and displaying the
	 * returned TINE error code and string value.
	 * 
	 * @param root_ Root container in which the dialog box should be centered. 
	 * 		If this is {@code null} then it is center in the middle of the 
	 * 		screen.
	 * @param tineAddress_ TINE address.
	 * @param error_ TINE error code.
	 */
	public final static void createTINEErrorDialog(
			final Container root_,
			final TLink link_, 
			final int error_)
	{
		final StringBuilder builder = new StringBuilder();
		
		if (link_ == null)
			builder.append("Es wurde der folgende Fehler zurckgegeben:<br><br>")
			.append(TErrorList.getErrorString(error_)).append("<br><br>")
			.append("Dabei liegen momentan keine Informationen ber den TINE-Link vor.");
		else
		{
			final String tineAddress = '/' + link_.getFullDeviceNameAndProperty();
			
			String host = null;
			if (link_.srvAddr != null && link_.srvAddr.getFecAddr() != null
					&& link_.srvAddr.getFecAddr().fecHost != null)
				host = link_.srvAddr.getFecAddr().fecHost.getHostName();
			
			if (host == null)
				builder.append("Die TINE Adresse ").append(tineAddress).append(" gab die folgende<br>")
				.append("Fehlermeldung zurck: ")
				.append(TErrorList.getErrorString(error_)).append(" (").append(error_).append(')');
			else
				builder.append("Die TINE Adresse ").append(tineAddress).append(" unter dem Host<br>")
				.append(host).append(" gab die folgende<br>")
				.append("Fehlermeldung zurck: ").append(TErrorList.getErrorString(error_)).append(" (").append(error_).append(')');
		}
		
		if (Utility.errorMessage(root_, "Verbindungsfehler", builder.toString(), null) == Utility.ERROR_EXIT)
			System.exit(-1);
	}

	/**
	 * Creates a TINE error dialog box popping up on screen and displaying the
	 * returned TINE error code and string value.
	 * 
	 * @param root_ Root container in which the dialog box should be centered. 
	 * 		If this is {@code null} then it is center in the middle of the 
	 * 		screen.
	 * @param tineAddress_ TINE address.
	 * @param error_ TINE error code.
	 */
	public final static void createTINEErrorDialog(
			final Container root_,
			final String tineAddress_, 
			final String host_, 
			final int error_)
	{
		final StringBuilder builder = new StringBuilder();
		final boolean isHost = host_ != null && !host_.isEmpty();
		final boolean isTineAddress = tineAddress_ != null && !tineAddress_.isEmpty();
		
		if (isHost && isTineAddress)
			builder.append("Die TINE Adresse ").append(tineAddress_).append(" unter dem Host<br>")
			.append(host_).append(" gab die folgende<br>")
			.append("Fehlermeldung zurck: ").append(TErrorList.getErrorString(error_)).append(" (").append(error_).append(')');
		else if (!isHost && isTineAddress)
			builder.append("Die TINE Adresse ").append(tineAddress_).append(" gab die folgende<br>")
			.append("Fehlermeldung zurck: ")
			.append(TErrorList.getErrorString(error_)).append(" (").append(error_).append(')');
		else
			builder.append("Es wurde ein Verbindungsfehler zurckgegeben, aber<br>")
			.append("keine Informationen ber die TINE Adresse oder den Hostnamen:<br>")
			.append(TErrorList.getErrorString(error_)).append(" (").append(error_).append(')');
		
		if (Utility.errorMessage(root_, "Verbindungsfehler", builder.toString(), null) == Utility.ERROR_EXIT)
			System.exit(-1);
	}

	/**
	 * Creates a TINE error dialog box popping up on screen and displaying the
	 * returned TINE error code and string value.
	 * 
	 * @param root_ Root container in which the dialog box should be centered. 
	 * 		If this is {@code null} then it is center in the middle of the 
	 * 		screen.
	 * @param tineAddress_ TINE address.
	 * @param error_ TINE error code.
	 */
	public final static void createTINEErrorDialog(
			final Container root_,
			final String tineAddress_, 
			final int error_)
	{
		final StringBuilder builder = new StringBuilder();
		final boolean isTineAddress = tineAddress_ != null && !tineAddress_.isEmpty();
		
		if (isTineAddress)
			builder.append("Die TINE Adresse ").append(tineAddress_).append(" gab die folgende<br>")
			.append("Fehlermeldung zurck: ")
			.append(TErrorList.getErrorString(error_)).append(" (").append(error_).append(')');
		else
			builder.append("Es wurde ein Verbindungsfehler zurckgegeben, aber<br>")
			.append("keine Informationen ber die TINE Adresse oder den Hostnamen:<br>")
			.append(TErrorList.getErrorString(error_)).append(" (").append(error_).append(')');
		
		if (Utility.errorMessage(root_, "Verbindungsfehler", builder.toString(), null) == Utility.ERROR_EXIT)
			System.exit(-1);
	}
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	/**
	 * Returns a string which has the transformed value of special characters
	 * into a %-notation. Currently these are only the values of a space (' ')
	 * and a backslash ('\').
	 *
	 * @param string_ The transforming string.
	 * @return The result.
	 */
	public static final String getNormalizedString(final String string_)
	{
		if (string_ == null)
			return "";

		StringBuffer normal = new StringBuffer();
		for(int i = 0; i < string_.length(); i++)
		{
			switch(string_.charAt(i))
			{
			case ' ':
				normal.append("%20");
				break;
			case '\\':
				normal.append("%5C");
				break;
			default:
				normal.append(string_.charAt(i));
			}
		}
		return normal.toString();
	}

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	/** Displays a moving label. The delay time between two states can also
	 * be delivered if it is greater then 50ms.
	 *
	 * @param delay_ The delay time between two states.
	 * @param str_ Is the string value to be printed in the Swing label.
	 * @param label_ Is the label itself.
	 */
	public static final void movingLabel(final int delay_, final String str_, final JLabel label_)
	{
		label_.setText(str_);
		Timer timer = new Timer((delay_ < 50)?50:delay_, new ActionListener() {

			private int count = 0;

			public void actionPerformed(final ActionEvent event_)
			{
				label_.setText(
						str_.substring(str_.length() - (count % str_.length()), str_.length())
						+
						str_.substring(0, str_.length() - (count % str_.length()))
						);
				count++;
			}
		});
		timer.start();
	}

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	/**
	 * Debugs all components and prints them all on the java console window.
	 */
	public static final void debugPrintAllComponents()
	{
        try
        {
	        System.out.println("\n\n");
	        for(IEData observer: ComponentSubject.getInstance().fetchObservers())
	        {
	        	System.out.println(observer.toString());
	        	for(IViewComponent component: ComponentSubject.getInstance()
	        			.fetchComponents(observer))
	        		System.out.println("\t" + component.getClass().getSimpleName() + ": " +
	        				((component.fetchText(observer) != null)
	        				? component.fetchText(observer)
	        				: component.toString()));
	        }
        }
        catch(Exception exception)
        {
        	Logger.SWApp.error("An error occurred on the debugging method for" +
        			" printing all components.", exception);
        }
	}

	// -------------------------------------------------------------------------

	private static long currentTimeMillis = 0;

	/**
	 * Prints and stores the current debug time in milliseconds.
	 *
	 * @param timeInMillis_ The current time in milliseconds.
	 * @param isDebugTime_ Depends if the current output should be displayed
	 * 		({@code true}) or not ({@code false}).
	 * @see #debugCurrentTime(long, String, boolean)
	 */
	public static final void debugCurrentTime(
			final long timeInMillis_,
			final boolean isDebugTime_)
	{
		debugCurrentTime(timeInMillis_, null, isDebugTime_);
	}

	/**
	 * Prints and stores the current debug time in milliseconds and prints out
	 * the message (if not {@code null} on the console. Depending on the call
	 * the output varies: the first time it only prints that the time counting
	 * has started. All other calls it prints the duration and sets the internal
	 * field to the current time in milliseconds.
	 *
	 * @param timeInMillis_ The current time in milliseconds.
	 * @param message_ A message (which can also be null).
	 * @param isDebugTime_ Depends if the current output should be displayed
	 * 		({@code true}) or not ({@code false}).
	 */
	public static final void debugCurrentTime(
			final long timeInMillis_,
			final String message_,
			final boolean isDebugTime_)
	{
		if (currentTimeMillis == 0 && isDebugTime_)
		{
			System.err.println("CURRENT APPLICATION TIME - START");
		}
		else if (message_ != null && isDebugTime_)
		{
			System.err.println("CURRENT APPLICATION TIME - DURATION=" + (timeInMillis_ - currentTimeMillis) + " - " + message_);
		}
		else if (isDebugTime_)
		{
			System.err.println("CURRENT APPLICATION TIME - DURATION=" + (timeInMillis_ - currentTimeMillis));
		}
		currentTimeMillis = timeInMillis_;
	}

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	private static long absoluteTimeMillis = 0;

	/**
	 * Prints the absolute time in milliseconds. The first call the current time
	 * is internally stored in a field.
	 *
	 * @param timeInMillis_ The current time in milliseconds.
	 * @param isDebugTime_ Depends if the current output should be displayed
	 * 		({@code true}) or not ({@code false}).
	 */
	public static final void debugAbsoluteTime(
			final long timeInMillis_,
			final boolean isDebugTime_)
	{
		if (absoluteTimeMillis == 0 && isDebugTime_)
		{
			System.err.println("ABSOLUTE APPLICATION TIME - START");
			absoluteTimeMillis = timeInMillis_;
		}
		else if (isDebugTime_)
		{
			System.err.println("ABSOLUTE APPLICATION TIME - DURATION=" + (timeInMillis_ - absoluteTimeMillis));
		}
	}
	
	// -------------------------------------------------------------------------

	/**
	 * Sets the MstApp logging level to the appropriated value.
	 * 
	 * @param loggingLevel_ Enumeration from MstApp which contains the logging
	 * 		level.
	 */
	public static final void setLoggingLevel(final LogLvl loggingLevel_)
	{
		LoggerManager.getConsolAppender().getFilter().setLogLvl(loggingLevel_);
	}
	
	// - - SHOWS TINE INFORMATION - - - - - - - - - - - - - - - - - - - - - - - 

	/**
	 * Shows exhaustive information about the links used by the called 
	 * application and an overview with permanently TINE requests from 
	 * {@link AcopTransport} instance.
	 * 
	 * @param frame_ Root component.
	 * @param links_ TINE links.
	 * @return The result depends on the shown selected buttons. Currently this
	 * 		dialog box will return a {@link DIALOG_RESULT#OK} if the dialog box
	 * 		was quit with the OK button or {@link DIALOG_RESULT#CHANCEL} if the 
	 * 		dialog box is canceled. The last one only occurs if the link is null
	 * 		or no elements are stored in it.
	 */
	@Deprecated
	public static final DIALOG_RESULT showLinkInfo(final Frame frame_, final AcopTransport transport_)
	{
		return DIALOG_RESULT.CANCEL;
	}

	/**
	 * Shows exhaustive information about the links used by the called 
	 * application and an overview with permanently TINE requests from the
	 * {@link ILinkData} instance.
	 * 
	 * @param frame_ Root component.
	 * @param linkData_ Link data instance.
	 * @param devices_ Collections of device group types.
	 * @return The result depends on the shown selected buttons. Currently this
	 * 		dialog box will return a {@link DIALOG_RESULT#OK} if the dialog box
	 * 		was quit with the OK button or {@link DIALOG_RESULT#CHANCEL} if the 
	 * 		dialog box is canceled. The last one only occurs if the link is null
	 * 		or no elements are stored in it.
	 */
	@Deprecated
	public static final DIALOG_RESULT showLinkInfo(final Frame frame_, final ILinkData linkData_, final IEDeviceGroup... devices_)
	{
		if (linkData_ == null || devices_ == null || devices_.length == 0)
		{
			Logger.SWApp.warn(CLAZZ + " - no device groups are defined or the device group is empty.");
			return DIALOG_RESULT.CANCEL;
		}
		
		final List<TLink> lstLinks = new ArrayList<TLink>();
		try
		{
			for(final IEDeviceGroup device: devices_)
				if (linkData_.hasLink(device, EAccess.READ))
						lstLinks.add(linkData_.getLink(device, EAccess.READ));
				else if (linkData_.hasLinkGroup(device, EAccess.READ))
					for(final TLink link: linkData_.getLinkGroup(device, EAccess.READ).values())
						lstLinks.add(link);
			
			return showLinkInfo(frame_, lstLinks.toArray(new TLink[lstLinks.size()]));
		}
		catch (final TineException exception)
		{
			Logger.SWApp.error(CLAZZ + " - TINE exception  - unable to get device" +
					" groups from ILinkData: " + TErrorList.getErrorString(exception.getErrorCode()));
		}
			
		return DIALOG_RESULT.CANCEL;
	}
	
	/**
	 * Shows exhaustive information about the links used by the called 
	 * application and an overview with permanently TINE requests from 
	 * {@link TLink} collection instances.
	 * 
	 * @param frame_ Root component.
	 * @param links_ TINE links.
	 * @return The result depends on the shown selected buttons. Currently this
	 * 		dialog box will return a {@link DIALOG_RESULT#OK} if the dialog box
	 * 		was quit with the OK button or {@link DIALOG_RESULT#CHANCEL} if the 
	 * 		dialog box is canceled. The last one only occurs if the link is null
	 * 		or no elements are stored in it.
	 */
	@Deprecated
	public static DIALOG_RESULT showLinkInfo(final Frame frame_, final TLink... links_)
	{
		if (links_ == null || links_.length == 0)
		{
			Logger.SWApp.warn(CLAZZ + " - no links are defined.");
			return DIALOG_RESULT.CANCEL;
		}
		
		final int numberOfLinks = links_.length;
		
		// ---------------------------------------------------------------------

		final JLabel lblHeader = new JLabel("<html><center>Verbindungsinformationen der einzelnen TINE links<br>werden im Sekundentakt aktualisiert:</center></html>");
		lblHeader.setIcon(CGeneralConstant.ICON_MESSAGE_INFORMATION_32x32);
		lblHeader.setHorizontalAlignment(SwingConstants.CENTER);
		lblHeader.setHorizontalTextPosition(SwingConstants.RIGHT);
		lblHeader.setVerticalTextPosition(SwingConstants.CENTER);
		
		// ---------------------------------------------------------------------
		
		final JButton btnOk = new JButton("OK");
		
		final JButton btnEMail = new JButton("Sende E-Mail an den Entwickler");
		btnEMail.setEnabled(false);
		
		@SuppressWarnings({ "rawtypes", "unchecked" })
		final JComboBox eMailSource = new JComboBox(new String[] { "UNBEKANNT", "Marcus Walla (marcus.walla@desy.de" });
		eMailSource.setSelectedIndex(0);
		eMailSource.setEnabled(false);
		final JLabel lblEMailSource = new JLabel("Benutzer:");
		
		// ---------------------------------------------------------------------
		
		final DefaultTableModel model = getLinkModel(links_);
		final JTable tblLinks = new JTable(model);
		tblLinks.setRowSelectionAllowed(false);
		tblLinks.setColumnSelectionAllowed(false);
		tblLinks.setShowGrid(false);
		tblLinks.setRowHeight(20);
		tblLinks.setDefaultRenderer(Integer.class, new TableCellRenderer() {

			public Component getTableCellRendererComponent(JTable table,
					Object value, boolean isSelected, boolean hasFocus,
					int row, int column)
			{
				final JLabel lbl;
				
				if (column == 1 && value instanceof Integer)
				{
					lbl = new JLabel(TErrorList.getErrorString((Integer) value) + " (" + value + ')');
					if (((Integer) value).intValue() == 0)
					{
						lbl.setOpaque(false);
						lbl.setForeground(CColor.MST_DEFAULT_FOREGROUND);
						lbl.setBackground(CColor.MST_DEFAULT_BACKGROUND);
					}
					else
					{
						lbl.setOpaque(true);
						lbl.setForeground(CColor.MST_ERROR_FOREGROUND);
						lbl.setBackground(CColor.MST_ERROR_BACKGROUND);
					}
				}
				else
					lbl = new JLabel(value.toString());
				
				return lbl;
			}
			
		});
		
		new Timer(1000, new ActionListener() {

			public void actionPerformed(ActionEvent e)
			{
				for(int i = 0; i < numberOfLinks; i++)
					tblLinks.getModel().setValueAt(new Integer(links_[i].linkStatus), 1, i);
			}
			
		});

		final JScrollPane scrTblLinks = new JScrollPane(tblLinks);
		
		// ---------------------------------------------------------------------

		final JPanel rootPanel = new JPanel();
		final GroupLayout layout = new GroupLayout(rootPanel);
		rootPanel.setLayout(layout);
		
		layout.setAutoCreateContainerGaps(true);
		layout.setAutoCreateGaps(true);

		layout.setHorizontalGroup(layout.createParallelGroup()
				.addComponent(lblHeader)
				.addComponent(scrTblLinks, 500, 500, 500)
				.addGroup(layout.createSequentialGroup()
						.addPreferredGap(ComponentPlacement.RELATED, Short.MAX_VALUE, Short.MAX_VALUE)
						.addComponent(lblEMailSource)
						.addComponent(eMailSource)
						)
				.addGroup(layout.createSequentialGroup()
						.addComponent(btnEMail)
						.addPreferredGap(btnEMail, btnOk, ComponentPlacement.RELATED)
						.addComponent(btnOk)
						)
		);
		layout.setVerticalGroup(layout.createSequentialGroup()
				.addComponent(lblHeader)
				.addComponent(scrTblLinks, 400, 400, 400)
				.addGroup(layout.createParallelGroup()
						.addComponent(lblEMailSource)
						.addComponent(eMailSource)
						)
				.addGroup(layout.createParallelGroup()
						.addComponent(btnEMail)
						.addComponent(btnOk)
						)
		);
		
		// ---------------------------------------------------------------------
		
		final JDialog dialog = new JDialog(frame_, "Verbindungsinformationen...", true);
		btnOk.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e)
			{
				dialog.dispose();
			}
			
		});

		dialog.setContentPane(rootPanel);
		dialog.pack();
		dialog.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
		dialog.setVisible(true);
		
		return DIALOG_RESULT.OK;
	}
	
	private static DefaultTableModel getLinkModel(final TLink... links_)
	{
		final DefaultTableModel model = new DefaultTableModel();

		final Vector<String> columnLinks = new Vector<String>(links_.length);
		final Vector<Integer> columnState = new Vector<Integer>(links_.length);
		for(final TLink link: links_)
		{
			columnLinks.add(link.toString());
			columnState.add(new Integer(link.getLinkStatus()));
		}
		model.addColumn("TINE link", columnLinks);
		model.addColumn("Zustand", columnState);

		return model;
	}
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	private final static String TINE_ERROR_NO_LINK_INFO = "[NO LINK INFORMATION]";
	
	/**
	 * Returns a transformed ACOP transport request informations as a string 
	 * value. There exists per access protocol of TINE one TINE link per
	 * ACOP transport request entry. 
	 * 
	 * @param request_ ACOP transport request.
	 * @return Human readable information about the internal TINE link on the 
	 * 		{@link AcopTransportRequest} instance.
	 */
	public final static String getTINEInfo(
			final AcopTransportRequest request_)
	{
		return getTINEInfo(true, request_);
	}
	
	/**
	 * Returns a transformed ACOP transport request informations as a string 
	 * value. There exists per access protocol of TINE one TINE link per
	 * ACOP transport request entry. 
	 * 
	 * @param isHTMLEscaped_ Delivers if the returned String value is HTML 
	 * 		escaped or not.
	 * @param request_ ACOP transport request.
	 * @return Human readable information about the internal TINE link on the 
	 * 		{@link AcopTransportRequest} instance.
	 */
	public final static String getTINEInfo(
			final boolean isHTMLEscaped_, 
			final AcopTransportRequest request_)
	{
		final StringBuilder builder = new StringBuilder();
		
		final String TINE_INFO_ACOP = 
			"On protocol %s with /%s/%s/%s [%s] @ <%s> with access %s on host %s (FEC ID=%s) - state=%s (%d)"; // 6+3=9 parameters
		final String TINE_INFO_ACOP_HTML = 
			"On protocol %s with /%s/%s/%s [%s] @ &lt;%s&gt; with access %s<br>on host %s (FEC ID=%s) - state=%s (%d)"; // 6+3=9 parameters

		if (request_ == null)
			builder.append(TINE_ERROR_NO_LINK_INFO);
		else
		{
			if (isHTMLEscaped_)
			{
				final TSrvEntry server = 
					TQuery.getModuleAddress(request_.deviceGroup, request_.deviceContext);

				builder
					.append(String.format(TINE_INFO_ACOP_HTML,
							request_.accessProtocol,
							
							request_.deviceContext,
							request_.deviceGroup,
							request_.deviceName,
							request_.deviceProperty,
							
							request_.accessRate,
							request_.accessMode.toString(),
							
							server.getFecAddr().fecHost.getHostName(),
							server.getFecName(),
							
							request_.getStatus(),
							request_.getStatusCode()));
			}
			else
			{
				// non HTML code ...
				final TSrvEntry server = 
					TQuery.getModuleAddress(request_.deviceGroup, request_.deviceContext);

				builder
					.append(String.format(TINE_INFO_ACOP,
							request_.accessProtocol,
							
							request_.deviceContext,
							request_.deviceGroup,
							request_.deviceName,
							request_.deviceProperty,

							request_.accessRate,
							request_.accessMode.toString(),
							
							server.getFecAddr().fecHost.getHostName(),
							server.getFecName(),
							
							request_.getStatus(),
							request_.getStatusCode()));
			}
		}
		
		return builder.toString();	
	}

	/**
	 * Returns a transformed TINE link information as a string value with 
	 * HTML escape sequences enabled. Depending
	 * on the length of the {@code AcopTransportRequest}'s from the 
	 * {@code AcopTransport} the output is formated in proper way.
	 * 
	 * @param transport_ ACOP transport.
	 * @return Human readable information about all internal TINE links on the 
	 * 		{@link AcopTransport} instance.
	 */
	public final static String getTINEInfo(
			final AcopTransport transport_)
	{
		return getTINEInfo(true, transport_, false);
	}

	/**
	 * Returns a transformed TINE link information as a string value with 
	 * HTML escape sequences enabled. Depending
	 * on the length of the {@code AcopTransportRequest}'s from the 
	 * {@code AcopTransport} the output is formated in proper way.
	 * 
	 * @param transport_ ACOP transport.
	 * @param isLastLink_ Returns on the list of all possible links in the 
	 * 		{@code AcopTransport} all ({@code false}) or the last one 
	 * 		({@link true}).
	 * @return Human readable information about all internal TINE links on the 
	 * 		{@link AcopTransport} instance.
	 */
	public final static String getTINEInfo(
			final AcopTransport transport_, 
			final boolean isLastLink_)
	{
		return getTINEInfo(true, transport_, isLastLink_);
	}
	
	/**
	 * Returns a transformed TINE link informations as a string value. Depending
	 * on the length of the {@code AcopTransportRequest}'s from the 
	 * {@code AcopTransport} the output is formated in proper way.
	 * 
	 * @param isHTMLEscaped_ Delivers if the returned String value is HTML 
	 * 		escaped or not.
	 * @param transport_ ACOP transport.
	 * @param isLastLink_ Returns on the list of all possible links in the 
	 * 		{@code AcopTransport} all ({@code false}) or the last one 
	 * 		({@link true}).
	 * @return Human readable information about all internal TINE links on the 
	 * 		{@link AcopTransport} instance.
	 */
	@SuppressWarnings("unchecked")
	public final static String getTINEInfo(
			final boolean isHTMLEscaped_, 
			final AcopTransport transport_, 
			final boolean isLastLink_)
	{
		final StringBuilder builder = new StringBuilder();
		
		if (transport_ == null || transport_.getAcopLinkTable().size() == 0)
			builder.append(TINE_ERROR_NO_LINK_INFO);
		else
		{
			final List<AcopTransportRequest> lstAcopTransportRequest;

			if (!isLastLink_)
				lstAcopTransportRequest = (Vector<AcopTransportRequest>) transport_.getAcopLinkTable();
			else
			{
				lstAcopTransportRequest = new ArrayList<AcopTransportRequest>();
				lstAcopTransportRequest.add((AcopTransportRequest) transport_.getAcopLinkTable().lastElement());
			}
			
			if (isHTMLEscaped_ && lstAcopTransportRequest.size() > 1)
			{
				// HTML code ...
				builder.append("<ol>");
				
				for(final AcopTransportRequest request: lstAcopTransportRequest)
					builder.append("<li>").append(getTINEInfo(true, request)).append("</li>");

				builder.append("</ol>");
			}
			else if (isHTMLEscaped_ && lstAcopTransportRequest.size() == 1)
			{
				// HTML code ...
				builder.append(getTINEInfo(true, lstAcopTransportRequest.get(0)));
			}
			else if (lstAcopTransportRequest.size() > 1)
			{
				// non HTML code ...
				for(int i = 0 ; i < lstAcopTransportRequest.size(); i++)
					builder.append(getTINEInfo(false, lstAcopTransportRequest.get(i))).append('\n');
			}
			else
			{
				// non HTML code ...
				builder.append(getTINEInfo(false, lstAcopTransportRequest.get(0)));
			}
		}
		
		return builder.toString();	
	}

	/**
	 * Returns a transformed TINE link information as a string value with HTML
	 * escape sequences enabled. Depending
	 * on the length of the TINE links the output is formated in proper way.
	 * 
	 * @param links_ TINE links.
	 * @return Human readable information on the TINE links.
	 */
	public final static String getTINEInfo(
			final TLink... links_)
	{
		return getTINEInfo(true, links_);
	}
	
	/**
	 * Returns a transformed TINE link informations as a string value. Depending
	 * on the length of the TINE links the output is formated in proper way.
	 * 
	 * @param isHTMLEscaped_ Delivers if the returned String value is HTML 
	 * 		escaped or not.
	 * @param links_ TINE links.
	 * @return Human readable information on the TINE links.
	 */
	public final static String getTINEInfo(
			final boolean isHTMLEscape_, 
			final TLink... links_)
	{
		final StringBuilder builder = new StringBuilder();
		
		final String TINE_INFO_LINK_ADDRESS = 
			"Protocol: TINE => /%s/%s/%s [%s] @ <%d> with access %s on host %s (FEC ID %s) - state=%s (%d)"; // 10 parameters
		final String TINE_INFO_LINK_ADDRESS_HTML = 
			"Protocol: TINE => /%s/%s/%s [%s] @ &lt;%d&gt; with access %s on host %s (FEC ID %s) - state=%s (%d)"; // 10 parameters

		if (links_ == null || links_.length == 0)
			builder.append(TINE_ERROR_NO_LINK_INFO);
		else if (isHTMLEscape_ && links_.length > 1)
		{
			builder.append("<ol>");
			
			// HTML code ...
			for(TLink link: links_)
				builder
					.append("<li>")
					
					.append(String.format(TINE_INFO_LINK_ADDRESS_HTML,
							link.getContext(), 
							link.getDeviceServer(), 
							link.getDeviceName(), 
							link.getProperty(),
							
							link.devTimeout,
							
							TMode.toString(link.getLinkAccessMode()),
				
							link.srvAddr.getFecAddr().fecHost.getHostName(),
							link.srvAddr.getFecName(),

							TErrorList.getErrorString(link.getLinkStatus()), 
							link.getLinkStatus()))
							
					.append("</li>");
			
			builder.append("</ol>");
		}
		else if (isHTMLEscape_ && links_.length == 1)
		{
			// HTML code ...
			builder.append(String.format(TINE_INFO_LINK_ADDRESS_HTML,
					links_[0].getContext(), 
					links_[0].getDeviceServer(), 
					links_[0].getDeviceName(), 
					links_[0].getProperty(),
					
					links_[0].devTimeout,
					
					TMode.toString(links_[0].getLinkAccessMode()),
		
					links_[0].srvAddr.getFecAddr().fecHost.getHostName(),
					links_[0].srvAddr.getFecName(),

					TErrorList.getErrorString(links_[0].getLinkStatus()), 
					links_[0].getLinkStatus()));
		}
		else if (links_.length > 1)
		{
			// non HTML code ...
			for(int i = 0; i < links_.length; i++)
			{
				builder.append(String.format(TINE_INFO_LINK_ADDRESS,
						links_[i].getContext(), 
						links_[i].getDeviceServer(), 
						links_[i].getDeviceName(), 
						links_[i].getProperty(),
						
						links_[i].devTimeout,
						
						TMode.toString(links_[i].getLinkAccessMode()),
			
						links_[i].srvAddr.getFecAddr().fecHost.getHostName(),
						links_[i].srvAddr.getFecName(),

						TErrorList.getErrorString(links_[i].getLinkStatus()), 
						links_[i].getLinkStatus()));
				if (i < links_.length)
					builder.append('\n');
			}
		}
		else
		{
			builder.append(String.format(TINE_INFO_LINK_ADDRESS,
					links_[0].getContext(), 
					links_[0].getDeviceServer(), 
					links_[0].getDeviceName(), 
					links_[0].getProperty(),
					
					links_[0].devTimeout,
					
					TMode.toString(links_[0].getLinkAccessMode()),
		
					links_[0].srvAddr.getFecAddr().fecHost.getHostName(),
					links_[0].srvAddr.getFecName(),

					TErrorList.getErrorString(links_[0].getLinkStatus()), 
					links_[0].getLinkStatus()));
		}
		
		return builder.toString();
				
	}

	// - - LOG TINE INFORMATION (ACOP TRANSPORT) - - - - - - - - - - - - - - - - 
	
	/**
	 * Log information about the established asynchronous link.
	 * 
	 * @param clazz_ Class name, leading the displayed output.
	 * @param transport_ ACOP transport.
	 * @param linkId_ Link ID.
	 */
	public final static void logTINEAttachedLink(
			final String clazz_, 
			final AcopTransport transport_, 
			final int linkId_)
	{
		if (transport_ != null && transport_.getAcopLinkTable().get(linkId_) != null)
		{
			AcopTransportRequest atr = 
				(AcopTransportRequest) transport_.getAcopLinkTable().get(linkId_);
			Logger.SWTine.info(clazz_ + " - established link on " + 
					atr.accessProtocol + " with /" + 
					atr.deviceContext + '/' +
					atr.deviceGroup + '/' +
					atr.deviceName + " [" +
					atr.deviceProperty + 
					"] @ <" + atr.accessRate + 
					"> on link ID " + linkId_);
		}
	}
	
	/**
	 * Log information about the executed link. The result must also be included
	 * as an argument because {@code AcopTransport} doesn't allow the developer
	 * to get the result as an integer value.
	 * 
	 * @param clazz_ Class name, leading the displayed output.
	 * @param transport_ ACOP transport instance.
	 */
	public final static void logTINEExecuteLink(
			final String clazz_, 
			final AcopTransport transport_)
	{
		if (transport_ != null)
		{
			Logger.SWTine.info(clazz_ + " - executed link on " + 
					transport_.getAccessProtocol() + " with /" + 
					transport_.getDeviceContext() + '/' +
					transport_.getDeviceGroup() + '/' +
					transport_.getDeviceName() + " [" +
					transport_.getDeviceProperty() + 
					"] @ <" + transport_.getAccessRate() + 
					"> => " + transport_.getStatus());
		}
	}
	
	/**
	 * Log information about the executed link. The result must also be included
	 * as an argument because {@code AcopTransport} doesn't allow the developer
	 * to get the result as an integer value.
	 * 
	 * @param clazz_ Class name, leading the displayed output.
	 * @param transport_ ACOP transport instance.
	 * @param result_ Result of the execute operation.
	 */
	public final static void logTINEExecuteLink(
			final String clazz_, 
			final AcopTransport transport_, 
			final int result_)
	{
		if (transport_ != null)
		{
			Logger.SWTine.info(clazz_ + " - executed link on " + 
					transport_.getAccessProtocol() + " with /" + 
					transport_.getDeviceContext() + '/' +
					transport_.getDeviceGroup() + '/' +
					transport_.getDeviceName() + " [" +
					transport_.getDeviceProperty() + 
					"] @ <" + transport_.getAccessRate() + 
					"> => " + TErrorList.getErrorString(result_) + " (" + result_ + ')');
		}
	}

	/**
	 * Log information about a TINE link error. Reports on the last installed link
	 * and displays its information in the logging environment of the application.
	 * 
	 * @param clazz_ Class name, leading the displayed output.
	 * @param transport_ Acop transport.
	 */
	public final static void logTINEErrorLink(
			final String clazz_, 
			final AcopTransport transport_)
	{
		logTINELink(LogLvl.error, clazz_, "error on link", transport_);
	}
	
	/**
	 * Logs information about the last link information in the {@code AcopTransport}
	 * data element. If their doesn't exist any link information a warning is 
	 * logged and this method is exited immediately.
	 * 
	 * @param logLevel_ MstApp logging level.
	 * @param clazz_ Class name.
	 * @param text_ A text field describing the responsibility of this logging
	 * 		 entry.
	 * @param transport_ AcopTransport.
	 */
	public final static void logTINELink(
			final LogLvl logLevel_, 
			final String clazz_, 
			final String text_, 
			final AcopTransport transport_)
	{
		if (transport_ != null && transport_.getAcopLinkTable().lastElement() != null)
		{
			final AcopTransportRequest atr = (AcopTransportRequest) transport_.getAcopLinkTable().lastElement();
			
			if (clazz_ != null && !clazz_.isEmpty() && text_ != null && !text_.isEmpty())
				Logger.doLog(logLevel_, Logger.SWTine,
						String.format("%s - %s on protocol %s with /%s/%s/%s [%s] @ <%s> - %s (%d)", 
								clazz_, 
								text_,
								atr.accessProtocol,
								atr.deviceContext,
								atr.deviceGroup,
								atr.deviceName,
								atr.deviceProperty,
								atr.accessRate,
								atr.getStatusCode()==0?"success":atr.getStatus(), 
								atr.getStatusCode()), 
						null);
			else if (clazz_ != null && !clazz_.isEmpty())
				Logger.doLog(logLevel_, Logger.SWTine, 
						String.format("%s - protocol %s with /%s/%s/%s [%s] @ <%s> - %s (%d)", 
								clazz_, 
								atr.accessProtocol,
								atr.deviceContext,
								atr.deviceGroup,
								atr.deviceName,
								atr.deviceProperty,
								atr.accessRate,
								atr.getStatusCode()==0?"success":atr.getStatus(), 
								atr.getStatusCode()), 
						null);
			else
				Logger.doLog(logLevel_, Logger.SWTine, 
						String.format("protocol %s with /%s/%s/%s [%s] @ <%s> - %s (%d)", 
								atr.accessProtocol,
								atr.deviceContext,
								atr.deviceGroup,
								atr.deviceName,
								atr.deviceProperty,
								atr.accessRate,
								atr.getStatusCode()==0?"success":atr.getStatus(), 
								atr.getStatusCode()), 
								null);
		}
		else
			Logger.SWApp.error(CLAZZ + " - AcopTransport or the returned element of the vector is null.\n" +
					getCallingMethod(Thread.currentThread()));
	}
	
	/**
	 * Logs information about the reflected TINE link ID.
	 * 
	 * @param logLevel_ MstApp logging level.
	 * @param clazz_ Class name.
	 * @param text_ A text field describing the responsibility of this logging
	 * 		 entry.
	 * @param transport_ AcopTransport.
	 * @param linkID_ TINE link ID.
	 */
	public final static void logTINELink(
			final LogLvl logLevel_, 
			final String clazz_, 
			final String text_, 
			final AcopTransport transport_, 
			final int linkID_)
	{
		if (transport_ != null && transport_.getAcopLinkTable().get(linkID_) != null)
		{
			final AcopTransportRequest atr = (AcopTransportRequest) transport_.getAcopLinkTable().get(linkID_);
			
			if (clazz_ != null && !clazz_.isEmpty() && text_ != null && !text_.isEmpty())
				Logger.doLog(logLevel_, Logger.SWTine, String.format("%s - %s %s => %s (%d)", 
						clazz_, 
						text_,
						atr.deviceInfo, 
						atr.getStatus(), 
						atr.getStatusCode()), 
						null);
			else if (clazz_ != null && !clazz_.isEmpty())
				Logger.doLog(logLevel_, Logger.SWTine, String.format("%s - %s => %s (%d)", 
						clazz_, 
						atr.deviceInfo, 
						atr.getStatus(), 
						atr.getStatusCode()), 
						null);
			else
				Logger.doLog(logLevel_, Logger.SWTine, String.format("%s => %s (%d)",
						atr.deviceInfo, 
						atr.getStatus(), 
						atr.getStatusCode()), 
						null);
		}
		else
			Logger.SWApp.error(CLAZZ + " - AcopTransport or the returned element of the vector is null.\n" +
					getCallingMethod(Thread.currentThread()));
	}
	
	// - - LOG TINE INFORMATION (TINE LINK) - - - - - - - - - - - - - - - - - - 
	
	/**
	 * Log information about the successful established asynchronous link.
	 * 
	 * @param clazz_ Class name, leading the displayed output.
	 * @param link_ TINE link.
	 * @param linkId_ Link ID.
	 */
	public final static void logTINEAttachedLink(
			final String clazz_, 
			final TLink link_, 
			final int linkId_)
	{
		if (linkId_ < 0)
			logTINELink(LogLvl.error, clazz_, "unable to established an asynchronous link", link_);
		else
			logTINELink(LogLvl.info, clazz_, "established asynchronous link on ID " + linkId_, link_);
	}
	
	/**
	 * Log information about the executed link.
	 * 
	 * @param clazz_ Class name, leading the displayed output.
	 * @param link_ TINE link.
	 */
	public final static void logTINEExecuteLink(
			final String clazz_, 
			final TLink link_)
	{
		logTINELink(LogLvl.info, clazz_, "execute link", link_);
	}
		
	/**
	 * Log information about a TINE link error.
	 * 
	 * @param clazz_ Class name, leading the displayed output.
	 * @param link_ TINE link.
	 */
	public final static void logTINEErrorLink(
			final String clazz_, 
			final TLink link_)
	{
		logTINELink(LogLvl.error, clazz_, "error on link", link_);
	}

	/**
	 * Logs information about the current TINE links.
	 * 
	 * @param links_ TINE links.
	 */
	public final static void logTINELink(
			final TLink... links_)
	{
		logTINELink(null, links_);
	}

	/**
	 * Logs information about the current TINE links.
	 * 
	 * @param clazz_ Calling clazz name.
	 * @param links_ TINE links.
	 */
	public final static void logTINELink(
			final String clazz_, 
			final TLink... links_)
	{
		logTINELink(LogLvl.info, clazz_, null, links_);
	}
	
	/**
	 * Logs information about the current TINE links.
	 * 
	 * @param logLevel_ MstApp logging level.
	 * @param clazz_ Calling clazz name.
	 * @param message_ Message.
	 * @param links_ TINE links.
	 */
	public final static void logTINELink(
			final LogLvl logLevel_, 
			final String clazz_, 
			final String message_, 
			final TLink... links_)
	{
		if (clazz_ != null && !clazz_.isEmpty() && message_ != null && !message_.isEmpty())
			for(final TLink link: links_)
				Logger.doLog(logLevel_, Logger.SWTine, String.format("%s - %s %s => %s (%d)", 
						clazz_, 
						message_,
						link.toString(),
						TErrorList.getErrorString(link.getLinkStatus()),
						link.getLinkStatus()), null);
		else if (message_ != null && !message_.isEmpty())
			for(final TLink link: links_)
				Logger.doLog(logLevel_, Logger.SWTine, String.format("[unspecified class] - %s %s => %s (%d)", 
						message_, 
						link.toString(),
						TErrorList.getErrorString(link.getLinkStatus()),
						link.getLinkStatus()), null);
		else
			for(final TLink link: links_)
				Logger.doLog(logLevel_, Logger.SWTine, String.format("[unspecified class] - %s => %s (%d)", 
						link.toString(),
						TErrorList.getErrorString(link.getLinkStatus()),
						link.getLinkStatus()), null);
	}
	
	// - - LOG INFORMATION ABOUT THE TINE ACCESSOR FROM TClientLib - - - - - - -
	
	/**
	 * Log information about the successful established asynchronous link.
	 * 
	 * @param clazz_ Class name, leading the displayed output.
	 * @param accessor_ TINE link.
	 */
	public final static void logTINEAttachedLink(
			final String clazz_, 
			final ITineAccessor accessor_)
	{
		logTINELink(LogLvl.info, clazz_, "established asynchronous link", accessor_);
	}
	
	/**
	 * Log information about the successful established asynchronous link.
	 * 
	 * @param clazz_ Class name, leading the displayed output.
	 * @param accessor_ TINE link.
	 * @param linkID_ Link ID.
	 */
	public final static void logTINEAttachedLink(
			final String clazz_, 
			final ITineAccessor accessor_, 
			final int linkId_)
	{
		if (linkId_ < 0)
			logTINELink(LogLvl.error, clazz_, "unable to established an asynchronous link", accessor_);
		else
			logTINELink(LogLvl.info, clazz_, "established asynchronous link on ID " + linkId_, accessor_);
	}
	
	/**
	 * Log information about the executed link.
	 * 
	 * @param clazz_ Class name, leading the displayed output.
	 * @param accessor_ TINE link.
	 */
	public final static void logTINEExecuteLink(
			final String clazz_, 
			final ITineAccessor accessor_)
	{
		logTINELink(LogLvl.info, clazz_, "execute synchronous link", accessor_);
	}
		
	/**
	 * Log information about a TINE link error.
	 * 
	 * @param clazz_ Class name, leading the displayed output.
	 * @param accessor_ TINE link.
	 */
	public final static void logTINEErrorLink(
			final String clazz_, 
			final ITineAccessor accessor_)
	{
		logTINELink(LogLvl.error, clazz_, "error on link", accessor_);
	}

	/**
	 * Logs information about the current TINE links.
	 * 
	 * @param lstAccessor_ Collection of TINE links.
	 */
	public final static void logTINELink(
			final ITineAccessor... lstAccessor_)
	{
		logTINELink(null, lstAccessor_);
	}

	/**
	 * Logs information about the current TINE links.
	 * 
	 * @param clazz_ Calling clazz name.
	 * @param lstAccessor_ Collection of TINE links.
	 */
	public final static void logTINELink(
			final String clazz_, 
			final ITineAccessor... lstAccessor_)
	{
		logTINELink(LogLvl.info, clazz_, null, lstAccessor_);
	}
	
	/**
	 * Logs information about the current TINE links.
	 * 
	 * @param logLevel_ MstApp logging level.
	 * @param clazz_ Calling clazz name.
	 * @param message_ Message.
	 * @param lstAccessor_ Collection of TINE accessor links.
	 */
	public final static void logTINELink(
			final LogLvl logLevel_, 
			final String clazz_, 
			final String message_, 
			final ITineAccessor... lstAccessor_)
	{
		if (clazz_ != null && !clazz_.isEmpty() && message_ != null && !message_.isEmpty())
			for(final ITineAccessor accessor: lstAccessor_)
				Logger.doLog(logLevel_, Logger.SWTine, String.format("%s - %s %s => %s (%d)", 
						clazz_, 
						message_,
						accessor.getLinkAddress().getFullPath(),
						TErrorList.getErrorString(accessor.getLinkStatus()),
						accessor.getLinkStatus()), null);
		else if (message_ != null && !message_.isEmpty())
			for(final ITineAccessor accessor: lstAccessor_)
				Logger.doLog(logLevel_, Logger.SWTine, String.format("[unspecified class] - %s %s => %s (%d)", 
						message_, 
						accessor.getLinkAddress().getFullPath(),
						TErrorList.getErrorString(accessor.getLinkStatus()),
						accessor.getLinkStatus()), null);
		else
			for(final ITineAccessor accessor: lstAccessor_)
				Logger.doLog(logLevel_, Logger.SWTine, String.format("[unspecified class] - %s => %s (%d)", 
						accessor.getLinkAddress().getFullPath(),
						TErrorList.getErrorString(accessor.getLinkStatus()),
						accessor.getLinkStatus()), null);
	}
}
/* end of file */