/*
 *
 * File: ACDIDevices.java
 * Date: Mar 20, 2007; 5:29:57 PM
 */
package de.desy.mst.libs.accelerator.model;

import static de.desy.mst.libs.accelerator.CGeneralConstant.DESCRIPTION_CHANNEL;
import static de.desy.mst.libs.accelerator.CGeneralConstant.DESCRIPTION_RECV;
import static de.desy.mst.libs.accelerator.CGeneralConstant.DESCRIPTION_SEND;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.SortedMap;

import de.desy.mst.libs.accelerator.CGeneralConstant;
import de.desy.mst.libs.accelerator.CounterGlobal;
import de.desy.mst.libs.accelerator.EName;
import de.desy.mst.libs.accelerator.EServiceElement;
import de.desy.mst.libs.accelerator.IECDIDevice;
import de.desy.mst.libs.accelerator.IEData;
import de.desy.mst.libs.accelerator.IEDataMpxChannel;
import de.desy.mst.libs.accelerator.IEDataPreference;
import de.desy.mst.libs.accelerator.IEDataSetValues;
import de.desy.mst.libs.accelerator.IEExpectList;
import de.desy.mst.libs.accelerator.exception.DeprecatedException;
import de.desy.mst.libs.accelerator.exception.EnumerationNotFoundException;
import de.desy.mst.libs.accelerator.exception.IllegalChannelException;
import de.desy.mst.libs.accelerator.exception.TineException;
import de.desy.mst.libs.accelerator.view.ComponentSubject;
import de.desy.mst.libs.framework.mstapp.logging.Logger;
import de.desy.mst.libs.ser.tclientlib.TConnectionMode;
import de.desy.mst.libs.ser.tclientlib.TineLinkException;
import de.desy.mst.libs.ser.tclientlib.readers.FloatArrayReader;
import de.desy.mst.libs.ser.tclientlib.readers.FloatReader;
import de.desy.mst.libs.ser.tclientlib.readers.NAME16Reader;
import de.desy.mst.libs.ser.tclientlib.readers.NAME64Reader;
import de.desy.tine.client.TLink;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TAccess;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.queryUtils.TQuery;
import de.desy.tine.types.ADDRESS;

/**
 * <p>This class represents all parameters of the corresponding CDI server or
 * the middle layer DB server. On a CDI server the
 * represented parameters can include the hardware address on SEDAC devices (i.e.
 * the crate and sub-address), the precision of the returned numerical device
 * (others then numerical devices are not sufficient for that), the conversion 
 * factor for receiving or sending data. A difference is for the interval ranges
 * and units, which depends on the current data element instance. When the data
 * element implements the interface of {@link IEDataPreference} then the interval
 * ranges and units are taken from the special DB middle layer server instead of
 * the CDI one.</p>   
 * <p>Note that for the CDI servers their exists a special meaning of the
 * property name, which are predefined by the common CDI device servers. For
 * example the property name of <tt>RECV</tt> or <tt>RECV.CLBR</tt> are predefined
 * by the CDI device server and represents the read back value, the last by
 * calculating the resulting value with a predefined expression.</p>
 *
 * @author <a href="marcus.walla(at)desy.de">walla</a>
 * @version $Id: ACDIDevices.java 55476 2013-06-19 16:17:37Z walla $
 */
public abstract class ACDIDevices implements ICDIDevice
{	
	// --- CLASS FIELDS --------------------------------------------------------
	
	/** Class name. */
	public final static String CLAZZ = ACDIDevices.class.getSimpleName();
	
	// --- INSTANCE FIELDS -----------------------------------------------------
	
	private int requestedTimeout = 1000;
	
	/** This is the preliminary TINE address array. */
	private String[] tineAddresses;

	/** This is a map which associates to each TINE preliminary address a list
	 * of device names. */
	private final Map<String, Map<String, Integer>> mapTINEDevices;

	/** This is a map which associates to each TINE preliminary address the
	 * number of all devices underneath it. */
	private final Map<String, Integer> mapNumberOfDevices;

	/** Is the map which has a key of the device number and the corresponding
	 * name as a value. */
	private final Map<String, Map<Integer, String>> mapNumbers;

	/** Is the map which has a key of the device number and the corresponding
	 * name as a value. */
	private final Map<String, Map<String, String>> mapUnits;

	/** Is the map which has a key of the device number and the corresponding
	 * name as a value. */
	private final Map<String, Map<String, String>> mapPrecisions;

	/** Is the map which has a key of the device number and the corresponding
	 * minimum range number as a value. */
	private final Map<String, Map<String, Float>> mapRangesMin;

	/** Is the map which has a key of the device number and the corresponding
	 * maximum range number as a value. */
	private final Map<String, Map<String, Float>> mapRangesMax;

	/** Is the map which has a key of the device number and the corresponding
	 * description as a string value. */
	private final Map<String, Map<String, String>> mapDescription;

	/** Is the map which has a key of the device number and the corresponding
	 * description of the receive value as a float. */
	private final Map<String, Map<String, Float>> mapRecvFactor;

	/** Is the map which has a key of the device number and the corresponding
	 * description of the send value as a float. */
	private final Map<String, Map<String, Float>> mapSendFactor;

	/** Is the map which has a key of the device number and the corresponding
	 * description of the channel as an integer. */
	private final Map<String, Map<String, Integer>> mapChannel;

	/** Is the map which has a key of the device number and the corresponding
	 * crate address number as a value. */
	private final Map<String, Map<String, Integer>> mapCrate;

	/** Is the map which has a key of the device number and the corresponding
	 * address number as a value. */
	private final Map<String, Map<String, Integer>> mapAddress;

	/** A collection of all devices which are defined in the concrete CDI
	 * table. The key element is the preliminary TINE address, i.e. the context
	 * and server. */
	private final Map<String, SortedMap<String, IEData>> mapAppDevices;

	// --- CONSTRUCTORS --------------------------------------------------------
	
	/**
	 * Constructor of this class. It initializes different
	 * maps and get all devices. A {@code RuntimeException} can be thrown by
	 * this constructor if an error occurred.
	 *
	 * @param device_ This enumeration interfaces includes all CDI devices
	 * 		which are registered to this instance.
	 */
	protected ACDIDevices(final IECDIDevice device_, final String[] tineAddresses_) {
		this(device_.getMap(), tineAddresses_);
	}

	/**
	 * Constructor of this class. It initialize different
	 * maps and get all devices. A {@code RuntimeException} can be thrown by
	 * this constructor if an error occurred.
	 *
	 * @param mapAppDevices_ Is the collection of devices which are defined in
	 * 		a CDI table. The first string element includes a list of the TINE
	 * 		addresses, where each TINE address has a map of the device name -
	 * 		data enumeration element type, which represents the device name.
	 * @param tineAddresses_ This is an array of all TINE preliminary address
	 * 		strings.
	 */
	protected ACDIDevices(final Map<String, SortedMap<String, IEData>> mapAppDevices_, final String[] tineAddresses_) {
		validate(mapAppDevices_, tineAddresses_);

		this.requestedTimeout = 1000;
		this.mapAppDevices = mapAppDevices_;
		this.tineAddresses = tineAddresses_;

		this.mapTINEDevices 	= new HashMap<String, Map<String, Integer>>();
		this.mapNumberOfDevices = new HashMap<String, Integer>();
		this.mapNumbers 		= new HashMap<String, Map<Integer, String>>();
		this.mapUnits 			= new HashMap<String, Map<String, String>>();
		this.mapPrecisions 		= new HashMap<String, Map<String, String>>();
		this.mapRangesMin 		= new HashMap<String, Map<String, Float>>();
		this.mapRangesMax 		= new HashMap<String, Map<String, Float>>();
		this.mapDescription 	= new HashMap<String, Map<String, String>>();
		this.mapRecvFactor 		= new HashMap<String, Map<String, Float>>();
		this.mapSendFactor 		= new HashMap<String, Map<String, Float>>();
		this.mapChannel 		= new HashMap<String, Map<String, Integer>>();
		this.mapCrate 			= new HashMap<String, Map<String, Integer>>();
		this.mapAddress 		= new HashMap<String, Map<String, Integer>>();
		
		for(final String tineAddress: tineAddresses)
			if (this.mapAppDevices.containsKey(tineAddress))
				for(final IEData data: this.mapAppDevices.get(tineAddress).values())
					ComponentSubject.getInstance().register(data);
			else
				throw new IllegalArgumentException("TINE address of " + tineAddress + " is not initialize in the delivered map of the sub constructor -> initialize the device names in this map!");
	}

	// --- INSTANCE METHODS ----------------------------------------------------

	@Override 
	public final void setTINETimeout(final int timeout) {
		this.requestedTimeout = (timeout < 1000) ? 1000 : ((timeout > 10000) ? 10000 : timeout);
	}
	
	@Override 
	public final int getTINETimeout() { return this.requestedTimeout; }
	
	@Deprecated
	public final void checkEnumElements() throws TineException, EnumerationNotFoundException {
		checkEnumElements(EApplicationModel.SERVER, null);
	}
	
	@Override
	public final void checkEnumElements(final EApplicationModel model_) throws TineException, EnumerationNotFoundException {
		checkEnumElements(model_, null);
	}

	@Deprecated
	public final void checkEnumElements(final String tineAddress_) throws TineException, EnumerationNotFoundException {
		checkEnumElements(EApplicationModel.SERVER, tineAddress_, true, true);
	}

	@Override
	public final void checkEnumElements(final EApplicationModel model_, final String tineAddress_) throws TineException, EnumerationNotFoundException {
		checkEnumElements(model_, tineAddress_, true, true);
	}

	@Deprecated
	public final void checkEnumElements(final String tineAddress_, final boolean checkEverytime_, final boolean initAllDevices_) throws TineException, EnumerationNotFoundException {
		checkEnumElements(EApplicationModel.SERVER, tineAddress_, checkEverytime_, initAllDevices_);
	}
	
	@Override
	public final void checkEnumElements(final EApplicationModel model_, final String tineAddress_, final boolean checkEverytime_, final boolean initAllDevices_)
		throws TineException, EnumerationNotFoundException {
		Logger.SWApp.info(CLAZZ + " - checking all device names on address " + tineAddress_ + " (model=" + model_ + ") ...");
		
		String[] tineInternalAddress;
		if (tineAddress_ != null || mapAppDevices.get(tineAddress_) != null)
			tineInternalAddress = new String[] {tineAddress_};
		else
			tineInternalAddress = tineAddresses;
	
		if (checkEverytime_) {
			refreshDevices(model_, initAllDevices_, tineInternalAddress);
		}
		
		if (mapTINEDevices.size() == 0 || mapNumbers.size() == 0)
			throw new TineException(TErrorList.code_failure, "Expected list of all devicesis null and this before calling method getDevices(..).");
	
		final StringBuffer buffer = checkInternalTineAddresses(tineInternalAddress);
	
		// If the buffer is longer then 0 throw the exception
		if (buffer.length() > 0)
			throw new EnumerationNotFoundException("Some enumerations cannot be found:\n" + buffer.toString());
	
		Logger.SWApp.debug(CLAZZ + " - successful checking of all enumeration elements: " + Arrays.toString(tineInternalAddress).toString());
	}
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	@Override
	public final String fetchDevice(final String tineAddress_, final int deviceNumber_) {
		if (!mapNumbers.containsKey(tineAddress_))
			throw new IllegalArgumentException("Device number of  prelimiary TINE address " + tineAddress_ + " doesn't exist.");
		return mapNumbers.get(tineAddress_).get(deviceNumber_);
	}

	@Override
	public final Collection<String> fetchDevices(final String tineAddress_) {
		if (mapTINEDevices.containsKey(tineAddress_))
			return mapTINEDevices.get(tineAddress_).keySet();
		throw new IllegalArgumentException("The delivered TINE address " + tineAddress_ + " doesn't have an associated entry.");
	}

	@Override
	public final int fetchNumber(final String tineAddress_, final String deviceName_) {
		if (mapTINEDevices.containsKey(tineAddress_) && mapTINEDevices.get(tineAddress_).containsKey(deviceName_))
			return mapTINEDevices.get(tineAddress_).get(deviceName_);
		throw new IllegalArgumentException("The delivered TINE address " + tineAddress_ + " doesn't have an associated entry.");
	}

	@Override
	public final int fetchNumberOfDevices() {
		return (mapNumberOfDevices.get("ALL") == null)?0:mapNumberOfDevices.get("ALL");
	}
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	
	@Deprecated
	public IExecute getExec(final IEExpectList exert_) {
		throw new DeprecatedException(CLAZZ);
	}
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	/**
	 * This method is deprecated. You should use {@link #refreshDevices(boolean, String...)}.
	 * 
	 * @param prefixTINEAddresses_ This is an array list of TINE
	 * 		CDI servers, which this method will get the required information from.
	 * @param initAllDevices_ Checks if all device elements should be initialized
	 * 		(<tt>true</tt>) or not (<tt>false</tt>). Is this boolean value 
	 * 		{@code false} then only the device names and numbers are initialize. If
	 * 		it set to {@code true} will initialize all device elements by getting the device
	 * 		names, device numbers, units, precision, minimum and max values (range),
	 * 		the addresses (crate and sub-address), and the description.
	 * @throws TineException A {@code TineException} is thrown if an error
	 * 		occurred on the TINE context (i.e. it is {@code null}), the current link
	 * 		is invalid or returns no result. Moreover there are also other
	 * 		runtime exception which can be thrown if the creating of the list
	 * 		fails.
	 * @see {@link #refreshDevices(boolean, String...)}
	 */
	@Deprecated
	public final synchronized void getDevices(final String[] prefixTINEAddresses_, final boolean initAllDevices_) throws TineException {
		refreshDevices(initAllDevices_, prefixTINEAddresses_);
	}
	
	/**
	 * This method is deprecated. You should use {@link #refreshDevices(boolean, String...)}.
	 * 
	 * @param prefixTINEAddresses_ This is an array list of TINE
	 * 		CDI servers, which this method will get the required information from.
	 * @param initAllDevices_ Checks if all device elements should be initialized
	 * 		(<tt>true</tt>) or not (<tt>false</tt>). Is this boolean value 
	 * 		{@code false} then only the device names and numbers are initialize. If
	 * 		it set to {@code true} will initialize all device elements by getting the device
	 * 		names, device numbers, units, precision, minimum and max values (range),
	 * 		the addresses (crate and sub-address), and the description.
	 * @throws TineException A {@code TineException} is thrown if an error
	 * 		occurred on the TINE context (i.e. it is {@code null}), the current link
	 * 		is invalid or returns no result. Moreover there are also other
	 * 		runtime exception which can be thrown if the creating of the list
	 * 		fails.
	 * @see {@link #refreshDevices(boolean, String...)}
	 */
	@Deprecated
	public final void refreshDevices(final String[] prefixTINEAddresses_, final boolean initAllDevices_) throws TineException {
		refreshDevices(initAllDevices_, prefixTINEAddresses_);
	}

	/**
	 * This method is <b>deprecated</b>. Instead use the method of {@link #refreshDevicesOnServer(boolean, String...)}.
	 * @param initAllDevices_ Checks if all device elements should be initialized
	 * 		(<tt>true</tt>) or not (<tt>false</tt>). Is this boolean value 
	 * 		{@code false} then only the device names and numbers are initialize. If
	 * 		it set to {@code true} will initialize all device elements by getting the device
	 * 		names, device numbers, units, precision, minimum and max values (range),
	 * 		the addresses (crate and sub-address), and the description.
	 * @param prefixTINEAddresses_ This is an array list of TINE
	 * 		CDI servers, which this method will get the required information from.
	 * @throws TineException A {@code TineException} is thrown if an error
	 * 		occurred on the TINE context (i.e. it is {@code null}), the current link
	 * 		is invalid or returns no result. Moreover there are also other
	 * 		runtime exception which can be thrown if the creating of the list
	 * 		fails.
	 * @see #refreshDevicesOnServer(boolean, String...)
	 */
	@Deprecated
	public final void refreshDevices(final boolean initAllDevices_, final String... prefixTINEAddresses_) throws TineException {
		refreshDevices(EApplicationModel.SERVER, initAllDevices_, prefixTINEAddresses_);
	}
	
	/**
	 * <p>Refresh all meta information of all devices by a given TINE context 
	 * and server (should be named in a TINE manner, i.e. separated by slashes) 
	 * for a middle layer server.</p>
	 * 
	 * </p>If an error occurred this method will throw a {@link TineException}. By 
	 * the context of all meta information I also mean that this function will 
	 * also refresh meta information about the units and precision and set it in 
	 * the require enumeration data.</p>
	 * <p>This connection is done via a TCP and not UDP.</p>
	 *
	 * @param model_ Depending on the current application model different meta
	 * 		values are read out from the CDI server. There currently exists two 
	 *      distinguish enumerations for updating and refreshing the meta 
	 *      information for all known TINE device names: one is for refreshing 
	 *      the <b>server</b> and the other is for <b>client</b> environment. 
	 *      <p>The server refreshing enumeration provides<ul>
	 *      <li>{@code numbers reflecting the device names} ({@link #refreshNumbersReflectingDeviceNames(String...)})</li>
	 *      <li>{@code interval ranges} ({@link #refreshIntervalRanges(String...)})</li>
	 *      <li>{@code SEDAC addresses} ({@link #refreshAddresses(String...)})</li>
	 *      <li>{@code description, send and receive factor, and multiplexer channel} ({@link #refreshDescriptions(String...)}).</li>
	 * 		<li>{@code interval ranges} ({@link #refreshIntervalRanges(String...)})</li>
	 *      </ul></p>
	 *      <p>The client refreshing enumeration provides<ul>
	 * 		<li>{@code numbers reflecting the device names} ({@link #refreshNumbersReflectingDeviceNames(String...)})</li>
	 * 		<li>{@code units} ({@link #refreshUnits(String...)})</li>
	 * 		<li>{@code precision} ({@link #refreshPrecisions(String...)})</li>
	 * 		<li>{@code SEDAC addresses} ({@link #refreshAddresses(String...)})</li>
	 * 		<li>{@code description, send and receive factor, and multiplexer channel} ({@link #refreshDescriptions(String...)})</li>
	 * 		</ul></p>
	 * @param initAllDevices_ Checks if all device elements should be initialized
	 * 		(<tt>true</tt>) or not (<tt>false</tt>). Is this boolean value 
	 * 		{@code false} then only the device names and numbers are initialize. If
	 * 		it set to {@code true} will initialize all device elements by getting the device
	 * 		names, device numbers, units, precision, minimum and max values (range),
	 * 		the addresses (crate and sub-address), and the description.
	 * @param prefixTINEAddresses_ This is an array list of TINE
	 * 		CDI servers, which this method will get the required information from.
	 * @throws TineException A {@code TineException} is thrown if an error
	 * 		occurred on the TINE context (i.e. it is {@code null}), the current link
	 * 		is invalid or returns no result. Moreover there are also other
	 * 		runtime exception which can be thrown if the creating of the list
	 * 		fails.
	 * @see #refreshDevicesOnClient(boolean, String...)
	 */
	public final void refreshDevices(final EApplicationModel model_, final boolean initAllDevices_, final String...prefixTINEAddresses_) throws TineException {
		switch(model_) {
		case CLIENT:
			refreshDevicesOnClient(initAllDevices_, prefixTINEAddresses_);
			break;
			
		case SERVER:
			refreshDevicesOnServer(initAllDevices_, prefixTINEAddresses_);
			break;
			
		default:
			refreshDevicesOnServer(initAllDevices_, prefixTINEAddresses_);
		}
	}

	/**
	 * <p>Refresh all meta information of all devices by a given TINE context 
	 * and server (should be named in a TINE manner, i.e. separated by slashes) 
	 * for a middle layer server.</p>
     *
	 * <p><b>ATTENTION</b>: There currently exists two distinguish methods for
	 * updating and refreshing the meta information for all known TINE device 
	 * names: one is for refreshing <b>server</b> and the other is
	 * for <b>client</b> environment, because both methods internally refreshes
	 * different capabilities of meta information. This server refreshing method 
	 * provides
	 * <b>numbers reflecting the device names</b> ({@link #refreshNumbersReflectingDeviceNames(String...)}), 
	 * <b>interval ranges</b> ({@link #refreshIntervalRanges(String...)}),
	 * <b>SEDAC addresses</b> ({@link #refreshAddresses(String...)}),
	 * and <b>description, send and receive factor, and multiplexer channel</b> ({@link #refreshDescriptions(String...)}).
	 * For the client environment see at the method {@link #refreshDevicesOnClient(boolean, String...)}.
	 * </p> 
	 * 
	 * </p>If an error occurred this method will throw a {@link TineException}. By 
	 * the context of all meta information I also mean that this function will 
	 * also refresh meta information about the units and precision and set it in 
	 * the require enumeration data.</p>
	 * <p>This connection is done via a TCP and not UDP.</p>
	 *
	 * @param initAllDevices_ Checks if all device elements should be initialized
	 * 		(<tt>true</tt>) or not (<tt>false</tt>). Is this boolean value 
	 * 		{@code false} then only the device names and numbers are initialize. If
	 * 		it set to {@code true} will initialize all device elements by getting the device
	 * 		names, device numbers, units, precision, minimum and max values (range),
	 * 		the addresses (crate and sub-address), and the description.
	 * @param prefixTINEAddresses_ This is an array list of TINE
	 * 		CDI servers, which this method will get the required information from.
	 * @throws TineException A {@code TineException} is thrown if an error
	 * 		occurred on the TINE context (i.e. it is {@code null}), the current link
	 * 		is invalid or returns no result. Moreover there are also other
	 * 		runtime exception which can be thrown if the creating of the list
	 * 		fails.
	 * @see #refreshDevicesOnClient(boolean, String...)
	 */
	private void refreshDevicesOnServer(final boolean initAllDevices_, final String... prefixTINEAddresses_) throws TineException {
		refreshDeviceNames(prefixTINEAddresses_);

		// after that call more special methods:
		refreshNumbersReflectingDeviceNames(prefixTINEAddresses_);
		if (initAllDevices_) {
			refreshIntervalRanges(prefixTINEAddresses_);
			refreshAddresses(prefixTINEAddresses_);
			refreshDescriptions(prefixTINEAddresses_);
		}
	}

	/**
	 * <p>Refresh all meta information of all devices by a given TINE context 
	 * and server (should be named in a TINE manner, i.e. separated by slashes) 
	 * for a middle layer server.</p>
     *
	 * <p><b>ATTENTION</b>: There currently exists two distinguish methods for
	 * updating and refreshing the meta information for all known TINE device 
	 * names: one is for refreshing <b>server</b> and the other is
	 * for <b>client</b> environment, because both methods internally refreshes
	 * different capabilities of meta information. This client refreshing method 
	 * provides
	 * <b>numbers reflecting the device names</b> ({@link #refreshNumbersReflectingDeviceNames(String...)}), 
	 * <b>units</b> ({@link #refreshUnits(String...)}),
	 * <b>precision</b> ({@link #refreshPrecisions(String...)}),
	 * <b>SEDAC addresses</b> ({@link #refreshAddresses(String...)}),
	 * <b>interval ranges<b> ({@link #refreshIntervalRanges(String...)}),
	 * and <b>description, send and receive factor, and multiplexer channel</b> ({@link #refreshDescriptions(String...)}).
	 * For the server environment see at the method {@link #refreshDevicesOnServer(boolean, String...)}.
	 * </p> 
	 * 
	 * </p>If an error occurred this method will throw a {@link TineException}. By 
	 * the context of all meta information I also mean that this function will 
	 * also refresh meta information about the units and precision and set it in 
	 * the require enumeration data.</p>
	 * <p>This connection is done via a TCP and not UDP.</p>
	 * 
	 * @param initAllDevices_ Checks if all device elements should be initialized
	 * 		(<tt>true</tt>) or not (<tt>false</tt>). Is this boolean value 
	 * 		{@code false} then only the device names and numbers are initialize. If
	 * 		it set to {@code true} will initialize all device elements by getting the device
	 * 		names, device numbers, units, precision, minimum and max values (range),
	 * 		the addresses (crate and sub-address), the interval, and the description.
	 * @param prefixTINEAddresses_ This is an array list of TINE
	 * 		CDI servers, which this method will get the required information from.
	 * @throws TineException A {@code TineException} is thrown if an error
	 * 		occurred on the TINE context (i.e. it is {@code null}), the current link
	 * 		is invalid or returns no result. Moreover there are also other
	 * 		runtime exception which can be thrown if the creating of the list
	 * 		fails.
	 * @see #refreshDevicesOnServer(boolean, String...)
	 */
	private void refreshDevicesOnClient(final boolean initAllDevices_, final String... prefixTINEAddresses_) throws TineException {
		refreshDeviceNames(prefixTINEAddresses_);

		// after that call more special methods:
		refreshNumbersReflectingDeviceNames(prefixTINEAddresses_);
		if (initAllDevices_) {
			refreshIntervalRanges(prefixTINEAddresses_);
			refreshUnits(prefixTINEAddresses_);
			refreshPrecisions(prefixTINEAddresses_);
			refreshAddresses(prefixTINEAddresses_);
			refreshDescriptions(prefixTINEAddresses_);
		}
	}

	/**
	 * Refresh the device names of a TINE device server.
	 * 
	 * @param prefixTINEAddresses_ This is an array list of TINE
	 * 		CDI servers, which this method will get the required information from.
	 * @throws TineException A {@code TineException} is thrown if an error
	 * 		occurred on the TINE context (i.e. it is {@code null}), the current link
	 * 		is invalid or returns no result. Moreover there are also other
	 * 		runtime exception which can be thrown if the creating of the list
	 * 		fails.
	 */
	private void refreshDeviceNames(final String... prefixTINEAddresses_) throws TineException {
		if (mapNumberOfDevices.get("ALL") == null || mapNumberOfDevices.get("ALL") <= 0) {
			try {
				getNumberOfDevices();
			}
			catch(final TineException exception) {
				throw new TineException(exception.getErrorCode(), "Reading the number of all devices on all servers failed.", exception);
			}
		}

		for (final String tineAddress: prefixTINEAddresses_) {
			if (mapNumberOfDevices.get(tineAddress) == null)
				throw new TineException(TErrorList.code_failure, "TINE address of " + tineAddress + " has no elements (possibly it doesn't" + " exists).");
			if(mapNumberOfDevices.get(tineAddress) <= 0)
				throw new TineException(TErrorList.code_failure, "TINE address " + tineAddress + " seemed to have no elements: " + mapNumberOfDevices.get(tineAddress) );

			// fetch the TINE devices on the particular TINE context and server
			final Map<String, Integer> mapTINEDevice = new HashMap<String, Integer>(mapNumberOfDevices.get(tineAddress));
			mapTINEDevices.put(tineAddress, mapTINEDevice);

			String[] elements = tineAddress.split("/");
			if (elements == null || elements.length != 3 || elements[0] == null || elements[1] == null || elements[2] == null || !elements[0].isEmpty() || elements[1].isEmpty() || elements[2].isEmpty())
				throw new TineException(TErrorList.code_failure, "Fetching TINE preliminary address fails: " + tineAddress);
			
			// get all device names ...
			final String[] deviceNames = TQuery.getDeviceNames(elements[1], elements[2]);

			for(final String deviceName: deviceNames)
				mapTINEDevice.put(deviceName, null);
		}
	}
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	private StringBuffer checkInternalTineAddresses(final String... prefixTINEAddresses_) {
		final StringBuffer buffer = new StringBuffer();

		for(final String prefixTINEAddress: prefixTINEAddresses_) {
			Map<String, Integer> mapTINEDevices = this.mapTINEDevices.get(prefixTINEAddress);
			Map<String, IEData> setAppDevices = mapAppDevices.get(prefixTINEAddress);
			for(String element: setAppDevices.keySet()) {
				if (!mapTINEDevices.containsKey(element))
					buffer.append("Enumeration element " + element + " (" + prefixTINEAddress + ") hasn't a TINE element!\n");
			}
		}

		return buffer;
	}

	/**
	 * Returns the number of all devices on the device context and device server.
	 * The queried number is request from a TINE meta information of the device 
	 * server and has nothing to do with a special device or property name. 
	 * Under TINE this is called a <b>stock property</b>.
	 *
	 * @return Returns the number of all devices for the initialized TINE 
	 * 		addresses.
	 * @throws TineException Is thrown if an error occurred on executing the
	 * 		TINE link.
	 */
	private int getNumberOfDevices() throws TineException {
		if (this.mapNumberOfDevices.get("ALL") != null && this.mapNumberOfDevices.get("ALL") > 0)
			return this.mapNumberOfDevices.get("ALL");

		int totalNumberOfDevices = 0;

		for (final String tineAddress: tineAddresses) {
			final int[] numberOfDevices = new int[1];
			final TDataType output = new TDataType(numberOfDevices);

			TLink link = null;
			try {
				// The following connection via TINE is done with TCP/IP via the
				// TAccess.CA_CONNECT flag.
				link = new TLink(tineAddress, ETineProperty.NDEVICES.getName(EName.PROPERTY), output, null, (short) (TAccess.CA_READ | TAccess.CA_CONNECT));
				final int cc = link.execute(this.requestedTimeout);
				if ( cc != 0)
					throw new TineException(cc, "Reading number of devices on " + tineAddress + " failed: "	+ TErrorList.getErrorString(cc) + " (" + cc + ") - may be the CDI server is not running?");
			}
			catch(final RuntimeException exception) {
				Logger.SWApp.error(CLAZZ + " - TLink exception: " + tineAddress + '/' + ETineProperty.NDEVICES.getName(EName.PROPERTY) + " with access READ doesn't exist.");
				throw new TineException(-1, tineAddress + '/' + ETineProperty.NDEVICES.getName(EName.PROPERTY) + " doesn't exist");
			}
			finally { if (link != null) link.close(); }

			totalNumberOfDevices += numberOfDevices[0];
			this.mapNumberOfDevices.put(tineAddress, numberOfDevices[0]);
		}

		this.mapNumberOfDevices.put("ALL", totalNumberOfDevices);
		Logger.SWApp.debug(CLAZZ + " - successful getting the number of all devices: " + mapNumberOfDevices.get("ALL"));
		return totalNumberOfDevices;
	}

	/**
	 * Gets the number of each TINE device over a TINE link for each enumeration element.
	 * This connection is done via a TCP/IP and not with UDP.
	 *
	 * @param prefixTINEAddresses_ TINE prefix address.
	 * @throws TineException A TINE exception is thrown if an error occurred on
	 * 		the network connection.
	 */
	private void refreshNumbersReflectingDeviceNames(final String... prefixTINEAddresses_) throws TineException {
		checkNumbers();
		int cc;

		for(String prefixTINEAddress: prefixTINEAddresses_) {
			final Map<Integer, String > mapNumbers = new HashMap<Integer, String>(this.mapNumberOfDevices.get(prefixTINEAddress));
			this.mapNumbers.put(prefixTINEAddress, mapNumbers);

			final Map<String, Integer> mapTINEDevices = this.mapTINEDevices.get(prefixTINEAddress);

			// TODO: here you should write the functionality of getting all CDI
			// numbers for the corresponding device name in ONE TINE link and not
			// like this way for EACH device in ONE TINE link (i.e. with a FOR loop!)
			for(final String deviceName: mapTINEDevices.keySet()) {
				final ADDRESS[] address = {new ADDRESS()};

				int current = 0;
				while(true) {
					TLink link = null;

					try {
						// The following connection via TINE is done with TCP/IP
						// via TAccess.CA_CONNECT flag.
						link = new TLink(prefixTINEAddress + '/' + deviceName, ETineProperty.NUMBER.getName(EName.PROPERTY), new TDataType(address), null, (short) (TAccess.CA_READ | TAccess.CA_CONNECT));

						// HINT: The execution of a TINE link can possibly also returned a value
						// which has the TErrorList.CE_SENDDATA set. This is not an error code.
						// It is rather a property about the link itself. For more information
						// you should contact Phil. There also exists in the Instant Client
						// application under the option Option - Property Query Precedence
						// a possible hint about its behavior.
						cc = link.execute(this.requestedTimeout);
						if ( cc != 0 && !((cc & TErrorList.CE_SENDDATA) == TErrorList.CE_SENDDATA) && current++ > CGeneralConstant.loopCDIArguments()) {
							throw new TineException(cc, "Execution of TINE" + " link " + prefixTINEAddress + '/' + deviceName + '/' + ETineProperty.NUMBER.getName(EName.PROPERTY) + " failed: " + TErrorList.getErrorString(cc));
						}
						else if (cc == 0) {
							mapNumbers.put(address[0].i1val, deviceName);
							mapTINEDevices.put(deviceName, address[0].i1val);
	
							link.close();
	
							// exit the while loop!
							break;
						}
					}
					catch(final RuntimeException exception) {
						Logger.SWApp.error(CLAZZ + " - error on initializing TINE link " + prefixTINEAddress + '/' + deviceName + '/' + ETineProperty.NUMBER.getName(EName.PROPERTY));
						throw new TineException(-1, "error on initializing TINE link " + prefixTINEAddress + '/' + deviceName + '/' + ETineProperty.NUMBER.getName(EName.PROPERTY));
					}
					finally { if (link != null) link.close(); }
				}
			}
		}

		Logger.SWApp.debug(CLAZZ + " - successful getting all device numbers: " + Arrays.toString(prefixTINEAddresses_));
	}

	/**
	 * Checks the number map. If it doesn't exist an {@code IllegalArgumentException} 
	 * is thrown.
	 */
	private void checkNumbers() {
		if (this.mapNumbers == null)
			throw new IllegalArgumentException("internal map of numbers for each CDI device is null.");
	}
	
	/**
	 * Gets the units over a TINE link for each enumeration element.
	 * This connection is done via a TCP/IP and not with UDP.
	 *
	 * @param prefixTINEAddresses_ TINE prefix addresses.
	 * @throws TineException A TINE exception is thrown if an error occurred on
	 * 		the network connection.
	 */
	private void refreshUnits(final String... prefixTINEAddresses_) throws TineException {
		checkUnits();
		
		for(final String prefixTINEAddress: prefixTINEAddresses_) {
			final Map<String, IEData> devices = this.mapAppDevices.get(prefixTINEAddress);

			final Map<String, String> mapUnits = new HashMap<String, String>();
			this.mapUnits.put(prefixTINEAddress, mapUnits);
	
			// TODO: here you should write the functionality of getting all CDI
			// numbers for the corresponding device name in ONE TINE link and not
			// like this way for EACH device in ONE TINE link (i.e. with a FOR loop!)
			for(String deviceName: this.mapTINEDevices.get(prefixTINEAddress).keySet()) {
				CounterGlobal.reset();
				
				while(true) {
					final NAME16Reader reader;
					
					if (devices.get(deviceName) instanceof IEDataPreference && ((IEDataPreference) devices.get(deviceName)).getPreferenceElement(EServiceElement.UNIT) != null) {
						final IEData dataUnit =  ((IEDataPreference) devices.get(deviceName)).getPreferenceElement(EServiceElement.UNIT);
						reader = new NAME16Reader(dataUnit.getName(EName.FULL_TINE_ADDRESS));
					}
					else {
						reader = new NAME16Reader(prefixTINEAddress + '/' + deviceName + '/' + ETineProperty.UNITS.getName(EName.PROPERTY));
					}
					
					// applying the TCP/IP connection ...
					reader.setConnectionMode(TConnectionMode.STREAM);
					reader.setTimeoutMs(this.requestedTimeout);
					
					try {
						mapUnits.put(deviceName, reader.getValue().getName());
	
						if (!mapUnits.get(deviceName).isEmpty() && devices.containsKey(deviceName))
							devices.get(deviceName).putName(EName.UNIT, mapUnits.get(deviceName));
						
						break;
					}
					catch (final TineLinkException exception) {
						if (CounterGlobal.isFinished()) {
							throw new TineException(reader.getLinkStatus(), "Execution of TINE link " + reader.getLinkAddress().getFullPath() + " failed:");
						}
					}
					finally { if (reader != null) reader.close(); }
					
					CounterGlobal.increment();
				}
			}
		}
		
		Logger.SWApp.debug(CLAZZ + " - successful getting all device units: " + Arrays.toString(prefixTINEAddresses_));
	}

	/**
	 * Checks the unit map. If it doesn't exist an {@code IllegalArgumentException} 
	 * is thrown.
	 */
	private void checkUnits() {
		if (this.mapUnits == null)
			throw new IllegalArgumentException("Internal unit map for" + " the devices is null.");
	}
	
	/**
	 * Gets the interval ranges over the predefined TINE prefix addresses.
	 * This is done via a TCP/IP and not UDP connection. If loading the interval
	 * ranges from a data elements fails this method will perform a check if
	 * the data element is an instance of {@link IEDataPerferences}. If yes it 
	 * sets the interval to the default values of {@code [0:1000]} before
	 * a {@link TINEException} is thrown. If not the interval ranges stay 
	 * untouched and it directly throws a {@link TineException}.
	 * 
	 * @param prefixTINEAddresses_ TINE prefix addresses.
	 * @throws TineException A TINE exception is thrown if an error occurred on
	 * 		the TINE connection.
	 */
	private void refreshIntervalRanges(final String... prefixTINEAddresses_) throws TineException {
		checkIntervalRanges();
	
		for(final String prefixTINEAddress: prefixTINEAddresses_) {
			final Map<String, IEData> devices = this.mapAppDevices.get(prefixTINEAddress);
	
			final Map<String, Float> mapRangesMin = new HashMap<String, Float>();
			this.mapRangesMin.put(prefixTINEAddress, mapRangesMin);
	
			final Map<String, Float> mapRangesMax = new HashMap<String, Float>();
			this.mapRangesMax.put(prefixTINEAddress, mapRangesMax);
	
			// TODO: here you should write the functionality of getting all CDI
			// numbers for the corresponding device name in ONE TINE link and not
			// like this way for EACH device in ONE TINE link (i.e. with a FOR loop!)
			for(final String deviceName: devices.keySet()) {
				CounterGlobal.reset();
				
				while(true) {
					if (devices.get(deviceName) instanceof IEDataPreference && ((IEDataPreference) devices.get(deviceName)).getPreferenceElement(EServiceElement.MAXIMUM) != null && ((IEDataPreference) devices.get(deviceName)).getPreferenceElement(EServiceElement.MINIMUM) != null) {
						final IEData dataMin = 
							((IEDataPreference) devices.get(deviceName)).getPreferenceElement(EServiceElement.MINIMUM);
						final IEData dataMax =
							((IEDataPreference) devices.get(deviceName)).getPreferenceElement(EServiceElement.MAXIMUM);
						
						final FloatReader readerMin = new FloatReader(dataMin.getName(EName.FULL_TINE_ADDRESS));
						// applying the TCP/IP connection ...
						readerMin.setConnectionMode(TConnectionMode.STREAM);
						readerMin.setTimeoutMs(this.requestedTimeout);

						final FloatReader readerMax = new FloatReader(dataMax.getName(EName.FULL_TINE_ADDRESS)); 
						// applying the TCP/IP connection ...
						readerMax.setConnectionMode(TConnectionMode.STREAM);
						readerMax.setTimeoutMs(this.requestedTimeout);
						
						try {
 							mapRangesMin.put(deviceName, readerMin.getValue());
							mapRangesMax.put(deviceName, readerMax.getValue());

							if ((mapRangesMin.get(deviceName) != Float.NaN || mapRangesMax.get(deviceName) != Float.NaN) && devices.containsKey(deviceName) && devices.get(deviceName) instanceof IEDataSetValues) {
								((IEDataSetValues) devices.get(deviceName)).putMinimum(mapRangesMin.get(deviceName));
								((IEDataSetValues) devices.get(deviceName)).putMaximum(mapRangesMax.get(deviceName));
							}
							
							break;
						}
						catch(final TineLinkException exception) {
							if (CounterGlobal.isFinished()) {
								if (readerMin.getLinkStatus() != 0)
									throw new TineException(readerMin.getLinkStatus(), "Execution of TINE link " +  readerMin.getLinkAddress().getFullPath() + " failed -> " + TErrorList.getErrorString(readerMin.getLinkStatus()));
								else
									throw new TineException(readerMax.getLinkStatus(), "Execution of TINE link " + readerMax.getLinkAddress().getFullPath() + " failed -> " + TErrorList.getErrorString(readerMax.getLinkStatus()));
							}
						}
						finally {
							if (readerMin != null) readerMin.close();
							if (readerMax != null) readerMax.close();
						}
					}
					else {
						final FloatArrayReader readerRange = new FloatArrayReader(prefixTINEAddress + '/' + deviceName + '/' + ETineProperty.RANGE.getName(EName.PROPERTY), 2);
						readerRange.setTimeoutMs(this.requestedTimeout);
		
						try {
							final float[] range = readerRange.getValues();
							mapRangesMin.put(deviceName, range[0]);
							mapRangesMax.put(deviceName, range[1]);

							if ((mapRangesMin.get(deviceName) != Float.NaN || mapRangesMax.get(deviceName) != Float.NaN) && devices.containsKey(deviceName) && devices.get(deviceName) instanceof IEDataSetValues) {
								((IEDataSetValues) devices.get(deviceName)).putMinimum(mapRangesMin.get(deviceName));
								((IEDataSetValues) devices.get(deviceName)).putMaximum(mapRangesMax.get(deviceName));
							}
							
							break;
						}
						catch(final TineLinkException exception) {
							if (CounterGlobal.isFinished())
								throw new TineException(readerRange.getLinkStatus(), "Execution of TINE link " + prefixTINEAddress + '/' + deviceName + '/' + ETineProperty.RANGE.getName(EName.PROPERTY) + " failed: " + TErrorList.getErrorString(exception.getErrorCode()));
						}
						finally { if (readerRange != null) readerRange.close(); }
					}
					
					CounterGlobal.increment();
				}
			}
		}
	
		Logger.SWApp.debug(CLAZZ + " - successful getting all interval ranges: " + Arrays.toString(prefixTINEAddresses_));
	}

	@SuppressWarnings("unused")
	private void refreshDefaultIntervalRanges(final String... prefixTINEAddresses_) {
		checkIntervalRanges();
		
		for(final String prefixTINEAddress: prefixTINEAddresses_) {
			final Map<String, IEData> devices = mapAppDevices.get(prefixTINEAddress);
			
			final Map<String, Float> mapRangesMin;
			if (this.mapRangesMin.containsKey(prefixTINEAddress))
				mapRangesMin = this.mapRangesMin.get(prefixTINEAddress);
			else {
				mapRangesMin = new HashMap<String, Float>();
				this.mapRangesMin.put(prefixTINEAddress, mapRangesMin);
			}
	
			final Map<String, Float> mapRangesMax;
			if (this.mapRangesMax.containsKey(prefixTINEAddress))
				mapRangesMax = this.mapRangesMax.get(prefixTINEAddress);
			else {
				mapRangesMax = new HashMap<String, Float>();
				this.mapRangesMax.put(prefixTINEAddress, mapRangesMax);
			}
	
			for(final String deviceName: devices.keySet()) {
				if (devices.get(deviceName) instanceof IEDataSetValues && devices.get(deviceName) instanceof IEDataPreference) {
					((IEDataSetValues) devices.get(deviceName)).putMinimum(0.0f);
					((IEDataSetValues) devices.get(deviceName)).putMaximum(1000.0f);
				}
			}
		}
		
		Logger.SWApp.info(CLAZZ + " - successful setting the default interval ranges from data elements of IEDataPreferences to [0:1000]");
	}

	/**
	 * Checks the range maps. If they doesn't exist an {@code IllegalArgumentException} 
	 * is thrown.
	 */
	private void checkIntervalRanges() {
		if (this.mapRangesMin == null)
			throw new IllegalArgumentException("Internal minimal range map of the devices is null.");
		if (this.mapRangesMax == null)
			throw new IllegalArgumentException("Internal maximal range map of the devices is null.");
	}
	
	/**
	 * Provides the precision over a TINE link for each enumeration element.
	 * This connection is done via a TCP/IP and not UDP.
	 *
	 * @param prefixTINEAddresses_ TINE prefix addresses.
	 * @throws TineException A TINE exception is thrown if an error occurred on
	 * 		the network connection.
	 */
	@SuppressWarnings("unused")
	private void refreshPrecisions(final String... prefixTINEAddresses_) throws TineException {
		checkPrecisions();

		for(final String prefixTINEAddress: prefixTINEAddresses_) {
			final Map<String, IEData> devices = this.mapAppDevices.get(prefixTINEAddress);

			final Map<String, String> mapPrecisions = new HashMap<String, String>();
			this.mapPrecisions.put(prefixTINEAddress, mapPrecisions);

			// TODO: here you should write the functionality of getting all CDI
			// numbers for the corresponding device name in ONE TINE link and not
			// like this way for EACH device in ONE TINE link (i.e. with a FOR loop!)
			for(String deviceName: devices.keySet()) {
				CounterGlobal.reset();
				
				while(true) {
					final NAME16Reader reader = new NAME16Reader(prefixTINEAddress + '/' + deviceName + '/' + ETineProperty.PRECISION.getName(EName.PROPERTY));
					// applying the TCP/IP connection ...
					reader.setConnectionMode(TConnectionMode.STREAM);
					reader.setTimeoutMs(this.requestedTimeout);
					
					try {
						mapPrecisions.put(deviceName, reader.getValue().getName());
						
						if (!mapPrecisions.get(deviceName).isEmpty() && devices.containsKey(deviceName))
							devices.get(deviceName).putName(EName.PRECISION, mapPrecisions.get(deviceName));
						
						break;
					}
					catch(final TineLinkException exception) {
						if (CounterGlobal.isFinished())
							throw new TineException(reader.getLinkStatus(), "Execution of TINE link " + prefixTINEAddress + " on " + deviceName + " failed");
					}
					finally { if (reader != null) reader.close(); }
					
					CounterGlobal.increment();
				}
			}
		}

		Logger.SWApp.debug(CLAZZ + " - successful getting all device precisions: " + Arrays.toString(prefixTINEAddresses_));
	}

	/**
	 * Checks the precision map. If it doesn't exist an {@code IllegalArgumentException} 
	 * is thrown.
	 */
	private void checkPrecisions() {
		if (this.mapPrecisions == null)
			throw new IllegalArgumentException("Internal precision map of the devices is null.");
	}
	
	/**
	 * Gets the description over a TINE link for all enumeration elements and 
	 * applying it. Moreover the send and receive factor and the multiplexer 
	 * channel of a TINE property is also refactored from the description string
	 * if it is recognized by a CDI server. This connection is done via a TCP/IP
	 * and not UDP.
	 *
	 * @param prefixTINEAddresses_ TINE prefix addresses.
	 * @throws TineException A TINE exception is thrown if an error occurred on
	 * 		the TINE connection.
	 */
	private void refreshDescriptions(final String... prefixTINEAddresses_) throws TineException {
		checkDescription();

		for(String prefixTINEAddress: prefixTINEAddresses_) {
			Map<String, IEData> mapAppDevices = this.mapAppDevices.get(prefixTINEAddress);

			Map<String, String> mapDescription = new HashMap<String, String>();
			this.mapDescription.put(prefixTINEAddress, mapDescription);
			Map<String, Float> mapRecvFactor = new HashMap<String, Float>();
			this.mapRecvFactor.put(prefixTINEAddress, mapRecvFactor);
			Map<String, Float> mapSendFactor = new HashMap<String, Float>();
			this.mapSendFactor.put(prefixTINEAddress, mapSendFactor);
			Map<String, Integer> mapChannel = new HashMap<String, Integer>();
			this.mapChannel.put(prefixTINEAddress, mapChannel);

			// TODO: here you should write the functionality of getting all CDI
			// numbers for the corresponding device name in ONE TINE link and not
			// like this way for EACH device in ONE TINE link (i.e. with a FOR loop!)
			for(final String deviceName: mapTINEDevices.get(prefixTINEAddress).keySet()) {
				int count = 0;
				while(true) {
					final NAME64Reader reader = new NAME64Reader(prefixTINEAddress + '/' + deviceName
							+ '/' +	ETineProperty.DESCRIPTION.getName(EName.PROPERTY));
					// applying the TCP/IP connection ...
					reader.setConnectionMode(TConnectionMode.STREAM);
					reader.setTimeoutMs(this.requestedTimeout);
					
					try {
						final String description = reader.getValue().getName();
						
						if (!description.isEmpty()) {
							mapDescription.put(deviceName, description);
							final String value[] = description.split(" ");
							
							for(final String current: value) {
								try {
									final String[] element = current.split("=");
									if (element.length == 2 && element[0].equals(DESCRIPTION_RECV)) {
										mapRecvFactor.put(deviceName, Float.parseFloat(element[1]));
										if (mapAppDevices.get(deviceName) instanceof IEDataSetValues)
											((IEDataSetValues) mapAppDevices.get(deviceName)).putRecvFactor( Float.parseFloat(element[1]));
									}
									else if(element.length == 2 && element[0].equals(DESCRIPTION_SEND)) {
										mapSendFactor.put(deviceName, Float.parseFloat(element[1]));
										if (mapAppDevices.get(deviceName) instanceof IEDataSetValues)
											((IEDataSetValues) mapAppDevices.get(deviceName)).putSendFactor( Float.parseFloat(element[1]));
									}
									else if(element.length == 2 && element[0].equals(DESCRIPTION_CHANNEL)) {
										mapChannel.put(deviceName, Integer.parseInt(element[1]));
										if (mapAppDevices.get(deviceName) instanceof IEDataMpxChannel)
											((IEDataMpxChannel) mapAppDevices.get(deviceName)).setChannel(Integer.parseInt(element[1]));
										else
											Logger.SWApp.warn(CLAZZ + " - data element of " + mapAppDevices.get(deviceName) + " isn't an instance of " + IEDataMpxChannel.class.getSimpleName());
									}
								}
								catch(IllegalChannelException exception) {
									Logger.SWApp.error(CLAZZ + " - failed to set channel" + " on " + mapAppDevices.get(deviceName).toString() + " :" + exception.getChannel());
								}
								catch(NumberFormatException exception) {
									Logger.SWApp.error(CLAZZ + " - failed to parse string" + " on " + mapAppDevices.get(deviceName).toString() + " into float/integer: " + current);
								}
							}
						}
						reader.close();
						break;
					}
					catch(final TineLinkException exception) {
						if (count++ > CGeneralConstant.loopCDIArguments())
							throw new TineException(-1, "The execution of the TINE link on " + deviceName + " failed (CDI column DESCRIPTION).");
					}
					reader.close();
				}
			}
		}

		Logger.SWApp.debug(CLAZZ + " - successful getting all describtions and possibly the send and" + " receive conversion factors: " + Arrays.toString(prefixTINEAddresses_));
	}

	/**
	 * Checks the description map. If it doesn't exist an {@code IllegalArgumentException} 
	 * is thrown.
	 */
	private void checkDescription() {
		if (this.mapDescription == null)
			throw new IllegalArgumentException("Internal description map of the devices is null.");
	}
	
	/**
	 * Gets all addresses over a TINE link for all enumeration elements and
	 * sets it. This is only responsible for such data elements which directly
	 * talk to a SEDAC interface instead of a TINE one.
	 * This connection is done via a TCP/IP and not UDP.
	 *
	 * @param prefixTINEAddresses_ TINE prefix addresses.
	 * @throws TineException A TINE exception is thrown if an error occurred on
	 * 		the TINE connection.
	 */
	private void refreshAddresses(final String... prefixTINEAddresses_) throws TineException {
		checkCrateAddresses();
		int cc = 0;

		for(final String prefixTINEAddress: prefixTINEAddresses_) {
			final Map<String, IEData> mapAppDevices = this.mapAppDevices.get(prefixTINEAddress);

			final Map<String, Integer> mapCrate = new HashMap<String, Integer>(this.mapNumberOfDevices.get(prefixTINEAddress));
			this.mapCrate.put(prefixTINEAddress, mapCrate);

			final Map<String, Integer> mapAddress = new HashMap<String, Integer>(this.mapNumberOfDevices.get(prefixTINEAddress));
			this.mapAddress.put(prefixTINEAddress, mapAddress);

			// TODO: here you should write the functionality of getting all CDI
			// numbers for the corresponding device name in ONE TINE link and not
			// like this way for EACH device in ONE TINE link (i.e. without a loop!)
			for(final String element: mapTINEDevices.get(prefixTINEAddress).keySet()) {
				final ADDRESS[] address = {new ADDRESS()};
				int current = 0;
				
				while(true) {
					TLink link = null;
					
					try {
						// The following connection via TINE is done with TCP/IP
						// via TAccess.CA_CONNECT flag.
						link = new TLink(prefixTINEAddress + '/' + element, ETineProperty.ADDRESS.getName(EName.PROPERTY), new TDataType(address), null, (short) (TAccess.CA_READ | TAccess.CA_CONNECT));

						// HINT: The execution of a TINE link can possibly also returned a value
						// which has the TErrorList.CE_SENDDATA set. This is not an error code.
						// It is rather a property about the link itself. For more information
						// you should contact Phil. There also exists in the Instant Client
						// application under the option Option - Property Query Precedence
						// a possible hint about its behavior.
						cc = link.execute(this.requestedTimeout);

						if ( cc != 0 && !((cc & TErrorList.CE_SENDDATA) == TErrorList.CE_SENDDATA) && current++ > CGeneralConstant.loopCDIArguments()) {
							throw new TineException(cc, "Execution of TINE link " + prefixTINEAddress + '/' + element + '/' + ETineProperty.ADDRESS.getName(EName.PROPERTY) + " failed: " + TErrorList.getErrorString(cc));
						}
						else if (cc == 0) {
							mapCrate.put(element, address[0].i2val);
							mapAddress.put(element, address[0].i3val);
							
							if (mapAppDevices.containsKey(element) && mapAppDevices.get(element) instanceof IEDataSetValues) {
								((IEDataSetValues) mapAppDevices.get(element)).putCrate(address[0].i2val);
								((IEDataSetValues) mapAppDevices.get(element)).putAddress(address[0].i3val);
							}
							
							link.close();
							break;
						}
					}
					catch(final RuntimeException exception) {
						Logger.SWApp.error(CLAZZ + " - error on initializing TINE link " + prefixTINEAddress + '/' + element + '/' + ETineProperty.ADDRESS.getName(EName.PROPERTY));
						throw new TineException(TErrorList.code_failure, "error on initializing TINE link " + prefixTINEAddress + '/' + element + '/' + ETineProperty.ADDRESS.getName(EName.PROPERTY));
					}
					finally { if (link != null) link.close(); }
				}
			}
		}
		
		Logger.SWApp.debug(CLAZZ + " - successful getting all device addresses (crate and subaddress): " + Arrays.toString(prefixTINEAddresses_));
	}

	/**
	 * Checks the crate address map. If it doesn't exist an {@code IllegalArgumentException} 
	 * is thrown.
	 */
	private void checkCrateAddresses() {
		if (this.mapCrate == null)
			throw new IllegalArgumentException("internal map of the create addresses for each CDI device is null.");
		if (this.mapAddress == null)
			throw new IllegalArgumentException("internal map of the  address for each CDI device is null.");
	}

	/**
	 * Validate the preliminary array of TINE addresses.
	 *
	 * @param tineAddress_ The validating preliminary TINE address array.
	 */
	private void validate(final String[] tineAddress_) {
		if (tineAddress_ == null)
			throw new NullPointerException("TINE address string array is null.");
		for(int i = 0; i < tineAddress_.length; i++)
			if (tineAddress_[i] == null && tineAddress_[i].isEmpty())
				throw new NullPointerException("TINE address string of " + i + " is null or empty.");
	}

	/**
	 * Validates the application device map.
	 *
	 * @param mapAppDevices_ Is the collection of devices which are defined in the
	 * 		CDI table.
	 * @param tineAddress_ Is the preliminary TINE address array.
	 */
	private void validate(final Map<String, SortedMap<String, IEData>> mapAppDevices_, final String[] tineAddress_) {
		if (mapAppDevices_ == null)
			throw new NullPointerException("Map of all application devices is null.");
		for(String tineAddress: mapAppDevices_.keySet())
			if (mapAppDevices_.get(tineAddress) == null || mapAppDevices_.get(tineAddress).isEmpty())
				throw new NullPointerException("Map of the application devices (TINE preliminary element " + tineAddress_ + ") is null.");
		validate(tineAddress_);
	}
}
/* end of file */