/*
 *
 * File: CbAcceleratorGroup.java
 * Date: Apr 27, 2007; 4:47:02 PM
 */
package de.desy.mst.libs.accelerator.model;

import de.desy.mst.libs.accelerator.CGeneralConstant;
import de.desy.mst.libs.accelerator.EDebug;
import de.desy.mst.libs.accelerator.IEData;
import de.desy.mst.libs.accelerator.exception.TineException;
import de.desy.mst.libs.framework.mstapp.logging.Logger;
import de.desy.tine.client.TLink;
import de.desy.tine.client.TLinkCallback;
import de.desy.tine.definitions.TErrorList;

/**
 * The call back class of a group of data values.
 *
 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
 * @version $Id: CbAcceleratorGroup.java 48667 2012-02-19 07:50:23Z walla $
 */
public final class CbAcceleratorGroup implements TLinkCallback
{
	public static final String CLAZZ = CbAcceleratorGroup.class.getSimpleName();
	
	// --- INSTANCE FIELDS -----------------------------------------------------

	/** Is the current time stamp in milliseconds. */
	private long _milliseconds;

	/** Is the current monitor instance.  */
	private final IMonitor _monitor;

	/** The value for which the monitor was created (SINGLE!!!). */
	private final IEData _data;

	// --- CONSTRUCTOR ---------------------------------------------------------
	
	/**
	 * The public call back constructor of an instance of this class.
	 *
	 * @param monitor_ Is the monitor who uses this call back method.
	 * @param data_ Is the data value the monitor uses to be displayed.
	 */
	public CbAcceleratorGroup(final IMonitor monitor_,
			final IEData data_)
	{
		if (monitor_ == null)
			throw new NullPointerException("The monitor should be set" +
					" but it is currently null.");
		if (data_ == null)
			throw new NullPointerException("The enumeration component should be set" +
					" but it is currently null.");
		_monitor = monitor_;
		_data = data_;
		_milliseconds = 0;
	}

	// --- INTSTANCE METHODS ---------------------------------------------------
	
	@Override
	public void callback(final TLink link_)
	{
		if (_milliseconds == 0)
			_milliseconds = System.currentTimeMillis();

		try
		{
			if (link_ != null && link_.getLinkStatus() == 0)
			{
				// calculate the result and show it on screen!
				_monitor.calculate(_data);
			}
			else if (link_ != null && link_.getLinkStatus() != 0)
			{
				// ATTENTION: it is illegal to throw here an exception because TINE
				// will recognize this exception and report it to the developer
				// (and this only when the corresponding debug session is set).
				// Also I don't want to report errors for link not open, I have
				// inserted an if-clause:
				if (!CGeneralConstant.ignoreTINEError(link_.getLinkStatus()) &&
						CGeneralConstant.isDebug(EDebug.DEBUG))
					Logger.SWApp.error(CLAZZ + " - CALLBACK (Group) - monitor=" +
							_monitor.getClass().getSimpleName() +
							" on " + _data + ", status=" +
							TErrorList.getErrorString(link_.getLinkStatus())
							+ " (" + link_.getLinkStatus() + ", time elapsed="
							+ (System.currentTimeMillis() - _milliseconds)/1000.0);
				_monitor.showError(link_.getLinkStatus(), _data);
			}
			else
			{
				// HINT: it is illegal to throw here an exception because TINE
				// will recognize this exception and report it to the developer
				// (and this only when the corresponding debug session is set).
				Logger.SWApp.error(CLAZZ + " - CALLBACK (Group) FATAL ERROR - monitor=" +
						_monitor.getClass().getSimpleName() +
						" - the callback TINE link is null.");
				_monitor.showError(-1, _data);
			}
		}
		catch(TineException exception)
		{
			Logger.SWApp.error(CLAZZ + " - TINE exception occurred on monitor "
					+ _monitor.getClass().getSimpleName());
			Logger.SWApp.error(CLAZZ + " - it is thrown by the following" +
					" TINE link with status="
					+ TErrorList.getErrorString(exception.getErrorCode())
					+ " (" + + exception.getErrorCode() + ")");
		}
		catch(RuntimeException exception)
		{
			Logger.SWApp.error(CLAZZ + " - CALLBACK RUNTIME - monitor= "
					+ _monitor.getClass().getSimpleName(), exception);
			_monitor.showError(link_.getLinkStatus(), null);
		}

		_milliseconds = System.currentTimeMillis();
	}
}
/* end of file */