/*
 *
 * File: ECavitySwitch.java
 * Date: Jan 27, 2008; 5:35:36 PM
 */
package de.desy.mst.libs.accelerator.model.cavity;

import de.desy.mst.libs.accelerator.CGeneralConstant;
import de.desy.mst.libs.accelerator.EName;
import de.desy.mst.libs.accelerator.exception.EnumerationNotUsedException;


/**
 * This is the output and input switching elements of a cavity.  The
 * following arguments can be selected in this enumeration:</p>
 *
 * <table cellpadding="4" cellspacing="4">
 * <tr><th>Enumeration {@link ECavitySwitch}</th><th>Description</th></tr>
 * <tr bgcolor="E0E0E0"><td>{@link ECavitySwitch#OFF_INPUT}</td><td>OFF switch (stops the movement of the stamps), input value.</td></tr>
 * <tr bgcolor="A0A0A0"><td>{@link ECavitySwitch#U_SOLL_INPUT}</td><td>Voltage SOLL switch, input value.</td></tr>
 * <tr bgcolor="E0E0E0"><td>{@link ECavitySwitch#F_MIN_INPUT}</td><td>F minimum switch, input value.</td></tr>
 * <tr bgcolor="A0A0A0"><td>{@link ECavitySwitch#F_MAX_INPUT}</td><td>F maximum switch, input value.</td></tr>
 * <tr bgcolor="E0E0E0"><td>{@link ECavitySwitch#F_RES_INPUT}</td><td>F resonance switch, input value.</td></tr>
 * <tr bgcolor="A0A0A0"><td>{@link ECavitySwitch#OFF_OUTPUT}</td><td>OFF switch (stops the movement of the stamps), output value.</td></tr>
 * <tr bgcolor="E0E0E0"><td>{@link ECavitySwitch#U_SOLL_OUTPUT}</td><td>Voltage SOLL switch, output value.</td></tr>
 * <tr bgcolor="A0A0A0"><td>{@link ECavitySwitch#F_MIN_OUTPUT}</td><td>F minimum switch, output value.</td></tr>
 * <tr bgcolor="E0E0E0"><td>{@link ECavitySwitch#F_MAX_OUTPUT}</td><td>F maximum switch, output value.</td></tr>
 * <tr bgcolor="A0A0A0"><td>{@link ECavitySwitch#F_RES_OUTPUT}</td><td>F resonance switch, output value.</td></tr>
 * </table>
 *
 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
 * @version $Id: ECavitySwitch.java 24115 2009-04-14 16:00:09Z walla $
 */
public enum ECavitySwitch implements IECavityElement {
	// --- ENUMERATIONS --------------------------------------------------------
	/** OFF switch (stops the movement of the stamps), input value. */
	OFF_INPUT("Aus", "Aus", "Aus", "", "%s")
	,
	/** Voltage SOLL switch, input value. */
	U_SOLL_INPUT("Spannung SOLL", "<html>U<sub>SOLL</sub></html>", "<html>U<sub>SOLL</sub></html>", "", "%s")
	,
	/** OF minimum switch, input value. */
	F_MIN_INPUT("f minimum", "<html>f<sub>min</sub></html>", "<html>f<sub>min</sub></html>", "", "%s")
	,
	/** F maximum switch, input value. */
	F_MAX_INPUT("f maximum", "<html>f<sub>max</sub></html>", "<html>f<sub>max</sub></html>", "", "%s")
	,
	/** F resonance switch, input value. */
	F_RES_INPUT("f resonanze", "<html>f<sub>res</sub></html>", "<html>f<sub>res</sub></html>", "", "%s")
	,

	/** OFF switch (stops the movement of the stamps), output value. */
	OFF_OUTPUT("Aus", "Aus", "Aus", "", "%s")
	,
	/** Voltage SOLL switch, output value.. */
	U_SOLL_OUTPUT("Spannung SOLL", "<html>U<sub>SOLL</sub></html>", "<html>U<sub>SOLL</sub></html>", "", "%s")
	,
	/** F minimum switch, output value. */
	F_MIN_OUTPUT("f minimum", "<html>f<sub>min</sub></html>", "<html>f<sub>min</sub></html>", "", "%s")
	,
	/** F maximum switch, output value. */
	F_MAX_OUTPUT("f maximum", "<html>f<sub>max</sub></html>", "<html>f<sub>max</sub></html>", "", "%s")
	,
	/** F resonance switch, output value. */
	F_RES_OUTPUT("f resonanze", "<html>f<sub>res</sub></html>", "<html>f<sub>res</sub></html>", "", "%s")
	;


	// --- INSTANCE FIELDS -----------------------------------------------------
	/** An HTML free name of the enumeration element. */
	private String _name;
	/** Cavity name, probably with HTML styled code. */
	private String _nameCavity;
	/** Table description name, probably in HTML styled code. */
	private String _nameTable;
	/** Unit of the enumeration element. */
	private String _unit;
	/** Precision of the enumeration element. */
	private String _precision;

	// --- CONSTRUCTOR ---------------------------------------------------------
	/**
	 * Constructs the enumeration instance.
	 *
	 * @param name_ An HTML free name of the enumeration instance.
	 * @param nameCavity_ Cavity name, probably in HTML styled code.
	 * @param nameTable_ Table description name, probably in HTML styled code.
	 * @param unit_ Unit of the enumeration instance.
	 * @param precision_ Precision of the enumeration instance.
	 */
	ECavitySwitch(
			final String name_,
			final String nameCavity_,
			final String nameTable_,
			final String unit_,
			final String precision_) {
		validate(name_, nameCavity_, nameTable_, unit_, precision_);
		initialize(name_, nameCavity_, nameTable_, unit_, precision_);
	}

	// --- INSTANCE METHODS ----------------------------------------------------
	/**
	 * Validates the enumeration instance.
	 *
	 * @param name_ An HTML free name of the enumeration instance.
	 * @param nameCavity_ Cavity name, probably in HTML styled code.
	 * @param nameTable_ Table description name, probably in HTML styled code.
	 * @param unit_ Unit of the enumeration instance.
	 * @param precision_ Precision of the enumeration instance.
	 */
	private void validate(
			final String name_,
			final String nameCavity_,
			final String nameTable_,
			final String unit_,
			final String precision_) {
		if (name_ == null)
			throw new NullPointerException("Name is null.");
		if (name_.isEmpty())
			throw new IllegalArgumentException("Name is empty.");

		if (nameCavity_ == null)
			throw new NullPointerException("Name cavity is null.");
		if (nameCavity_.isEmpty())
			throw new IllegalArgumentException("Name cavity is empty.");

		if (nameTable_ == null)
			throw new NullPointerException("Name table is null.");
		if (nameTable_.isEmpty())
			throw new IllegalArgumentException("Name table is empty.");

		if (unit_ == null)
			throw new NullPointerException("Unit is null.");

		if (precision_ == null)
			throw new NullPointerException("Precision is null.");
		if (precision_.isEmpty())
			throw new IllegalArgumentException("Precision is empty.");
		if (!precision_.matches(CGeneralConstant.PATTERN_CAVITY_PRECISION))
			throw new IllegalArgumentException("Precision of " + this.toString()
					+ " doesn't match the cavity precision: " + precision_);
	}

	/**
	 * Initialize the enumeration instance.
	 *
	 * @param name_ An HTML free name of the enumeration instance.
	 * @param nameCavity_ Cavity name, probably in HTML styled code.
	 * @param nameTable_ Table description name, probably in HTML styled code.
	 * @param unit_ Unit of the enumeration instance.
	 * @param precision_ Precision of the enumeration instance.
	 */
	private void initialize(
			final String name_,
			final String nameCavity_,
			final String nameTable_,
			final String unit_,
			final String precision_) {
		_name = name_;
		_nameCavity = nameCavity_;
		_nameTable = nameTable_;
		_unit = unit_;
		_precision = precision_;
	}

	@Override public String getName(final EName description_) throws EnumerationNotUsedException {
		switch(description_)
		{
		case NAME:
			return _name;
		case NAME_CAVITY:
			return _nameCavity;
		case NAME_TABLE:
			return _nameTable;
		case UNIT:
			return _unit;
		case PRECISION:
			return _precision;
		default:
			throw new EnumerationNotUsedException("Enumeration of " +
					description_.toString() + " exists but is not used on " +
					this.toString() + ".");
		}
	}

	/** DON'T DELETE THIS CONTENT BECAUSE IT IS NEEDED BY SOME CALLS! */
	@Override public void putName(final EName description_, final String name_) throws EnumerationNotUsedException {
		if (name_ == null)
			throw new NullPointerException("Setting " + description_.toString()
					+ " of " + this.toString() + " is null.");

		switch(description_)
		{
		case NAME:
			if (name_.isEmpty())
				throw new IllegalArgumentException("Setting " + description_.toString()
						+ " of " + this.toString() + " is empty.");
			_name = name_;
			break;
		case NAME_CAVITY:
			if (name_.isEmpty())
				throw new IllegalArgumentException("Setting " + description_.toString()
						+ " of " + this.toString() + " is empty.");
			_nameCavity = name_;
			break;
		case NAME_TABLE:
			if (name_.isEmpty())
				throw new IllegalArgumentException("Setting " + description_.toString()
						+ " of " + this.toString() + " is empty.");
			_nameTable = name_;
			break;
		case UNIT:
			_unit = name_;
			break;
		case PRECISION:
			if (name_.isEmpty())
				throw new IllegalArgumentException("Setting " + description_.toString()
						+ " of " + this.toString() + " is empty.");
			if (!name_.matches(CGeneralConstant.PATTERN_CAVITY_PRECISION))
				throw new IllegalArgumentException("Precision of " + this.toString()
						+ " doesn't match the cavity precision: " + name_);
			_precision = name_;
			break;
		default:
			throw new EnumerationNotUsedException("Enumeration of " +
					description_.toString() + " exists but is not used on " +
					this.toString() + ".");
		}
	}

	@Override public int size() { return 1; }
	@Override public boolean isAdjusting() { return false; }
	@Override public boolean isCDI() { return false; }
	@Override public boolean isFireEvent() { return false; }
}
/* end of file */