/*
 * File: Converter.java
 * Date: Jun 12, 2008; 9:32:59 AM
 */
package de.desy.mst.libs.accelerator.tine.types;

import de.desy.mst.libs.accelerator.exception.TineUnknownDataTypeException;
import de.desy.tine.types.TCompoundDataObject;

/**
 * Converts a primitive TINE data object array into a primitive Java type.
 *
 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
 * @version $Id: Converter.java 18950 2008-11-25 15:10:52Z walla $
 */
public final class Converter
{
	/**
	 * Converts the TINE data element array of the interface
	 * {@code TCompoundDataObject} into a primitive {@code double} array.
	 *
	 * @param array_ TINE data array to convert into a {@code double} array.
	 * @return A {@code double} array is returned.
	 */
	public static double[] convertDouble(final TCompoundDataObject[] array_)
	{
		validate(array_);

		double[] doubleArray = new double[array_.length];

		switch(getDataType(array_))
		{
		case SHORT:
			for(int i = 0; i < array_.length; i++)
				doubleArray[i] = ((SHORT) array_[i]).sval;
			break;
		case INTEGER:
			for(int i = 0; i < array_.length; i++)
				doubleArray[i] = ((INTEGER) array_[i]).ival;
			break;
		case FLOAT:
			for(int i = 0; i < array_.length; i++)
				doubleArray[i] = ((FLOAT) array_[i]).fval;
			break;
		default:
			throw new TineUnknownDataTypeException("TINE data type is" +
					" unknown.");
		}

		return doubleArray;
	}

	/**
	 * Converts the TINE data element array of the interface
	 * {@code TCompoundDataObject} into a primitive {@code float} array.
	 *
	 * @param array_ TINE data array to convert into a {@code float} array.
	 * @return A {@code float} array is returned.
	 */
	public static float[] convertFloat(final TCompoundDataObject[] array_)
	{
		validate(array_);

		float[] floatArray = new float[array_.length];

		switch(getDataType(array_))
		{
		case SHORT:
			for(int i = 0; i < array_.length; i++)
				floatArray[i] = ((SHORT) array_[i]).sval;
			break;
		case INTEGER:
			for(int i = 0; i < array_.length; i++)
				floatArray[i] = ((INTEGER) array_[i]).ival;
			break;
		case FLOAT:
			for(int i = 0; i < array_.length; i++)
				floatArray[i] = ((FLOAT) array_[i]).fval;
			break;
		default:
			throw new TineUnknownDataTypeException("TINE data type is" +
					" unknown.");
		}

		return floatArray;
	}

	/**
	 * Converts the TINE data element array of the interface
	 * {@code TCompoundDataObject} into a primitive {@code int} array.
	 *
	 * @param array_ TINE data array to convert into a {@code int} array.
	 * @return A {@code int} array is returned.
	 */
	public static int[] convertInt(final TCompoundDataObject[] array_)
	{
		validate(array_);

		int[] intArray = new int[array_.length];

		switch(getDataType(array_))
		{
		case SHORT:
			for(int i = 0; i < array_.length; i++)
				intArray[i] = ((SHORT) array_[i]).sval;
			break;
		case INTEGER:
			for(int i = 0; i < array_.length; i++)
				intArray[i] = ((INTEGER) array_[i]).ival;
			break;
		case FLOAT:
			for(int i = 0; i < array_.length; i++)
				intArray[i] = (int) ((FLOAT) array_[i]).fval;
			break;
		default:
			throw new TineUnknownDataTypeException("TINE data type is" +
					" unknown.");
		}

		return intArray;
	}

	/**
	 * Converts the TINE data element array of the interface
	 * {@code TCompoundDataObject} into a primitive {@code short} array.
	 *
	 * @param array_ TINE data array to convert into a {@code short} array.
	 * @return A {@code short} array is returned.
	 */
	public static short[] convertShort(final TCompoundDataObject[] array_)
	{
		validate(array_);

		short[] shortArray = new short[array_.length];

		switch(getDataType(array_))
		{
		case SHORT:
			for(int i = 0; i < array_.length; i++)
				shortArray[i] = ((SHORT) array_[i]).sval;
			break;
		case INTEGER:
			for(int i = 0; i < array_.length; i++)
				shortArray[i] = (short) ((INTEGER) array_[i]).ival;
			break;
		case FLOAT:
			for(int i = 0; i < array_.length; i++)
				shortArray[i] = (short) ((FLOAT) array_[i]).fval;
			break;
		default:
			throw new TineUnknownDataTypeException("TINE data type is" +
					" unknown.");
		}

		return shortArray;
	}

	/**
	 * Returns the corresponding enumeration data type of the TINE array
	 * depending on its instance.
	 *
	 * @param array_ TINE data array.
	 * @return Enumeration of the corresponding data element.
	 */
	private static EDataType getDataType(final TCompoundDataObject[] array_)
	{
		if(array_ instanceof SHORT[])
			return EDataType.SHORT;
		throw new IllegalArgumentException("Code failure - unknown array " +
				"implementation: " + array_.getClass().getSimpleName());
	}

	/**
	 * Private enumeration of different TINE primitve data types.
	 *
	 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
	 * @version $Id: Converter.java 18950 2008-11-25 15:10:52Z walla $
	 */
	private enum EDataType
	{
		SHORT, INTEGER, FLOAT;
	}

	/**
	 * Validating.
	 *
	 * @param array_ TINE array.
	 */
	private static void validate(final TCompoundDataObject[] array_)
	{
		if (array_ == null)
			throw new NullPointerException("Array is null.");
		if (array_.length == 0)
			throw new IllegalArgumentException("Array has no elements.");
	}
}
/* end of file */