/*
 *
 * File: SHORT.java
 * Date: Jun 11, 2008; 5:18:07 PM
 */
package de.desy.mst.libs.accelerator.tine.types;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import de.desy.tine.definitions.TFormat;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.types.TCompoundDataObject;

/**
 * <p>The primitive TINE data type called {@code SHORT} which currently doesn't 
 * exist in the TINE framework.</p>
 * <p><b>ATTENTION:</b> The methods of {@link #toStruct()} and 
 * {@link #toByteArray()} doesn't react as expected, because the TINE method of
 * {@code Swap.Short(..)} seems to be a little bit misleading. So both will 
 * throw a runtime exception if they are called.</p>
 *
 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
 * @version $Id: SHORT.java 48102 2012-01-17 14:03:18Z walla $ 
 */
public final class SHORT extends ACompoundDataObject
{
	/** Currently the primitive type of this object. */
	public short sval;
	  
	/** The size in bytes of this field. */
	public static final int sizeInBytes = 2;
	  
	/** TINE data type constant format. */
	public static final short dFormat = TFormat.CF_SHORT;
	  
	/** Buffered array output stream. */
	ByteArrayOutputStream dBuffer = new ByteArrayOutputStream(sizeInBytes);
	  
	/** Byte array of the primitive {@code short} field. */
	public byte[] hByteArray;
	  
	/**
	 * Initializes the current object with the value of 0.
	 */
	public SHORT() 
	{
		sval = 0; 
	}  
	
	/**
	 * Initializes the current object with the value of {@code short_}.
	 * 
	 * @param short_ Current {@code short} value.
	 */
	public SHORT(final short short_) 
	{ 
		sval = short_; 
	}  
	
	/**
	 * Initializes the current object with a {@code SHORT} instance.
	 * 
	 * @param instance_ The current instance of {@code SHORT}.
	 */
	public SHORT(final SHORT instance_) 
	{ 
		sval = instance_.sval; 
	}
	  
	/**
	 * Sets the current value to the delivered {@code short} one.
	 * 
	 * @param newValue_ Current value.
	 */
	public void setValues(final short newValue_)
	{
		sval = newValue_;
	}
	  
	/*
	 * (non-Javadoc)
	 * @see de.desy.tine.types.TCompoundDataObject#getSizeInBytes()
	 */
	public int getSizeInBytes() 
	{ 
		return sizeInBytes; 
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.tine.types.TCompoundDataObject#getFormat()
	 */
	public short getFormat() 
	{ 
		return dFormat; 
	}    
	  
	/*
	 * (non-Javadoc)
	 * @see de.desy.tine.types.TCompoundDataObject#setByteArray(byte[])
	 */
	public void setByteArray(byte[] byteArray) 
	{
		hByteArray = byteArray; 
	}
	  
	/*
	 * (non-Javadoc)
	 * @see de.desy.tine.types.TCompoundDataObject#getByteArray()
	 */
	public byte[] getByteArray() 
	{ 
		return hByteArray; 
	}
	
	/*
	 * (non-Javadoc)
	 * @see de.desy.tine.types.TCompoundDataObject#getIntegerValue()
	 */
	public int getIntegerValue()
	{
		return sval;
	}
	
	/*
	 * (non-Javadoc)
	 * @see de.desy.tine.types.TCompoundDataObject#getDoubleValue()
	 */
	public double getDoubleValue()
	{
		return sval;
	}
	
	/*
	 * (non-Javadoc)
	 * @see de.desy.tine.types.TCompoundDataObject#getTag()
	 */
	public String getTag()
	{
		return null;
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override public String toString()
	{
		return "["+ sval + "]";
	}
	  
	/**
	 * Returns the equality of this instance with the delivered one.
	 * 
	 * @param instance_ Delivered instance.
	 * @return Returns a {@code true} if both instances are the same, i.e.
	 * 		all fields are equal. Otherwise a {@code false} is returned.
	 */
	public boolean equals(final SHORT instance_)
	{
		if (instance_.sval != sval) return false;
		return true;
	}
	
	/*
	 * (non-Javadoc)
	 * @see de.desy.tine.types.TCompoundDataObject#toByteArray()
	 */
	public byte[] toByteArray()
	{
		try
		{
			dBuffer.reset();
			DataOutputStream ds = new DataOutputStream(dBuffer);
			ds.writeShort(Swap.Short(sval));
			ds.close();
		}
		catch (IOException e)
		{
			e.printStackTrace();
		}
		hByteArray = dBuffer.toByteArray();
		return hByteArray;
	}  
	
	/*
	 * (non-Javadoc)
	 * @see de.desy.tine.types.TCompoundDataObject#toStruct()
	 */
	public void toStruct()
	{
		try
		{
			ByteArrayInputStream dis = new ByteArrayInputStream(hByteArray);
			DataInputStream ds = new DataInputStream(dis);
			sval = Swap.Short(ds.readInt());
			ds.close(); dis.close();
		}
		catch (IOException e)
		{
			e.printStackTrace();
		}    
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override public int hashCode()
	{
		final int PRIME = 31;
		int result = 1;
		result = PRIME * result + sval;
		return result;
	}
	 
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override public boolean equals(Object obj)
	{
		if (this == obj) return true;
		if (obj == null) return false;
		if (getClass() != obj.getClass()) return false;
		final SHORT other = (SHORT) obj;
		return equals(other);
	}

	@Override public TCompoundDataObject clone() { return new SHORT(this); }
}
/* end of file */