/*
 *
 * File: IListModel.java
 * Date: Jun 21, 2007; 5:53:57 PM
 */
package de.desy.mst.libs.accelerator.view;

import javax.swing.ListModel;
import javax.swing.event.ListDataListener;

/**
 * <p>This is the interface of a list model. This one is different from that in
 * the swing framework. It allows also to add elements to the model and to
 * remove them. Also you can get the size of the list and query if it is empty.
 * But mostly there are methods for adding and removing elements which are not
 * realized in the swing interface of <tt>javax.swing.ListModel</tt>.</p>
 * <p>Aside it also allows to set or delete the listener method for data 
 * notification, called {@link #addListDataListener(ListDataListener)} and
 * {@link #removeListDataListener(ListDataListener)}. Both methods have the
 * same signature as the swing ones.</p> 
 *  
 * @author <a href="mailto:marcus.walla at desy.de">walla</a>
 * @version $Id: IListModel.java 51593 2012-08-22 15:37:13Z walla $
 */
@SuppressWarnings("rawtypes")
public interface IListModel<E> extends ListModel
{
	/**
	 * Adds an element at the index to the list. If the index is not in the
	 * stored list it will join it at the end. The index should never be 
	 * negative.
	 * 
	 * @param index_ The index to add the element.
	 * @param element_ The element to add to the model.
	 */
	public abstract void add(int index_, E element_);
	
	/**
	 * Adds an element to the end of the list.
	 * 
	 * @param element_ The element to add to the model.
	 */
	public abstract void addElement(E element_);
	
	/**
	 * Clears all element from the list.
	 */
	public abstract void clear();
	
	/**
	 * Queries if the corresponding element is in the model (<tt>true</tt>) or
	 * not (<tt>false</tt>).
	 * 
	 * @param element_ Is the element to query for.
	 * @return Returns a boolean status value.
	 */
	public abstract boolean contains(E element_);
	
	/**
	 * Gets the current element at the position of index. The index should be
	 * in the range of the stored elements.
	 * 
	 * @param index Is the index. Should be in the range of the model indexes.
	 * 		If it is outside a <tt>IllegalArgumentException</tt> will be thrown
	 * 		or a <tt>null</tt> value is returned.
	 * @return Returns the element.
	 */
	public abstract E get(int index);
	
	/**
	 * Returns the size of the model.
	 * 
	 * @return The size of the model.
	 */
	public abstract int size();
	
	/**
	 * Queries if the list is empty.
	 * 
	 * @return If the list is empty a <tt>true</tt> is returned otherwise a 
	 * 		<tt>false</tt>.
	 */
	public abstract boolean isEmpty();
	
	/**
	 * Removes the element from the model. If the element could not be found
	 * in the model a <tt>false</tt> is returned otherwise a <tt>true</tt>.
	 * 
	 * @param element_ The element to be deleted.
	 * @return A boolean value which tells the caller if the element was in the
	 * 		model.
	 */
	public abstract boolean remove(E element_);
	
	/**
	 * Removes element at the position of index and returns the deleted element.
	 * If the index is not in the range a <tt>null</tt> value is returned.
	 * 
	 * @param index_ The index in the model.
	 * @return The element which has been deleted.
	 */
	public abstract E remove(int index_);

	/**
	 * Adds a new <tt>ListDataListener</tt> to the model.
	 * 
	 * @param listener_ A <tt>ListDataListener</tt>
	 */
	public abstract void addListDataListener(ListDataListener listener_);
	
	/**
	 * Removes a new <tt>ListDataListener</tt> from the model.
	 * 
	 * @param listener_ A <tt>ListDataListener</tt>
	 */
	public abstract void removeListDataListener(ListDataListener listener_);
	
//	public abstract Iterator<E> iterator();
}
/* end of file */