/*
 *
 * File: IViewContainer.java
 * Date: May 3, 2007; 3:14:21 PM
 */
package de.desy.mst.libs.accelerator.view;

import java.io.Serializable;

import javax.swing.JComponent;

import de.desy.mst.libs.accelerator.EViewElement;
import de.desy.mst.libs.accelerator.IEData;
import de.desy.mst.libs.accelerator.IEValue;
import de.desy.mst.libs.accelerator.exception.EnumerationException;

/**
 * <p>A container can carry different labels, buttons etc. To get or set each of
 * these components to the container there exists different getter and setter 
 * methods inside this interface, which allows you to get and set the value. As 
 * an argument one has to deliver the enumeration interface of 
 * {@link IEValue} to control the element one wants to get or set.</p>
 * <p>There also exists a reset method (
 * {@link IViewContainer#reset(IEValue)}) 
 * which will reset all values to the old ones.</p>
 * <p><u><i>ATTENTION:</u></i> Remember that not all methods must be implemented 
 * by a concret class and can also throw a <tt>RuntimeException</tt>.</p>
 * 
 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
 * @version $Id: IViewContainer.java 12066 2008-05-29 08:35:57Z walla $
 */
public interface IViewContainer extends Serializable
{

	/**
	 * Will reset all current values to the old ones. The only argument
	 * to this method is the enumeration interface of the viewable container
	 * element.  For example, you can get the content of a button 
	 * by enabling it with the concrete enumeration of 
	 * <tt>EViewContainer.BUTTON</tt>.
	 * 
	 * @param container_ Is the viewable container enumeration interface.
	 */
	public abstract void reset(final IEValue container_);
	
	/**
	 * Returns the component of the viewable container.  For example, you can 
	 * get the content of a button by enabling it with the concrete enumeration 
	 * of <tt>EViewContainer.BUTTON</tt> and setting the second argument to
	 * <tt>null</tt>. Or you can just set the first viewable argument to
	 * <tt>null</tt> and set the data component to <tt>ERF104.HF104SRESET</tt>.
	 * 
	 * @param container_ Is the viewable container enumeration interface.
	 * @param data_ Is the data component enumeration interface.
	 * @return returns the viewable component as an interface.
	 * @throws EnumerationException An enumeration exceptions is thrown if
	 * 		an error occurred on the enumeration interface.
	 */
	public abstract IViewComponent getComponent(final IEValue container_, 
			final IEData data_) throws EnumerationException;

	/**
	 * This will return for the corresponding viewable enumeration interface
	 * the numeric value.  For example, you can get the content of a button 
	 * by enabling it with the concrete enumeration of 
	 * <tt>EViewContainer.BUTTON</tt>.
	 * 
	 * @param container_ Is the viewable container enumeration interface.
	 * @return The numeric value is returned.
	 */
	public abstract Number getValue(final IEValue container_);
	
	/**
	 * Sets the viewable container as an enumeration interface and the
	 * corresponding element with a numeric value.
	 * 
	 * @param element_ Is the viewable container enumeration interface.
	 * @param value_ Is the numeric value.
	 */
	public abstract void setValue(final IEValue element_, 
			final Number value_);

	/**
	 * <p>This allows to set a boolean value on the enumeration interface of 
	 * {@link IEValue}. For example, you can get the enabling of the
	 * container by calling it with <tt>EViewContainer.BUTTON</tt> and the 
	 * enumeration interface type of <tt>EViewElement.ENABLE</tt>.</p>
	 * 
	 * @param container_ Is the viewable container enumeration interface.
	 * @param element_ Is the enumeration of the viewable element.
	 * @return Returns the boolean value.
	 */
	public abstract boolean getBoolean(final IEValue container_, 
			final EViewElement element_);
	
	/**
	 * Sets for the viewable container as an enumeration interface and the
	 * corresponding element with a boolean value.  For example, you can set the 
	 * content of a button by enabling it with the concrete enumeration of 
	 * <tt>EViewContainer.BUTTON</tt>.
	 * 
	 * @param container_ Is the viewable container enumeration interface.
	 * @param element_ Is the enumeration of the viewable element.
	 * @param value_ Is the boolean value to be set.
	 */
	public abstract void setBoolean(final IEValue container_, 
			final EViewElement element_, final boolean value_);
		
	/**
	 * Returns the component element.
	 * 
	 * @return Returns the current component of this viewable container.
	 */
	public abstract JComponent getPanel();

}
/* end of file */