/*
 *
 * File: WriteArmed.java
 * Date: Dec 4, 2007; 4:42:21 PM
 */
package de.desy.mst.libs.accelerator.view;

import java.awt.Container;
import java.awt.Font;

import javax.swing.BorderFactory;
import javax.swing.GroupLayout;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.SwingConstants;

import de.desy.mst.libs.accelerator.CGeneralConstant;
import de.desy.mst.libs.accelerator.EServer;
import de.desy.mst.libs.accelerator.exception.MissingEnumerationException;

/**
 * There exists different write flags of server connections. These values are
 * displayed in a container, which is returned by the static method of
 * {@link #getInstance(de.desy.mst.libs.accelerator.view.WriteArmed.E_ARMED)}.
 * If the write flag enables the server connection it is shown with a red
 * background color, otherwise the  * corresponding component is set to green
 * background color.
 *
 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
 * @version $Id: WriteArmed.java 19706 2008-12-14 12:41:25Z walla $
 */
public final class WriteArmed
{
	private static final String CDI =
		"CDI: " + (CGeneralConstant.isWriteArmed(EServer.CDI)?"WRITE/READ":"READ");

	private static final String TINE =
		"TINE: " + (CGeneralConstant.isWriteArmed(EServer.TINE)?"WRITE/READ":"READ");

	private static final String EVENT =
		"Event Archive: " + (CGeneralConstant.isWriteArmed(EServer.EVENTARCHIVE)?"WRITE/READ":"READ");

	/**
	 * The enumeration of different instances of the read/write state. Here
	 * it is the menu items and a panel.
	 *
	 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
	 * @version $Id: WriteArmed.java 19706 2008-12-14 12:41:25Z walla $
	 */
	public enum E_ARMED { WRITE_ARMED_PANEL, WRITE_ARMED_MENU};

	/**	Is the static and final instance of this class. Need for creating the
	 * corresponding sub-classes. */
	private static final WriteArmed _writeArmed = new WriteArmed();

	/**
	 * Returns the instance of the corresponding enumeration as a factory.
	 *
	 * @param eArmed_ Is the enumeration of the container one wants to return.
	 * @return The corresponding container.
	 * @throws MissingEnumerationException Is thrown if the enumeration is
	 * 		not know.
	 */
	public static final Container getInstance(final E_ARMED eArmed_)
		throws MissingEnumerationException
	{
		switch(eArmed_)
		{
		case WRITE_ARMED_MENU:
			return _writeArmed.new WriteArmedMenu();
		case WRITE_ARMED_PANEL:
			return _writeArmed.new WriteArmedPanel();
		}
		throw new MissingEnumerationException("The corresponding container of" +
				" the enumeration " + eArmed_.toString() + " cannot be found.");
	}

	/**
	 * This constructor is only specified with the modification of private to
	 * disable the construction of any object from this class.
	 */
	private WriteArmed()
	{
	}

	/**
	 * This internal private class will return an instance of the write armed
	 * panel.
	 *
	 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
	 * @version $Id: WriteArmed.java 19706 2008-12-14 12:41:25Z walla $
	 */
	private class WriteArmedPanel extends JPanel
	{
		/** This is the current unique serial version UID. It helps the
		 * application to serialize this version. */
		public final static long serialVersionUID = 2643619945841929592l;

		/**
		 * The constructor of the different write/read flags of the server
		 * connections. Currently there are three values that are displayed by
		 * this class.
		 */
		private WriteArmedPanel()
		{
			initialize();
		}

		/**
		 * Initializes the whole panel.
		 */
		private void initialize()
		{
			GroupLayout layout = new GroupLayout(this);
			this.setLayout(layout);

			boolean cdi = CGeneralConstant.isWriteArmed(EServer.CDI);
			boolean tine = CGeneralConstant.isWriteArmed(EServer.TINE);
			boolean event = CGeneralConstant.isWriteArmed(EServer.EVENTARCHIVE);

			Font font = new Font("Arial", Font.PLAIN, 8);

			JLabel cdiLabel = new JLabel(CDI);
			cdiLabel.setBorder(BorderFactory.createLineBorder(CColor.BLACK));
			cdiLabel.setFont(font);
			cdiLabel.setOpaque(true);
			cdiLabel.setBackground((cdi?CColor.RED:CColor.GREEN));
			cdiLabel.setForeground((cdi?CColor.BLACK:CColor.BLACK));
			cdiLabel.setHorizontalAlignment(SwingConstants.CENTER);
			cdiLabel.setVerticalAlignment(SwingConstants.CENTER);
			cdiLabel.setToolTipText("CDI " + (cdi?"schreibt/liest":"liest") + " daten.");

			JLabel tineLabel = new JLabel(TINE);
			tineLabel.setBorder(BorderFactory.createLineBorder(CColor.BLACK));
			tineLabel.setFont(font);
			tineLabel.setOpaque(true);
			tineLabel.setBackground((tine?CColor.RED:CColor.GREEN));
			tineLabel.setForeground((tine?CColor.BLACK:CColor.BLACK));
			tineLabel.setHorizontalAlignment(SwingConstants.CENTER);
			tineLabel.setVerticalAlignment(SwingConstants.CENTER);
			tineLabel.setToolTipText("TINE " + (tine?"schreibt/liest":"liest") + " daten.");

			JLabel eventLabel = new JLabel(EVENT);
			eventLabel.setBorder(BorderFactory.createLineBorder(CColor.BLACK));
			eventLabel.setFont(font);
			eventLabel.setOpaque(true);
			eventLabel.setBackground((event?CColor.RED:CColor.GREEN));
			eventLabel.setForeground((event?CColor.BLACK:CColor.BLACK));
			eventLabel.setHorizontalAlignment(SwingConstants.CENTER);
			eventLabel.setVerticalAlignment(SwingConstants.CENTER);
			eventLabel.setToolTipText("EVENT ARCHIVE " + (event?"schreibt/liest":"liest") + " daten.");

			layout.setHorizontalGroup(layout.createParallelGroup()
					.addComponent(cdiLabel)
					.addComponent(tineLabel)
					.addComponent(eventLabel)
			);
			layout.setVerticalGroup(layout.createSequentialGroup()
					.addComponent(cdiLabel)
					.addComponent(tineLabel)
					.addComponent(eventLabel)
			);
		}
	}

	/**
	 * Creates the write armed menu.
	 *
	 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
	 * @version $Id: WriteArmed.java 19706 2008-12-14 12:41:25Z walla $
	 */
	private class WriteArmedMenu extends JMenu
	{
		/** This is the current unique serial version UID. It helps the application
		 * to control serialization compatibility between versions. */
		public final static long serialVersionUID = 239295185123488514l;

		/**
		 * The constructor of the different write read flags of server
		 * connections. Currently there are three values that are displayed by
		 * this class in an menu entry.
		 */
		private WriteArmedMenu()
		{
			initialize();
		}

		/**
		 * Initializes the whole menu item.
		 */
		private void initialize()
		{
			this.setText("Lese/Schreib-Zugriff Einstellungen");
			this.add(new JMenuItem(CDI));
			this.add(new JMenuItem(TINE));
			this.add(new JMenuItem(EVENT));
		}
	}

}
/* end of file */