/*
 *
 * File: RFCheckBox.java
 * Date: Feb 1, 2007; 9:39:18 AM
 */
package de.desy.mst.libs.accelerator.view.generic;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.event.ActionListener;
import java.util.EventListener;

import javax.swing.Action;
import javax.swing.Icon;
import javax.swing.JCheckBox;
import javax.swing.border.Border;

import de.desy.mst.libs.accelerator.EGraphicComponent;
import de.desy.mst.libs.accelerator.ESeverity;
import de.desy.mst.libs.accelerator.IEData;
import de.desy.mst.libs.accelerator.exception.OperationNotAllowedException;
import de.desy.mst.libs.accelerator.view.CFrame;
import de.desy.mst.libs.accelerator.view.IViewComponent;

/**
 * This is the manipulated {@code JCheckBox} which allows beside calling
 * the constructor with its normal java values also to provide its dimension.
 * Moreover there are also there inherited methods from the class
 * {@code JComponent} which all will return the delivered dimension value, so
 * that there no resize will happen. There can also be delivered a new font.
 * If this value is not set the normal font is dialed.
 *
 * @author <a href="marcus.walla(at)desy.de">walla</a>
 * @version $Id: RFCheckBox.java 36465 2010-06-04 08:04:07Z walla $
 * @see javax.swing.JCheckBox
 */
public class RFCheckBox extends JCheckBox implements IViewComponent
{
	// --- CLASS FIELDS --------------------------------------------------------
	
	/** This is the current unique serial version UID. It helps the application
	 * to control serialization compatibility between versions. */
	final static long serialVersionUID = 5694949230407096934l;

	/** Class name. */
	public static final String CLAZZ = RFCheckBox.class.getSimpleName();
	
	// --- INSTANCE FIELDS -----------------------------------------------------
	
	/** This is the dimension of this container. */
    private Dimension _dim;

    /** The severity of the check box component. This has no direct effect on the
     * GUI element. */
    private ESeverity _severity = ESeverity.NORMAL;

    // --- CONSTRUCTOR ---------------------------------------------------------
    
    /**
     * Creates an initially unselected check box button with no text, no icon but
     * with the font and dimension possibly set.
     * 
	 * The dimension is currently {@code null} and therefore this instance can
	 * vary in its size. Also the default font is used.
	 * 
     * @see #RFCheckBox(Font, Dimension)
     */
    public RFCheckBox()
    {
    	this((Font) null, (Dimension) null);
    }
    
    /**
     * Creates an initially unselected check box button with no text, no icon but
     * with the font and dimension possibly set.
     *
     * @param font_ Is the current font. This can also be {@code null}.
     * @param dim_ This is the dimension of this container. If the dimension is 
     * 		{@code null} the size of the instance is taken via its super method
     * 		and so fore this instance isn't fixed in its size.
     */
    public RFCheckBox(
    		final Font font_, 
    		final Dimension dim_)
	{
		super();
		initialization(dim_, font_);
    	debug();
	}

    /**
     * Creates an initially unselected check box with an icon and the font and
     * dimension possibly set.
     *
     * @param icon_ The Icon image to display.
     * @param dim_ This is the dimension of this container. If the dimension is 
     * 		{@code null} the size of the instance is taken via its super method
     * 		and so fore this instance isn't fixed in its size.
     * @see #RFCheckBox(Icon, boolean, Font, Dimension)
     */
	public RFCheckBox(
			final Icon icon_, 
			final Dimension dim_)
	{
		super(icon_);
		initialization(dim_, null);
    	debug();
	}

	/**
	 * Creates an initially unselected check box with text, and the font and
	 * dimension possibly set.
	 *
	 * @param text_ The text of the check box.
     * @param font_ Is the current font. This can also be {@code null}.
     * @param dim_ This is the dimension of this container. If the dimension is 
     * 		{@code null} the size of the instance is taken via its super method
     * 		and so fore this instance isn't fixed in its size.
	 */
	public RFCheckBox(
			final String text_, 
			final Font font_, 
			final Dimension dim_)
	{
		super(text_);
		initialization(dim_, font_);
    	debug();
	}

	/**
	 * Creates a check box where properties are taken from the Action supplied,
	 * and the font and dimension possibly set.
	 *
	 * @param action_ The {@code Action} of this constructed check box.
     * @param font_ Is the current font. This can also be {@code null}.
     * @param dim_ This is the dimension of this container. If the dimension is 
     * 		{@code null} the size of the instance is taken via its super method
     * 		and so fore this instance isn't fixed in its size.
	 */
	public RFCheckBox(
			final Action action_, 
			final Font font_, 
			final Dimension dim_)
	{
		super(action_);
		initialization(dim_, font_);
    	debug();
	}

	/**
	 * Creates a check box with an icon and specifies whether or not it is
	 * initially selected, and the font and dimension possibly set.
	 *
	 * @param icon_ The Icon image to display.
	 * @param selected_ A boolean value indicating the initial selection state.
	 * 			If {@code true} the check box is selected.
     * @param font_ Is the current font. This can also be {@code null}.
     * @param dim_ This is the dimension of this container. If the dimension is 
     * 		{@code null} the size of the instance is taken via its super method
     * 		and so fore this instance isn't fixed in its size.
	 */
	public RFCheckBox(
			final Icon icon_, 
			final boolean selected_, 
			final Font font_, 
			final Dimension dim_)
	{
		super(icon_, selected_);
		initialization(dim_, font_);
    	debug();
	}

	/**
	 * Creates a check box with text and specifies whether or not it is initially
	 * selected, and the font and dimension possibly set.
	 *
	 * @param text_ The text of the check box.
	 * @param selected_ A boolean value indicating the initial selection state.
	 * 			If {@code true} the check box is selected.
     * @param font_ Is the current font. This can also be {@code null}.
     * @param dim_ This is the dimension of this container. If the dimension is 
     * 		{@code null} the size of the instance is taken via its super method
     * 		and so fore this instance isn't fixed in its size.
	 */
	public RFCheckBox(
			final String text_, 
			final boolean selected_, 
			final Font font_, 
			final Dimension dim_)
	{
		super(text_, selected_);
		initialization(dim_, font_);
    	debug();
	}

	/**
	 * Creates an initially unselected check box with the specified text and
	 * icon, and the font and dimension possibly set.
	 *
	 * @param text_ The text of the check box.
	 * @param icon_ The Icon image to display.
     * @param font_ Is the current font. This can also be {@code null}.
     * @param dim_ This is the dimension of this container. If the dimension is 
     * 		{@code null} the size of the instance is taken via its super method
     * 		and so fore this instance isn't fixed in its size.
	 */
	public RFCheckBox(
			final String text_, 
			final Icon icon_, 
			final Font font_, 
			final Dimension dim_)
	{
		super(text_, icon_);
		initialization(dim_, font_);
    	debug();
	}

	/**
	 * Creates a check box with text and icon, and specifies whether or not it
	 * is initially selected, and the font and dimension possibly set.
	 *
	 * @param text_ The text of the check box.
	 * @param icon_ The Icon image to display.
	 * @param selected_ A boolean value indicating the initial selection state.
	 * 			If {@code true} the check box is selected.
     * @param font_ Is the current font. This can also be {@code null}.
     * @param dim_ This is the dimension of this container. If the dimension is 
     * 		{@code null} the size of the instance is taken via its super method
     * 		and so fore this instance isn't fixed in its size.
	 */
	public RFCheckBox(
			final String text_, 
			final Icon icon_, 
			final boolean selected_, 
			final Font font_,
			final Dimension dim_)
	{
		super(text_, icon_, selected_);
		initialization(dim_, font_);
    	debug();
	}

	// --- INSTANCE METHODS ----------------------------------------------------

	/**
	 * Initializes this instance.
	 *
	 * @param dim_ Dimension.
	 * @param font_ Font.
	 */
	private void initialization(final Dimension dim_, final Font font_)
	{
		if (font_ != null)
			setFont(font_);
   		_dim = dim_;
	}

	@Override public Dimension getMinimumSize()
	{
		if (_dim != null)
			return _dim;
		return super.getMinimumSize();
	}

	@Override public Dimension getMaximumSize()
	{
		if (_dim != null)
			return _dim;
		return super.getMaximumSize();
	}

	@Override public Dimension getPreferredSize()
	{
		if (_dim != null)
			return _dim;
		return super.getPreferredSize();
	}
	
	@Override public void setSize(final Dimension dim_)
	{
		_dim = dim_;
		super.setSize(dim_);
	}
	
	/*
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#debug()
	 */
	public void debug()
	{
		CFrame.DEBUG_BORDER(this);
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putSeverity(int)
	 */
	public void putSeverity(final ESeverity severity_)
	{
		_severity = severity_;
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#fetchSeverity()
	 */
	public ESeverity fetchSeverity()
	{
		return _severity;
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#fetchGraphicComponent()
	 */
	public EGraphicComponent fetchGraphicComponent()
	{
		return EGraphicComponent.CHECKBOX_FIXED;
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#fetchLabel()
	 */
	public String fetchText(IEData view_)
	{
		return super.getText();
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#isEnable()
	 */
	public boolean isEnable()
	{
		return super.isEnabled();
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#isVisual()
	 */
	public boolean isVisual()
	{
		return super.isVisible();
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#hasOpaque()
	 */
	public boolean hasOpaque()
	{
		return super.isOpaque();
	}

	/*
	 * Will always return the {@code false} value.
	 *
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#isValueAdjusting(de.desy.mst.libs.accelerator.model.IEDataComponent)
	 */
	public boolean isValueAdjusting(IEData view_)
	{
		return false;
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putListener(java.util.EventListener)
	 */
	public void putListener(final EventListener listener_)
	{
		if (listener_ instanceof ActionListener)
			super.addActionListener((ActionListener) listener_);
		else
			throw new IllegalArgumentException("The listener model you are" +
					" adding doesn't exist for this instance: " +
					listener_.toString());
	}

	/*
	 * Ignores the parameter of {@code IEData}.
	 *
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putLabel(java.lang.String, de.desy.mst.libs.accelerator.IEData)
	 */
	public void putText(String label_, IEData view_)
	{
		super.setText(label_);
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putNumber(java.lang.Number, de.desy.mst.libs.accelerator.IEData)
	 */
	public void putNumber(Number number_, IEData view_)
	{
		throw new OperationNotAllowedException("The operation is not allowed.");
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putEnable(boolean)
	 */
	public void putEnable(boolean enable_)
	{
		super.setEnabled(enable_);
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putVisual(boolean)
	 */
	public void putVisual(boolean visual_)
	{
		super.setVisible(visual_);
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putOpaque(boolean)
	 */
	public void putOpaque(boolean opaque_)
	{
		super.setOpaque(opaque_);
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putForeground(java.awt.Color)
	 */
	public void putForeground(Color foreground_)
	{
		super.setForeground(foreground_);
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putBackground(java.awt.Color)
	 */
	public void putBackground(Color background_)
	{
		super.setBackground(background_);
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putTooltip(java.lang.String)
	 */
	public void putTooltip(String tooltip_)
	{
		this.setToolTipText(tooltip_);
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putSelected(boolean)
	 */
	public void putSelected(boolean isSelected_)
	{
		this.setSelected(isSelected_);
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putBorder(javax.swing.border.Border)
	 */
	public void putBorder(final Border border_)
	{
		super.setBorder(border_);
	}

	@Override
	public String toString()
	{
		return CLAZZ + "[text=" + this.getText() + ']';
	}
	
}
/* end of file */