/*
 *
 * File: RFDoubleSlider.java
 * Date: Jan 10, 2008; 1:21:44 PM
 */
package de.desy.mst.libs.accelerator.view.generic;

import java.awt.Color;
import java.beans.PropertyChangeListener;
import java.util.EventListener;

import javax.swing.BoundedRangeModel;
import javax.swing.border.Border;

import com.cosylab.events.SetListener;
import com.cosylab.gui.components.Slider;
import com.cosylab.gui.components.SliderSetMode;

import de.desy.mst.libs.accelerator.EGraphicComponent;
import de.desy.mst.libs.accelerator.EName;
import de.desy.mst.libs.accelerator.ESeverity;
import de.desy.mst.libs.accelerator.IEData;
import de.desy.mst.libs.accelerator.IEDataFactor;
import de.desy.mst.libs.accelerator.IEDataRange;
import de.desy.mst.libs.accelerator.view.CFrame;
import de.desy.mst.libs.accelerator.view.IViewComponent;

/**
 * This component creates from the derivated component of the ACOP slider a
 * new slider which has lots of its properties set by the delivered data
 * component enumeration type.
 *
 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
 * @version $Id: RFDoubleSlider.java 36385 2010-06-02 11:43:05Z walla $
 */
public class RFDoubleSlider extends Slider implements IViewComponent
{
	// --- CLASS FIELDS --------------------------------------------------------
	
	/** This is the current unique serial version UID. It helps the application
	 * to control serialization compatibility between versions. */
	public static final long serialVersionUID = 1209329393213952345L;

	/** Class name. */
	public static final String CLAZZ = RFDoubleSlider.class.getSimpleName();
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	
	/** Small step of a slider. */
	public static final double SMALL_STEP = 10.0;

	/** Large step of a slider. */
	public static final double LARGE_STEP = 25.0;

	// --- INSTANCE FIELDS -----------------------------------------------------
	
	/** Is the delivered argument of the data component. For these argument
	 * the slider holds its value. */
	private IEData _data;

    /** The severity of the double slider component. This has no direct effect on the
     * GUI element. */
    private ESeverity _severity = ESeverity.NORMAL;

    // --- CONSTRUCTOR ---------------------------------------------------------
    
    /**
	 *
	 * @param data_ The data component enumeration element. This should never
	 * 		by {@code null}.
	 */
	public RFDoubleSlider(final IEData data_)
	{
		validate(data_);

		_data = data_;

		initialize(null, 0, SMALL_STEP, LARGE_STEP);
	}

	/**
	 *
	 * @param data_ The data component enumeration element. This should never
	 * 		by {@code null}.
	 * @param majorTicks_ Is an array of major ticks of the slider. This can
	 * 		also be {@code null} if the major ticks doesn't need to be
	 * 		displayed.
	 * @param minorTicksNumber_ Is the number of minor ticks between two major
	 * 		ones. This should be greater or equal to zero.
	 */
	public RFDoubleSlider(final IEData data_, final double[] majorTicks_,
			final double minorTicksNumber_)
	{
		validate(data_);

		_data = data_;

		initialize(majorTicks_, minorTicksNumber_, SMALL_STEP, LARGE_STEP);
	}

	/**
	 *
	 * @param data_ The data component enumeration element. This should never
	 * 		by {@code null}.
	 * @param majorTicks_ Is an array of major ticks of the slider. This can
	 * 		also be {@code null} if the major ticks doesn't need to be
	 * 		displayed.
	 * @param minorTicksNumber_ Is the number of minor ticks between two major
	 * 		ones. This should be greater or equal to zero.
	 * @param smallStep_ Small steps for positioning to a new value.
	 * @param largeStep_ Large steps for positioning to a new value.
	 */
	public RFDoubleSlider(final IEData data_, final double[] majorTicks_,
			final double minorTicksNumber_,
			final double smallStep_, final double largeStep_)
	{
		validate(data_, smallStep_, largeStep_);

		_data = data_;

		initialize(majorTicks_, minorTicksNumber_, smallStep_, largeStep_);
	}

	// --- INSTANCE METHODS ----------------------------------------------------
	
	/* (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#debug()
	 */
	public void debug()
	{
		CFrame.DEBUG_BORDER(this);
	}

	/**
	 * Checks all relevant values if they are valid.
	 *
	 * @param data_ The data component.
	 */
	private void validate(final IEData data_)
	{
		if (data_ == null)
			throw new NullPointerException("The value of the data component is" +
					" null.");
	}

	/**
	 * Checks all relevant values if they are valid.
	 *
	 * @param data_ The data component.
	 * @param smallStep_ Small steps for positioning to a new value.
	 * @param largeStep_ Large steps for positioning to a new value.
	 */
	private void validate(final IEData data_,
			final double smallStep_, final double largeStep_)
	{
		validate(data_);

		if (smallStep_ < 0.0 || smallStep_ > 100.0 || smallStep_ >= largeStep_)
			throw new IllegalArgumentException("Small step is invalid: " + smallStep_);
		if (largeStep_ < 0.0 || largeStep_> 1000.0 || largeStep_ <= smallStep_)
			throw new IllegalArgumentException("Large step is invalid: " + largeStep_);
	}

	/**
	 *
	 * @param majorTicks_ Is an array of major ticks of the slider. This can
	 * 		also be {@code null} if the major ticks doesn't need to be
	 * 		displayed.
	 * @param minorTicksNumber_ Is the number of minor ticks between two major
	 * 		ones. This should be greater or equal to zero.
	 * @param smallStep_ Small steps for positioning to a new value.
	 * @param largeStep_ Large steps for positioning to a new value.
	 */
	private void initialize(final double[] majorTicks_,
			final double minorTicksNumber_,
			final double smallStep_, final double largeStep_)
	{
		String dataName = _data.getName(EName.NAME_TABLE);
		String dataUnit = "";
		String dataFactorBit = "";
		String dataFactorSlow = "";
		String dataFactorFast = "";

		if (_data.getName(EName.NAME_TABLE) != null &&
				!_data.getName(EName.NAME_TABLE).isEmpty())
		{
			super.setTitle(_data.getName(EName.NAME_TABLE));
		}
		if (_data.getName(EName.UNIT) != null &&
				!_data.getName(EName.UNIT).isEmpty())
		{
			super.setUnits(_data.getName(EName.UNIT));
			dataUnit = _data.getName(EName.UNIT);
		}
		if (_data.getName(EName.PRECISION) != null &&
				!_data.getName(EName.PRECISION).isEmpty())
		{
			super.setFieldFormat(_data.getName(EName.PRECISION));
			super.setTicksFormat(_data.getName(EName.PRECISION));
		}
		if (_data instanceof IEDataRange &&
				((IEDataRange) _data).maximum() != Float.MAX_VALUE)
			super.setMaximum(((IEDataRange) _data).maximum());
		if (_data instanceof IEDataRange &&
				((IEDataRange) _data).minimum() != Float.MIN_VALUE)
			super.setMinimum(((IEDataRange) _data).minimum());

		if (_data instanceof IEDataFactor &&
				((IEDataFactor) _data).recvFactor() != Float.MAX_VALUE &&
				((IEDataFactor) _data).recvFactor() != Float.MIN_VALUE &&
				((IEDataFactor) _data).recvFactor() != 0.0)
		{
			dataFactorBit = Float.toString(((IEDataFactor) _data).recvFactor());
			dataFactorSlow = Float.toString(((IEDataFactor) _data).recvFactor() * (float) smallStep_);
			dataFactorFast = Float.toString(((IEDataFactor) _data).recvFactor() * (float) largeStep_);

			super.setBitStep(((IEDataFactor) _data).recvFactor());
			super.setSmallStep(((IEDataFactor) _data).recvFactor() * smallStep_);
			super.setLargeStep(((IEDataFactor) _data).recvFactor() * largeStep_);
			super.setButtonTooltip(Slider.Button.BIT_DECREMENT,
					"Dekrement " + dataName + " um " + dataFactorBit + ' ' + dataUnit + '.');
			super.setButtonTooltip(Slider.Button.SLOW_DECREMENT,
					"Dekrement " + dataName + " um " + dataFactorSlow + ' ' + dataUnit + '.');
			super.setButtonTooltip(Slider.Button.FAST_DECREMENT,
					"Dekrement " + dataName + " um " + dataFactorFast + ' ' + dataUnit + '.');
			super.setButtonTooltip(Slider.Button.BIT_INCREMENT,
					"Inkrement " + dataName + " um " + dataFactorBit + ' ' + dataUnit + '.');
			super.setButtonTooltip(Slider.Button.SLOW_INCREMENT,
					"Inkrement " + dataName + " um " + dataFactorSlow + ' ' + dataUnit + '.');
			super.setButtonTooltip(Slider.Button.FAST_INCREMENT,
					"Inkrement " + dataName + " um " + dataFactorFast + ' ' + dataUnit + '.');
		}

		if (majorTicks_ != null)
		{
			super.setMajorTicks(majorTicks_);
			super.setMajorTickLabels(majorTicks_);
		}
		if (minorTicksNumber_ >= 0)
			super.setMinorTicksStep(minorTicksNumber_);

		super.setSetMode(SliderSetMode.SET_ON_RELEASE);
		super.setBitStepRelative(false);
		super.setSmallStepRelative(false);
		super.setLargeStepRelative(false);
		super.setEnabled(true);
		super.setEditable(true);
		super.setBackground(null);
		// Sets the auto-synchronization to false. This will NOT set the slider
		// after a delay time to the new value.
		super.setAutoSynchronize(false);
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putSeverity(int)
	 */
	public void putSeverity(final ESeverity severity_)
	{
		_severity = severity_;
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#fetchSeverity()
	 */
	public ESeverity fetchSeverity()
	{
		return _severity;
	}

	/**
	 * Returns the current range model of the slider.
	 *
	 * @return The model of the slider is returned.
	 */
	public final BoundedRangeModel getModel()
	{
		return getSlider().getModel();
	}

	/* (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#fetchGraphicComponent()
	 */
	public EGraphicComponent fetchGraphicComponent()
	{
		return EGraphicComponent.SLIDER_FIXED;
	}

	/* (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#fetchText(de.desy.mst.libs.accelerator.IEData)
	 */
	public String fetchText(IEData view_)
	{
		return this.getTitle();
	}

	/* (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#hasOpaque()
	 */
	public boolean hasOpaque()
	{
		return this.isOpaque();
	}

	/* (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#isEnable()
	 */
	public boolean isEnable()
	{
		// returns the value if the ACOP slider is enabled by returning if it
		// is editable (has the same meaning is enabled).
		return super.isEditable();
	}

	/* (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#isValueAdjusting(de.desy.mst.libs.accelerator.IEData)
	 */
	public boolean isValueAdjusting(final IEData view_)
	{
		return super.isAutoSynchronize();
	}

	/* (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#isVisual()
	 */
	public boolean isVisual()
	{
		return super.isVisible();
	}

	/* (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putBackground(java.awt.Color)
	 */
	public void putBackground(final Color background_)
	{
		super.setBackground(background_);
	}

	/* (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putEnable(boolean)
	 */
	public void putEnable(final boolean enable_)
	{
		super.setEnabled(enable_);
	}

	/* (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putForeground(java.awt.Color)
	 */
	public void putForeground(final Color foreground_)
	{
		super.setForeground(foreground_);
	}

	/**
	 * Adds a {@code PropertyChangeListener} or {@code SetListener} interface to
	 * the listener array. All other listeners throw an
	 * {@code IllegalArgumentException}.
	 *
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putListener(java.util.EventListener)
	 */
	public void putListener(final EventListener listener_)
	{
		if (listener_ == null)
			throw new NullPointerException("Listener is null.");

		if (listener_ instanceof PropertyChangeListener)
			super.addPropertyChangeListener("value",
					(PropertyChangeListener) listener_);
		else if (listener_ instanceof SetListener)
			super.addSetListener((SetListener) listener_);
		else
			throw new IllegalArgumentException("Listener is not an instance" +
					" of PropertyChangeListener or SetListener: "
					+ listener_.toString());
	}

	/* (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putNumber(java.lang.Number, de.desy.mst.libs.accelerator.IEData)
	 */
	public void putNumber(final Number number_, final IEData view_)
	{
		if (number_ instanceof Double || number_ instanceof Float
				|| number_ instanceof Integer)
		{
			if (!super.isAutoSynchronize())
				super.setValue(number_.doubleValue());
			super.setReadback(number_.doubleValue());
		}
	}

	/* (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putOpaque(boolean)
	 */
	public void putOpaque(final boolean opaque_)
	{
		this.setOpaque(opaque_);
	}

	/* (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putSelected(boolean)
	 */
	public void putSelected(final boolean isSelected_)
	{
	}

	/* (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putText(java.lang.String, de.desy.mst.libs.accelerator.IEData)
	 */
	public void putText(final String label_, final IEData view_)
	{
		this.setTitle(label_);
	}

	/**
	 * Possibly this will not work because the getToolTip(void):String method is
	 * overwritten by the ACOP slider class.
	 *
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putTooltip(java.lang.String)
	 */
	public void putTooltip(final String tooltip_)
	{
		super.setToolTipText(tooltip_);
	}

	/* (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putVisual(boolean)
	 */
	public void putVisual(boolean visual_)
	{
		super.setVisible(visual_);
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.mst.libs.accelerator.view.IViewComponent#putBorder(javax.swing.border.Border)
	 */
	public void putBorder(final Border border_)
	{
		super.setBorder(border_);
	}
	
	@Override
	public String toString()
	{
		return CLAZZ + "[value=" + this.getValue() + ']';
	}
}
/* end of file */