/*
 *
 * File: RFLabel.java
 * Date: Jan 19, 2007; 3:04:33 PM
 */
package de.desy.mst.libs.accelerator.view.generic;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.util.EventListener;

import javax.swing.BorderFactory;
import javax.swing.Icon;
import javax.swing.JLabel;
import javax.swing.border.Border;

import de.desy.mst.libs.accelerator.CGeneralConstant;
import de.desy.mst.libs.accelerator.EDebug;
import de.desy.mst.libs.accelerator.EGraphicComponent;
import de.desy.mst.libs.accelerator.EName;
import de.desy.mst.libs.accelerator.ESeverity;
import de.desy.mst.libs.accelerator.IEData;
import de.desy.mst.libs.accelerator.exception.NotImplementedException;
import de.desy.mst.libs.accelerator.view.CColor;
import de.desy.mst.libs.accelerator.view.IViewComponent;

/**
 * This is the manipulated {@code JLabel} which allows beside calling
 * the constructor with its normal java values also to provide its dimension.
 * Moreover there are also there inherited methods from the class
 * {@code JComponent} which all will return the delivered dimension value, so
 * that there no resize will happen. There can also be delivered a new font.
 * If this value is not set the normal font is dialed.
 *
 * @author <a href="marcus.walla(at)desy.de">walla</a>
 * @version $Id: RFLabel.java 51593 2012-08-22 15:37:13Z walla $
 * @see javax.swing.JLabel
 */
public class RFLabel extends JLabel implements IViewComponent {
	// --- CLASS FIELDS --------------------------------------------------------
	
	/** This is the current unique serial version UID. It helps the application
	 * to control serialization compatibility between versions. */
	static final long serialVersionUID = 256030102948585201L;

	/** Class name. */
	public static final String CLAZZ = RFLabel.class.getSimpleName();
	
	// --- INSTANCE FIELDS -----------------------------------------------------
	
	/** This is the dimension of this container. */
	private Dimension _dim;

    /** The severity of the label component. This has no direct effect on the
     * GUI element. */
    private ESeverity _severity = ESeverity.NORMAL;

    // --- CONSTRUCTOR ---------------------------------------------------------
    
	/**
	 * Creates a JLabel instance without any special arguments.
	 * 
	 * The dimension is currently {@code null} and therefore this instance can
	 * vary in its size. Also the default font is used.
	 * 
	 * @see #RFLabel(String, Font, Dimension)
	 */
	public RFLabel() {
		super();
		initialization(null, null);
		debug();
	}

	/**
	 * Creates a JLabel instance with the specified text. The font and dimension
	 * are set to its default values.
	 *
	 * The dimension is currently {@code null} and therefore this instance can
	 * vary in its size. Also the default font is used.
	 *
	 * @param name_ The text to be displayed by the label.
	 * @see #RFLabel(String, Font, Dimension)
	 */
	public RFLabel(final String name_) {
		super(name_);
		initialization(null, null);
		debug();
	}

	/**
	 * Creates a JLabel instance with the specified text. The font is possibly
	 * set.
	 *
	 * The dimension is currently {@code null} and therefore this instance can
	 * vary in its size.
	 * 
	 * @param name_ The text to be displayed by the label.
     * @param font_ Is the current font. This can also be {@code null}.
	 */
	public RFLabel(final String name_, final Font font_) {
		super(name_);
		initialization(font_, null);
    	debug();
	}

	/**
	 * Creates a JLabel instance with the specified text. The font and dimension
	 * is possibly also set.
	 *
	 * @param name_ The text to be displayed by the label.
     * @param font_ Is the current font. This can also be null.
     * @param dim_ This is the dimension of this container. If the dimension is 
     * 		{@code null} the size of the instance is taken via its super method
     * 		and so fore this instance isn't fixed in its size.
	 */
	public RFLabel(final String name_, final Font font_, final Dimension dim_) {
		super(name_);
		initialization(font_, dim_);
    	debug();
	}

	/**
	 * Creates a JLabel instance with the specified text. The font and dimension
	 * is possibly also set.
	 *
	 * @param name_ The text to be displayed by the label.
	 * @param horizontalAlignment_ Is the horizontal alignment of this label.
	 * 		This value can be set to the constant values of {@code SwingConstants}
	 * 		to {@code RIGHT}, {@code LEFT}, {@code CENTER}, {@code LEADING},
	 * 		and {@code TRAILING}.
     * @param font_ Is the current font. This can also be {@code null}.
     * @param dim_ This is the dimension of this container. If the dimension is 
     * 		{@code null} the size of the instance is taken via its super method
     * 		and so fore this instance isn't fixed in its size.
	 */
	public RFLabel(final String name_, final int horizontalAlignment_,
			final Font font_, final Dimension dim_) {
		super(name_, horizontalAlignment_);
		initialization(font_, dim_);
    	debug();
	}

	/**
	 * Creates a JLabel instance with the specified image. The font and dimension
	 * is possibly also set.
	 *
	 * @param icon_ The image to be displayed by the label.
     * @param dim_ This is the dimension of this container. If the dimension is 
     * 		{@code null} the size of the instance is taken via its super method
     * 		and so fore this instance isn't fixed in its size.
	 */
	public RFLabel(final Icon icon_, final Dimension dim_) {
		super(icon_);
		initialization(null, dim_);
    	debug();
	}

	/**
	 * Creates a JLabel instance with the specified text. The font and dimension
	 * is possibly also set.
	 *
	 * @param name_ The text to be displayed by the label.
	 * @param icon_ The image to be displayed by the label.
	 * @param horizontalAlignment_ One of the following constants defined in
	 * 		SwingConstants: {@code LEFT}, {@code CENTER}, {@code RIGHT},
	 * 		{@code LEADING} or {@code TRAILING}.
     * @param font_ Is the current font. This can also be {@code null}.
     * @param dim_ This is the dimension of this container. If the dimension is 
     * 		{@code null} the size of the instance is taken via its super method
     * 		and so fore this instance isn't fixed in its size.
	 */
	public RFLabel(final String name_, final Icon icon_,
			final int horizontalAlignment_, final Font font_,
			final Dimension dim_) {
		super(name_, icon_, horizontalAlignment_);
		initialization(font_, dim_);
		setFont(font_);
    	debug();
	}

	// --- INSTANCE METHODS ----------------------------------------------------

	/**
	 * Initialization.
	 *
     * @param font_ Is the current font. This can also be null.
     * @param dim_ This is the dimension of this container. If it is null the
     * 		dimension of the current container is selected.
	 */
	private void initialization(
			final Font font_, 
			final Dimension dim_) {
		if (font_ != null)
			setFont(font_);
		_dim = dim_;
	}

	@Override public Dimension getMinimumSize() {
		if (_dim == null)
			return super.getMinimumSize();
		return _dim;
	}

	@Override public Dimension getMaximumSize() {
		if (_dim == null)
			return super.getMaximumSize();
		return _dim;
	}

	@Override public Dimension getPreferredSize() {
		if (_dim == null)
			return super.getPreferredSize();
		return _dim;
	}
	
	@Override public void setSize(final Dimension dim_) {
		_dim = dim_;
		super.setSize(dim_);
	}
	
	@Override public void debug() {
		if (CGeneralConstant.isDebug(EDebug.DEBUG_GRAPHIC))
			super.setBorder(BorderFactory.createLineBorder(CColor.RED));
	}

	@Override public void putSeverity(final ESeverity severity_) {
		_severity = severity_;
	}

	@Override public ESeverity fetchSeverity() {
		return _severity;
	}

	@Override public EGraphicComponent fetchGraphicComponent() {
		return EGraphicComponent.LABEL_FIXED;
	}

	@Override public String fetchText(IEData view_) {
		return super.getText();
	}

	@Override public boolean isEnable() {
		return super.isEnabled();
	}

	@Override public boolean isVisual() {
		return super.isVisible();
	}

	@Override public boolean hasOpaque() {
		return super.isOpaque();
	}

	@Override public boolean isValueAdjusting(IEData view_) {
		return false;
	}
	
	@Override public void putListener(final EventListener listener_) {
		throw new NotImplementedException();
	}

	@Override public void putText(String label_, IEData view_) {
		super.setText(label_);
	}

	@Override public void putNumber(Number number_, IEData data_) {
		super.setText(String.format(data_.getName(EName.PRECISION), number_));
	}

	@Override public void putEnable(boolean enable_) {
		if (super.isEnabled() != enable_)
			super.setEnabled(enable_);
	}

	@Override public void putVisual(boolean visual_) {
		if (super.isVisible() != visual_)
			super.setVisible(visual_);
	}

	@Override public void putOpaque(boolean opaque_) {
		if (super.isOpaque() != opaque_)
			super.setOpaque(opaque_);
	}

	@Override public void putForeground(Color foreground_) {
		if (!super.getForeground().equals(foreground_))
			super.setForeground(foreground_);
	}

	@Override public void putBackground(Color background_) {
		if (!super.getBackground().equals(background_))
			super.setBackground(background_);
	}

	@Override public void putTooltip(String tooltip_) {
		super.setToolTipText(tooltip_);
	}

	@Override public void putSelected(boolean isSelected_) {
		throw new NotImplementedException();
	}

	@Override public void putBorder(final Border border_) {
		super.setBorder(border_);
	}
	
	@Override public String toString() {
		return CLAZZ + "[text=" + this.getText() + ']';
	}
}
/* end of file */