/**
 * Aug  6, 2008 - Aug 27, 2008	ws:	Creation from scratch
 */
package de.desy.mst.libs.framework.mstapp.logging.gui;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import de.desy.mst.libs.framework.mstapp.ini.LocalHome;
import de.desy.mst.libs.framework.mstapp.logging.Logger;
import de.desy.mst.libs.framework.mstapp.logging.appender.file.PfadeProtokoll;

/**
 * Es existieren eine Reihe von *.log Dateien mit verschiedenen Formaten. Diese
 * Klasse versucht die richtige Factory zu finden.
 * @author ws
 * @version 1.0.0
 */
@SuppressWarnings({"rawtypes"})
abstract class File_LogSourceFactory {
	private static final File_LogSourceFactory DEFAULT_FACTORY;
	private static final File_LogSourceFactory[] factories;
	private static final List<File_LogSourceFactory> DEFAULT_LIST;
	//-------------------------------------------------------------------------
	static{
		DEFAULT_FACTORY = new CSV_SF();

		factories = new File_LogSourceFactory[] { new MstLogSF()
										  		, DEFAULT_FACTORY
										  		, new TineLogSF()
												, new WDogLogSF()
		};

		final LinkedList<File_LogSourceFactory> list = new LinkedList<File_LogSourceFactory>();
		list.add(DEFAULT_FACTORY);
		DEFAULT_LIST = Collections.unmodifiableList(list);
	}
//=============================================================================
		/**file wird geoeffnet*/
	public static ALogSource createBestLogSource(final File file) throws IOException{
		final List<File_LogSourceFactory> factories = getFactories(file);
		for (final File_LogSourceFactory factory : factories) {
			final ALogSource logSource = factory.createLogSource(file);
			if(!logSource.hasMassiveDecodingProblems())
				return logSource;	//fertig
			Logger.SWFW.info("File_LogSourceFactory.createBestLogSource(\""
							+ file.getAbsolutePath()
							+ "\") hat massive Dekodierungsprobleme mit \""
							+ factory.getName() + "\".");
		}
		return CSVSource.create(file);	//geht immer, sollte hier nie hinkommen
	}
	//-------------------------------------------------------------------------
	/**alle bekannten factories sortiert mit der besten im Index 0*/
	static List<File_LogSourceFactory> getFactories(final File file){
		return findByFileProperties(file);	//can improve to test read
	}
	static File_LogSourceFactory[] getFactories() {
		return File_LogSourceFactory.factories;
	}
	/**alle bekannten factories sortiert mit der besten im Index 0*/
	static List<File_LogSourceFactory> findByFileProperties(final File file) {
		if(file == null )
			return DEFAULT_LIST;

		final String filePath = file.getParentFile()	//ohne fileName nicht
		.getAbsolutePath()
		.toLowerCase();
		if(filePath.length() == 0)
			return DEFAULT_LIST;

		final ScoreFactoryPair[] pairs = new ScoreFactoryPair[getFactories().length];
		final String fileName = file.getName().toLowerCase();
		for (int i = 0; i < pairs.length; i++) {
			final File_LogSourceFactory factory = getFactories()[i];
			final int score = score(filePath, fileName, factory);
			pairs[i] = new ScoreFactoryPair(score, factory);
		}
		Arrays.sort(pairs);
		final ArrayList<File_LogSourceFactory> list = extractFactories(pairs);
		return list;
	}
	//-------------------------------------------------------------------------
	private static int score(final String filePath, final String fileName, final File_LogSourceFactory factory) {
		return factory.scoreByFileName(fileName)
			 + factory.scoreByFilePath(fileName, filePath);
	}
	private static ArrayList<File_LogSourceFactory> extractFactories(final ScoreFactoryPair[] pairs) {
		final ArrayList<File_LogSourceFactory> list = new ArrayList<File_LogSourceFactory>(pairs.length);
		for (final ScoreFactoryPair scoreFactoryPair : pairs) {
			list.add(scoreFactoryPair.factory);
		}
		return list;
	}
//=============================================================================
	
	public abstract ALogSource createLogSource(final File file) throws IOException;
	//-------------------------------------------------------------------------
		   abstract String getName();
		   abstract int    scoreByFileName(final String fileName);
		   abstract int    scoreByFilePath(final String fileName, final String filePath);
	//-------------------------------------------------------------------------
    File_LogSourceFactory getDefaultFactory(){return DEFAULT_FACTORY;}
	int pathPartBonus(final String pathPart, final int bonus, final int neutral, final String filePath) {
		return (filePath.contains(pathPart) ? bonus : neutral);
	}

//=============================================================================

	private static class ScoreFactoryPair implements Comparable<ScoreFactoryPair>{
		final int score;
		final File_LogSourceFactory factory;
		private ScoreFactoryPair(final int score, final File_LogSourceFactory factory) {
			this.score = score;
			this.factory = factory;
		}
		@Override public int compareTo(final ScoreFactoryPair other) {
			final Integer  thisScore = Integer.valueOf( this.score);
			final Integer otherScore = Integer.valueOf(other.score);
			return otherScore.compareTo(thisScore);
		}
	}// End class ScoreFactoryPair

}// End class File_LogSourceFactory
//=============================================================================
//=============================================================================
//=============================================================================



class MstLogSF extends File_LogSourceFactory{
	private static final String FILE_SUFFIX 	= PfadeProtokoll.FILE_SUFFIX.toLowerCase();
	private static final String SRV_NAME 		= PfadeProtokoll.SRV_NAME	.toLowerCase();
	private static final String HOME_DIR_NAME 	= LocalHome.HOME_DIR_NAME	.toLowerCase();
	//-------------------------------------------------------------------------
	
	@SuppressWarnings("rawtypes")
	@Override public ALogSource createLogSource(final File file) throws IOException {
		return ALogSource.createLogSource_ReaderProvider(file);
	}
	@Override String getName() {return "MstLog Factory";}
	@Override int scoreByFileName(final String name) {
		if(name.endsWith(FILE_SUFFIX))
			return 7;
		if(name.endsWith(".log"))
			return 5;
		return 0;
	}
	@Override int scoreByFilePath(final String fileName, final String filePath) {
		return pathPartBonus(SRV_NAME,      3, 0, filePath)	 	//protokoll
			 + pathPartBonus(HOME_DIR_NAME, 3, 0, filePath);	//mst_app
	}
}// End class MstLogSF
//=============================================================================
class CSV_SF extends File_LogSourceFactory{
	
	@SuppressWarnings("rawtypes")
	@Override public ALogSource createLogSource(final File file) throws IOException {
		return CSVSource.create(file);
	}
	@Override String getName() {return "CSV Factory";}
	@Override int scoreByFileName(final String name) {
		if(name.endsWith(".csv"))
			return 20;
		return 2;	//can read most formats
	}
	@Override int scoreByFilePath(final String fileName, final String filePath) {
		return pathPartBonus("database", 3, 0, filePath)
			 + pathPartBonus("controls", 3, 0, filePath);
	}
}// End class CSV_SF
//=============================================================================
@SuppressWarnings({"rawtypes"})
class TineLogSF extends File_LogSourceFactory{
	@Override public ALogSource createLogSource(final File file) throws IOException {
		return TineStdLogSource.create(file);
	}
	@Override String getName() {return "TineLog Factory";}
	@Override int scoreByFileName(final String name) {
		if(name.endsWith(".log") || name.endsWith(".bak"))
			return 7;
		if(name.contains(".log.") || name.contains(".bak."))
			return 5;
		return 0;
	}
	@Override int scoreByFilePath(final String fileName, final String filePath) {
		return pathPartBonus("log", 	 3, 0, filePath)
			 + pathPartBonus("controls", 3, 0, filePath);
	}

}// End class TineLogSF
//=============================================================================
@SuppressWarnings({"rawtypes"})
class WDogLogSF extends File_LogSourceFactory{
	@Override public ALogSource createLogSource(final File file) throws IOException {
		return WDogStdLogSource.create(file);
	}
	@Override String getName() {return "WDog Log Factory";}
	@Override int scoreByFileName(final String name) {
		if(name.startsWith("wdog") && name.endsWith(".log"))
				return 20;
		if(name.endsWith(".log"))
			return 4;
		return 0;
	}
	@Override int scoreByFilePath(final String fileName, final String filePath) {
		return pathPartBonus("log", 	 3, 0, filePath)
			 + pathPartBonus("controls", 3, 0, filePath);
	}
}// End class WDogLogSF
