/*
 * Created on May 27, 2009
 *
 */
package de.desy.mst.libs.ser.tclientlib.connection;

/**
 * Manages a connection.
 * 
 * The connection manager tries to keep up a connection by means of a state
 * machine. It maintains information about the current state of the connection
 * and error statistics. The ConnectionManager.update() method must be called
 * periodically. In this method, the connection state is checked.
 * 
 * @author jwlg
 */
public class ConnectionManager {

    private IObservableConnection observedConnection = null;

    // Counters for connection state management
    private LimitedCounter failureDelayCounter = new LimitedCounter(5);
    private LimitedCounter connectTimeoutCounter = new LimitedCounter(60);
    private LimitedCounter reconnectDelayCounter = new LimitedCounter(30);
    private LimitedCounter stabilizationDelayCounter = new LimitedCounter(3);
    private LimitedCounter subsequentReconnectionsCounter = new LimitedCounter(0);

    private boolean isFirstConnection = true;

    private StatisticsCounter disabledCounter = new StatisticsCounter();
    private StatisticsCounter disconnectedCounter = new StatisticsCounter();
    private StatisticsCounter connectingCounter = new StatisticsCounter();
    private StatisticsCounter failedCounter = new StatisticsCounter();
    private StatisticsCounter readyCounter = new StatisticsCounter();

    private IConnectionStateTransition requestedTransition = null;
    private IConnectionStateTransition currentTransition = null;

    private final IConnectionStateTransition disconnectedState =
        new IConnectionStateTransition() {

        @Override
        public ConnectionState getState() {
            return ConnectionState.DISCONNECTED;
        }

        @Override
        public IConnectionStateTransition update() {

            IConnectionStateTransition requestedTransition = null;
            // If disconnectedCounter is active, increment and wait until
            // Counter
            // has
            // expired. If counter is expired or not active,
            // open connections,
            // set state to CONNECTING and start reconnectCounter.
            // Don't wait if this is the first connection.

            reconnectDelayCounter.increment();

            if (isFirstConnection || reconnectDelayCounter.isExpired()) {

                isFirstConnection = false;

                subsequentReconnectionsCounter.increment();
                if (subsequentReconnectionsCounter.isActive()
                    && subsequentReconnectionsCounter.isExpired()) {
                    // NEVER try to re-establish this connection again!
                    requestedTransition = disabledState;
                } else {
                    if (observedConnection.openConnection()) {
                        requestedTransition = connectingState;
                    } else {
                        observedConnection.closeConnection();
                        reconnectDelayCounter.reset();
                        // We stay disconnected. We do not have to request a
                        // transition.
                        // requestTransition(ConnectionState.DISCONNECTED);
                    }
                }
            }
            return requestedTransition;
        }

        @Override
        public void enter() {
            disconnectedCounter.increment();
        }

        @Override
        public void leave() {
            resetCounters();
        }

    };

    private final IConnectionStateTransition disabledState =
        new IConnectionStateTransition() {

        @Override
        public ConnectionState getState() {
            return ConnectionState.DISABLED;
        }

        @Override
        public IConnectionStateTransition update() {
            // No transition from here!
            return null;
        }

        @Override
        public void enter() {
            disabledCounter.increment();            
        }

        @Override
        public void leave() {
            resetCounters();
        }

    };

    private final IConnectionStateTransition readyState =
        new IConnectionStateTransition() {

        @Override
        public ConnectionState getState() {
            return ConnectionState.READY;
        }

        @Override
        public IConnectionStateTransition update() {
            IConnectionStateTransition requestedTransition = null;
            if (!observedConnection.isDataAvailable()) {
                requestedTransition = failedState;
            }
            return requestedTransition;
        }

        @Override
        public void enter() {
            readyCounter.increment();
        }

        @Override
        public void leave() {
            resetCounters();
        }

    };

    private final IConnectionStateTransition failedState =
        new IConnectionStateTransition() {

        @Override
        public ConnectionState getState() {
            return ConnectionState.FAILED;
        }

        @Override
        public IConnectionStateTransition update() {

            // If connection is still failed,
            // increment failedCounter.
            // Else, stop failCounter and set State to READY.
            // If failCounter has expired, close connections and
            // setState to DISCONNECTED.

            IConnectionStateTransition requestedTransition = null;

            if (observedConnection.isDataAvailable()) {
                // FAILED can go to READY immediately.
                requestedTransition = readyState;
            } else {
                failureDelayCounter.increment();
                if (failureDelayCounter.isExpired()) {
                    observedConnection.closeConnection();
                    requestedTransition = disconnectedState;
                }
            }
            return requestedTransition;

        }

        @Override
        public void enter() {
            failedCounter.increment();
        }

        @Override
        public void leave() {
            resetCounters();
        }

    };

    private final IConnectionStateTransition connectingState =
        new IConnectionStateTransition() {

        @Override
        public ConnectionState getState() {
            return ConnectionState.CONNECTING;
        }

        @Override
        public IConnectionStateTransition update() {

            IConnectionStateTransition requestedTransition = null;

            // if reconnectCounter is not expired, try to reconnect.
            // If connection OK, set state to ready. reset.reconnectCounter.
            // if connection is not OK, increment counter.
            // If counter is expired, set state to DISCONNECTED.

            if (observedConnection.isDataAvailable()) {
                stabilizationDelayCounter.increment();
                if (stabilizationDelayCounter.isExpired()) {
                    subsequentReconnectionsCounter.reset();
                    requestedTransition = readyState;
                }
            } else {
                connectTimeoutCounter.increment();
                if (connectTimeoutCounter.isExpired()) {
                    observedConnection.closeConnection();
                    requestedTransition = disconnectedState;
                }
            }
            return requestedTransition;

        }

        @Override
        public void enter() {
            connectingCounter.increment();
        }

        @Override
        public void leave() {
            resetCounters();
        }

    };

    public ConnectionManager(IObservableConnection connection) {
        this.currentTransition = this.disconnectedState;
        this.observedConnection = connection;
    }

    /**
     * Updates the connection state.
     * 
     * This method must be called periodically to check the current state and
     * make a state transition if necessary.
     * 
     */
    public final void update() {

        requestedTransition = null; // reset
        if (currentTransition == null) {
            currentTransition = disconnectedState;
        }
        requestedTransition = currentTransition.update();

        performTransition();

    }

    private void resetCounters() {
        failureDelayCounter.reset();
        connectTimeoutCounter.reset();
        reconnectDelayCounter.reset();
        stabilizationDelayCounter.reset();

    }

    private void performTransition() {
        if ((requestedTransition != null) && (requestedTransition != currentTransition)) {
            currentTransition.leave();
            currentTransition = requestedTransition;
            currentTransition.enter();
        }
        requestedTransition = null;
    }

    public int getFailureDelay() {
        return failureDelayCounter.getMaxValue();
    }

    /**
     * Specifies the time to wait until a failure is recognized as an error.
     * This value allows 'glitches' to occur during the specified tolerance
     * time. If the connection fails, the state goes to FAILED. It can switch
     * back to READY during this time if the connection recovers. When the time
     * has expired and the connection is still not OK, the state switches to
     * DISCONNECTED and the observed connection is closed.
     * 
     * @param failureDelay
     *            Delay in number of updates.
     */
    public void setFailureDelay(int failureDelay) {
        this.failureDelayCounter.setMaxValue(failureDelay);
    }

    public int getConnectingTimeout() {
        return connectTimeoutCounter.getMaxValue();
    }

    /**
     * Sets the connection timeout. The connection timeout is the maximal time
     * to wait after the connection has opened until the connection is working.
     * 
     * @param connectingTimeout
     *            The timeout in update intervals.
     */
    public void setConnectingTimeoutCounter(int connectingTimeout) {
        this.connectTimeoutCounter.setMaxValue(connectingTimeout);
    }

    public int getReconnectDelay() {
        return reconnectDelayCounter.getMaxValue();
    }

    /**
     * Assigns a reconnect delay. This is the time to wait after a connection
     * has been recognized as DISCONNECTED until the ConnectionManager tries to
     * re-establish it.
     * 
     * @param reconnectDelay
     */
    public void setReconnectDelay(int reconnectDelay) {
        this.reconnectDelayCounter.setMaxValue(reconnectDelay);
    }

    public int getStabilizationDelay() {
        return stabilizationDelayCounter.getMaxValue();
    }

    /**
     * Assigns a stabilization delay.
     * 
     * After a connection has been opened and works, the ConnectionManager may
     * wait a few intervals until it assumes that the connection is stable. Only
     * after this delay it switches from CONNECTING to READY, otherwise back to
     * DISCONNECTED.
     * 
     * The default value is 3.
     * 
     * @param connectingDelay
     *            The delay to wait for the stabilization of a connection.
     */
    public void setStabilizationDelay(int connectingDelay) {
        this.stabilizationDelayCounter.setMaxValue(connectingDelay);
    }

    public ConnectionState getConnectionState() {
        return currentTransition.getState();
    }

    /**
     * Same as setDisabled();
     */
    public void closeConnection() {
        setDisabled();
    }

    /**
     * Closes the connection and sets the state to DISABLED. The next update
     * will try to reconnect.
     */
    public void resetConnection() {
        setDisabled();
        setEnabled();
    }

    /**
     * Explicitly the connection.
     * 
     * A connection can only be enabled when the connection state is DISABLED.
     * The state is then changed to DISCONNECTED, and the next update() will try
     * to build up the connection.
     * 
     */
    public void setEnabled() {
        if (this.currentTransition == disabledState) {
            resetCounters();
            isFirstConnection = true;
            this.currentTransition = disconnectedState;
        }
    }

    /**
     * Explicitly disables the connection.
     * 
     * Closes the connection and sets the connection state to DISABLED. Has no
     * function if the connection state already is DISABLED.
     */
    public void setDisabled() {
        if (this.currentTransition != disabledState) {
            observedConnection.closeConnection();
            resetCounters();
            this.currentTransition = disabledState;
        }
    }

    /**
     * Assigns a maximal number of subsequent reconnections.
     * 
     * If the connection manager cannot open a connection, it retries to open it
     * after a certain period of time. This process repeats forever.
     * 
     * If a maximal number of subsequent reconnections is set, the connection
     * manager stops retrying when no connection can be built up and sets the
     * connection state to DISABLED. After that it will never try to connect
     * again unless setEnabled() is called.
     * 
     * @param maxReconnects
     *            Maximal number of subsequent reconnects. <=0 means no limit.
     */
    public void setMaxSubsequentReconnects(int maxReconnects) {
        if (maxReconnects < 0)
            maxReconnects = 0;
        subsequentReconnectionsCounter.setMaxValue(maxReconnects);
    }

    public int getMaxSubsequentReconnects() {
        return subsequentReconnectionsCounter.getMaxValue();
    }

    public long getDisabledCounter() {
        return disabledCounter.getValue();
    }

    public long getFailedCounter() {
        return failedCounter.getValue();
    }

    public long getDisconnectedCounter() {
        return disconnectedCounter.getValue();
    }

    public long getReadyCounter() {
        return readyCounter.getValue();
    }

    public long getConnectingCounter() {
        return connectingCounter.getValue();
    }

}
