/*
 * Created on Mar 8, 2006
 *
 */
package de.desy.mst.libs.ser.tclientlib.utils;

import de.desy.mst.libs.ser.tclientlib.TLinkAddress;
import de.desy.tine.definitions.TStrings;

/**
 * TINE Name Utilities.
 * 
 * Utilities to validate address components.
 * 
 * Note: This is a singleton. Sample Usage:
 * TineNameUtils.getInstance().isContextNameCorrect();
 * 
 * @author jwlg
 */
public class TineNameUtils {

    private static TineNameUtils theInstance = new TineNameUtils();
    
    /**
     * Private constructor. This is a singleton.
     */
    private TineNameUtils() {
        super();
    }

    /**
     * Returns the instance.
     */
    public static TineNameUtils getInstance() {
        return theInstance;
    }

    /**
     * Returns true if a specified name is a syntactically valid alarm name.
     * 
     * @param name
     *            The name to be examined.
     */
    public boolean isAlarmTagSyntaxOK(String name) {
        return (name != null) && name.matches("[A-Z]\\w*");
    }

    /**
     * Returns true if a specified name is a syntactically valid context name.
     * 
     * @param name
     *            The name to be examined.
     */
    public boolean isContextNameSyntaxOK(String name) {
        return (name != null) && name.matches("[A-Za-z][A-Za-z0-9_.]*");
    }

    /**
     * Returns true if a specified name is a syntactically valid device name.
     * 
     * @param name
     *            The name to be examined.
     */
    public boolean isDeviceNameSyntaxOK(String name) {
        return (name != null) && name.matches("[A-Za-z][A-Za-z0-9_.]*");
    }

    /**
     * Returns true if a specified name is a syntactically valid server name (=
     * TINE device server = TINE export name = ACOP device group) .
     * 
     * @param name
     *            The name to be examined.
     */
    public boolean isServerNameSyntaxOK(String name) {
        return (name != null) && name.matches("[A-Za-z][a-zA-Z0-9_.-]*");
    }

    /**
     * Returns true if a specified name is a syntactically valid property name.
     * 
     * @param name
     *            The name to be examined.
     */
    public boolean isPropertyNameSyntaxOK(String name) {
        return (name != null) && name.matches("[A-Za-z][a-zA-Z0-9_.-]*");
    }

    /**
     * Returns true if the size of the property name is permitted of TINE.
     * 
     * @param name
     *            The name to be examined.
     */
    public boolean isPropertyNameSizeOK(String name) {
        return ((name != null) &&
            (name.length() > 0) && (name.length() <= TStrings.PROPERTY_NAME_SIZE));
    }

    /**
     * Returns true if the size of the device name is permitted of TINE.
     * 
     * @param name
     *            The name to be examined.
     */
    public boolean isDeviceNameSizeOK(String name) {
        return ((name != null) &&
            (name.length() > 0) && (name.length() <= TStrings.DEVICE_NAME_MAXSIZE));
    }

    /**
     * Returns true if the size of the context name is permitted of TINE.
     * 
     * @param name
     *            The name to be examined.
     */
    public boolean isContextNameSizeOK(String name) {
        return ((name != null) &&
            (name.length() > 0) && (name.length() <= TStrings.CONTEXT_NAME_SIZE));
    }

    /**
     * Returns true if the size of the alarm tag is permitted of TINE.
     * 
     * @param name
     *            The name to be examined.
     */
    public boolean isAlarmTagSizeOK(String name) {
        return ((name != null) &&
            (name.length() > 0) && (name.length() <= TStrings.ALARM_TAG_SIZE));
    }

    /**
     * Returns true if the size of the module name is permitted of TINE.
     * 
     * @param name
     *            The name to be examined.
     */
    public boolean isServerNameSizeOK(String name) {
        return ((name.length() > 0) && (name.length() <= TStrings.EXPORT_NAME_SIZE));
    }

    /**
     * Creates a TINE property name. Ensures that a name is short enough. Cuts
     * off characters which exceed the maximum size.
     * 
     * @param propertyName
     *            A property name.
     */
    public String generateTineProperty(String propertyName) {
        String result;
        result = new String(propertyName);
        if (result.length() > TStrings.PROPERTY_NAME_SIZE) {
            result = result.substring(0, TStrings.PROPERTY_NAME_SIZE - 1);
        }
        return result;
    }

    /**
     * Returns true if an alarm tag is syntactically correct and has a valid
     * length.
     * 
     * @param name
     *            An alarm tag.
     * @return True if the name is OK.
     */
    public boolean isAlarmTagValid(String name) {
        return isAlarmTagSizeOK(name) && isAlarmTagSyntaxOK(name);
    }

    /**
     * Returns true if an context name is syntactically correct and has a valid
     * length.
     * 
     * @param name
     *            A context name.
     * @return True if the name is OK.
     */
    public boolean isContextNameValid(String name) {
        return isContextNameSizeOK(name) && isContextNameSyntaxOK(name);
    }

    /**
     * Returns true if a server name is syntactically correct and has a valid
     * length.
     * 
     * @param name
     *            A server name.
     * @return True if the name is OK.
     */
    public boolean isServerNameValid(String name) {
        return isServerNameSizeOK(name) && isServerNameSyntaxOK(name);
    }

    /**
     * Returns true if a device name is syntactically correct and has a valid
     * length.
     * 
     * @param name
     *            A device name.
     * @return True if the name is OK.
     */
    public boolean isDeviceNameValid(String name) {
        return isDeviceNameSizeOK(name) && isDeviceNameSyntaxOK(name);
    }

    /**
     * Returns true if a property name is syntactically correct and has a valid
     * length.
     * 
     * @param name
     *            A property name.
     * @return True if the name is OK.
     */
    public boolean isPropertyNameValid(String name) {
        return isAlarmTagSizeOK(name) && isPropertyNameSyntaxOK(name);
    }
    
    /**
     * Checks if a TINE address is valid. Checks syntax, limits of field length,
     * and checks if there are null or empty components.
     * 
     * @param address
     *            TINE address.
     */
    public boolean isAddressValid(TLinkAddress address) {
        TineNameUtils nu = TineNameUtils.getInstance();
        return nu.isContextNameValid(address.getContext()) &&
            nu.isServerNameValid(address.getServer()) &&
            nu.isDeviceNameValid(address.getDevice()) &&
            nu.isPropertyNameValid(address.getProperty());
    }

    /**
     * Checks if a TINE address is valid.
     * 
     * @param fullPath
     *            TINE address as full path.
     */
    public boolean isAddressValid(String fullPath) {
        TLinkAddress address = new TLinkAddress(fullPath);
        return isAddressValid(address);
    }

}
