package de.desy.mst.libs.ser.tclientlib.writers;
import java.util.List;
import de.desy.tine.dataUtils.TDataType;
import de.desy.mst.libs.ser.tclientlib.ATineWriter;
import de.desy.mst.libs.ser.tclientlib.TineLinkException;
import de.desy.mst.libs.ser.tclientlib.TLinkAddress;

/**
 * An float array writer.
 *
 * The FloatArrayWriter writes an array of type float
 * to a TINE server.
 *
 */
public class FloatArrayWriter extends ATineWriter {
    
    private float[] outputBuffer = null; // Data buffer for TINE
    
    /**
     * Constructor.
     * Constructs a FloatArrayWriter. The array size is 
     * determined dynamically by a server request.
     * @param linkAddress The TINE address
     */
    public FloatArrayWriter(TLinkAddress linkAddress) {
        super(linkAddress);
    }

    /**
     * Constructor.
     * Constructs a FloatArrayWriter. The array size is 
     * determined dynamically by a server request.
     * @param fullDevicePath The full TINE path in the 
     *  form "/context/server/device/property"
     */
    public FloatArrayWriter(String fullDevicePath) {
        super(fullDevicePath);
    }

    
    /**
     * Writes an array of values to the server.
     * If the write operation was not successful,
     * a TineLinkException is thrown.
     * <BR><B>Note:</B> This method uses the size of
     * the given array. To determine which array size the
     * server expects, use getArraySize().
     * @throws TineLinkException
     */
    public synchronized void setValues(float[] newValues) throws TineLinkException {
        
        if (outputBuffer == null || (newValues.length != outputBuffer.length)) {
            outputBuffer = new float[newValues.length];
        }
        for (int i = 0; i< outputBuffer.length; i++) {
            outputBuffer[i] = newValues[i];
        }
        
        if (writeData(createTDataType()) != 0) {
            throw new TineLinkException(getLinkAddress(), getLinkStatus(), getErrorMessage());
        }
    }

    /**
     * Writes an array of values to the server.
     * This is an alternative to setValues(float[] newValues)
     * which takes a list of values instead an array.
     * If the write operation was not successful,
     * a TineLinkException is thrown.
     * <BR><B>Note:</B> This method uses the size of
     * the given array. To determine which array size the
     * server expects, use getArraySize().
     * @throws TineLinkException
     */
    public void setValueList(List<Float> newValues) throws TineLinkException {
        
        int listSize = newValues.size();
        if (outputBuffer == null || ( listSize != outputBuffer.length)) {
            outputBuffer = new float[listSize];
        }
        
        int i = 0;
        for (Float value : newValues) {
            outputBuffer[i] = value.floatValue();
            i++;
        }
        
        if (writeData(createTDataType()) != 0) {
            throw new TineLinkException(getLinkAddress(), getLinkStatus(), getErrorMessage());
        }
    }
    
    /**
     * Returns the array size.
     * This method returns the array size as 
     * specified or dynamically detected.
     */
    public int getArraySize() {
        int result = 0;
        if (outputBuffer != null) {
            result = outputBuffer.length;
        }
        return result;
    }
    
    private TDataType createTDataType() {
        TDataType result = null;
        result = new TDataType(outputBuffer);
        return result;
    }
    
}
