package de.desy.tine.apps.daqdata;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Map;

import javax.swing.JOptionPane;

import de.desy.acop.launcher.Launcher;
import de.desy.acop.launcher.Utilities;
import de.desy.tine.client.TLink;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TAccess;

/**
 * 
 * <code>Starter</code> is the main class which starts the DAQDataGUI. It expects
 * certain parameters to be supplied, then it creates an xml on the fly and starts
 * the jnlp.
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public class Starter {

	private static final String XML = 
			"<DAQREQ>\n" + 
			"   <TStart time=\"<start>\"/>\n" + 
			"   <TStop time=\"<stop>\"/>\n"	+ 
			"   <Exp name=\"<experiment>\"/>\n" + 
			"   <CDir name=\"<directory>\"/>\n" + 
			"<channels>" + 
			"</DAQREQ>";
	private static final String MAIN_XML = 
			"<DAQGUICONF>\n" +
			"   <DATE date='<date>'/>\n" +
			"   <CURDIR dir='<curdir>'/>\n"+
			"   <SVRADDR addr='<address>/'/>\n" +
			"   <INPUTFILE name='<file>'/>\n" +
//			"   <TIMEPERIOD min='<time>'/>\n" +
			"   <CYCLEPERIOD sec='<cycle>'/>\n" +
			"</DAQGUICONF>";
	
	public static final String TIMESTAMP = "time";
	public static final String START_TIME = "startTime";
	public static final String END_TIME = "stopTime";
	public static final String CHANNELS = "channels";
	public static final String EXPERIMENT = "experiment";
	public static final String DIRECTORY = "directory";
	public static final String DESTINATION = "destinationFile";
	
	public static final String TIME_PERIOD = "timePeriod";
	public static final String CYCLE_PERIOD = "cyclePeriod";
	public static final String ADDRESS = "address";
	
	public static final String P_DAQ_CONF_DIR = "-daqconfdir";
	public static final String P_DAQ_CONF = "-conf";
	public static final String P_DAQ_DESCRI = "-daqdescr";
	
	public static final String SP_DAQ_CONF_DIR = "daqconfdir";
	public static final String SP_DAQ_CONF = "conf";
	public static final String SP_DAQ_DESCRI = "daqdescriptionfile";
		
	public static final String JNLP_URL = "jnlpURL"; 

	public static void main(String[] args) {
		start(args);
	}
	
	/**
	 * Starts the DAQDataGUI using the specified parameters. Any of the fields in this class to be
	 * used as a parameter. If not provided, the default values will be used. For details see
	 * {@link #generateMainXML(Map)} and {@link #generateDataXML(Map)}.
	 * 
	 * @param args the parameters in form (key=value)
	 */
	public static void start(String[] args) {
		try {
			Map<String,String> parameters = Utilities.argsToParameters(args,System.getProperties());
//			String ver = parameters.get("java.runtime.version");
//			String[] versions = ver.split("[.]");
//			if (Integer.parseInt(versions[0]) == 1 && Integer.parseInt(versions[1]) < 7) {
//				JOptionPane.showMessageDialog(null,"Startup parameters will only work in Java 1.7.0 or later.","Java Version Problem",JOptionPane.WARNING_MESSAGE);
//			}
			generateDataXML(parameters);
			File startupFile = generateMainXML(parameters);
			
			String confDir = startupFile.getParent();
			String conf = startupFile.getName();
			confDir = confDir.replace("\\","/");
			
			String url = parameters.get(JNLP_URL);
			if (url == null) {
				url = "http://ttfinfo2.desy.de/common/applications/DAQdataGUI-experimental_arg_jnlp.jsp";
			}
//			String[] jvmparam = {SP_DAQ_CONF+"="+conf, "\""+SP_DAQ_CONF_DIR+"="+confDir + "\""};			
			url = url+'?'+SP_DAQ_CONF + '='+conf+'&' + SP_DAQ_CONF_DIR + '='+confDir;
			if (System.getProperty("os.name").toLowerCase().contains("win")) {
				url = '"' + url + '"';
			} 
			Launcher.runWebstartApplication(url,false);
		}
		catch (IOException e) {
			JOptionPane.showMessageDialog(null,"Cannot start the DataDAQGUI. Could not create the XML file.","Invalid Parameters",
					JOptionPane.ERROR_MESSAGE);
		} catch (Exception e) {
			JOptionPane.showMessageDialog(null,"Cannot start the DataDAQGUI. Parameters are not valid.","Invalid Parameters",
					JOptionPane.ERROR_MESSAGE);
		}
	}
	
	/**
	 * Generates the application startup xml using the content of the parameters map. The 
	 * parameters are expected to provide the {@value #DESTINATION}. Other parameters are
	 * optional and the default values will be used if not provided ({@value #ADDRESS},
	 * {@value #TIME_PERIOD}, {@value #CYCLE_PERIOD}).
	 * <p>
	 * The method returns the File into which the data was stored.
	 * 
	 * @param parameters the content
	 * @return the File into which the data was stored
	 * @throws IOException if there was an error writing the xml or extracting the parameters
	 */
	private static File generateMainXML(Map<String,String> parameters) throws IOException {
		long timePeriod = 12;
		long cyclePeriod = 0;
		if (parameters.get(TIME_PERIOD) != null) {
			timePeriod = Long.parseLong(parameters.get(TIME_PERIOD));
		}
		if (parameters.get(CYCLE_PERIOD) != null) {
			cyclePeriod = Long.parseLong(parameters.get(CYCLE_PERIOD));
		}
		String destination = parameters.get(DESTINATION).replace("\\","/");
		String address = "TTF2.DAQ/DAQ.SERVER1/DAQ.DATA.SVR/:TTF2.DAQ/DAQ.SERVER3/DAQ.DATA.SVR/:TTF2.DAQ/DAQ.SERVER5/DAQ.DATA.SVR/:TTF2.DAQ/DAQ.SERVER7/DAQ.DATA.SVR/:TTF2.DAQ/DAQ.SERVER2/DAQ.DATA.SVR/:TTF2.DAQ/DAQ.SERVER4/DAQ.DATA.SVR/:TTF2.DAQ/DAQ.SERVER6/DAQ.DATA.SVR/:TTF2.DAQ/DAQ.SERVER8/DAQ.DATA.SVR/";
		if (parameters.get(ADDRESS) != null) {
			address = parameters.get(ADDRESS);
		}
		String currentDir = new File(".").getAbsoluteFile().getParentFile().getAbsolutePath().replace("\\","/");
		
		SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
		String xml = MAIN_XML.replace("<date>",formatter.format(new Date()));
		xml = xml.replace("<time>",String.valueOf(timePeriod));
		xml = xml.replace("<cycle>",String.valueOf(cyclePeriod));
		xml = xml.replace("<file>",destination);
		xml = xml.replace("<address>",address);
		xml = xml.replace("<curdir>",currentDir);
		
		File file = File.createTempFile("DAQStarter",".xml");
		
		
		PrintWriter pw = new PrintWriter(file);
		pw.println(xml);
		pw.close();
		
		return file;
	}
	
	/**
	 * Generates the data xml using the content of the parameters map. The map is expected
	 * to provide the {@value #START_TIME}, {@value #END_TIME}, {@value #CHANNELS}.
	 * {@value #EXPERIMENT} and {@value #DIRECTORY} are optional.
	 * 
	 * @param parameters the parameters used for the xml content
	 * @throws Exception if there was an error in writing the xml or extracting the contents
	 */
	private static void generateDataXML(Map<String,String> parameters) throws Exception {
		long time = Long.parseLong(parameters.get(TIMESTAMP));
		long start = time - 15*60000;//Long.parseLong(parameters.get(START_TIME));
		long stop = time + 15*60000;//Long.parseLong(parameters.get(END_TIME));
		String[] channels = parameters.get(CHANNELS).split(",");
		String experiment = parameters.get(EXPERIMENT);
		if (experiment == null)
			experiment = "linac";
		String dir = parameters.get(DIRECTORY);
		if (dir == null)
			dir = "/daq/ttf2/admtemp/";
		StringBuilder sb = new StringBuilder();
		for (String c : channels) {
			if (c == null || c.length() == 0) continue;
			sb.append("   <Chan name=\"");
			sb.append(toDAQName(c));
			sb.append("\"/>\n");
		}
		SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
		String xml = XML.replace("<start>",formatter.format(new Date(start)));
		xml = xml.replace("<stop>",formatter.format(new Date(stop)));
		xml = xml.replace("<experiment>",experiment);
		xml = xml.replace("<directory>",dir);
		xml = xml.replace("<channels>",sb.toString());
		File file = File.createTempFile("DAQData",".xml");
		parameters.put(DESTINATION,file.getAbsolutePath());
		
		PrintWriter pw = new PrintWriter(file);
		pw.println(xml);
		pw.close();
	}

	private static String toDAQName(String channel) {
		char[] daqchr = new char[128];
		TLink d2dlnk = new TLink("/FLASH/DAQCrossReferencer","DAQChannel",new TDataType(daqchr),
				new TDataType(channel),TAccess.CA_READ);
		if(d2dlnk.executeAndClose() == 0) {
			return new String(daqchr).trim();
		} 
		return channel;
	}
}
